
#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <D3DX10math.h>
#include <assert.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include "Utility.h"
#include "OffsetMap.h"


OffsetMap::OffsetMap()
    :   m_pTex(NULL),
        m_pSRV(NULL),
        m_pSampler(NULL),
        m_MapMask(0)
{

}

OffsetMap::~OffsetMap()
{
    Release();
}

HRESULT OffsetMap::Set(ID3D11DeviceContext* pd3dImmediateContext, UINT iShaderSlot)
{
    HRESULT hr = S_OK;
    assert(m_MapMask != kMapNone); // is this really what you meant?

    if(m_MapMask & kMapPS)
    {
        pd3dImmediateContext->PSSetShaderResources( iShaderSlot, 1, &m_pSRV );
        pd3dImmediateContext->PSSetSamplers( iShaderSlot, 1, &m_pSampler);
    }
    if(m_MapMask & kMapVS)
    {
        pd3dImmediateContext->VSSetShaderResources( iShaderSlot, 1, &m_pSRV );
        pd3dImmediateContext->VSSetSamplers( iShaderSlot, 1, &m_pSampler);
    }
    if(m_MapMask & kMapGS)
    {
        pd3dImmediateContext->GSSetShaderResources( iShaderSlot, 1, &m_pSRV );
        pd3dImmediateContext->GSSetSamplers( iShaderSlot, 1, &m_pSampler);
    }

    return hr;
}

HRESULT OffsetMap::Create( ID3D11Device* pd3dDevice, int length )
{
    HRESULT hr = S_OK;

    Release();

    int totalLength = length * length;
    D3DXVECTOR2* offsets = GetRandomOffsets(totalLength);

    // Create the texture with the data pass in
    D3D11_TEXTURE2D_DESC dstex;
    dstex.Width = length;
    dstex.Height = length;
    dstex.MipLevels = 1;
    dstex.SampleDesc.Count = 1;
    dstex.SampleDesc.Quality = 0;
    dstex.Format = DXGI_FORMAT_R32G32_FLOAT;
    dstex.Usage = D3D11_USAGE_DEFAULT;
    dstex.BindFlags = D3D11_BIND_SHADER_RESOURCE;
    dstex.CPUAccessFlags = 0;
    dstex.MiscFlags = 0;
    dstex.ArraySize = 1;
    D3D11_SUBRESOURCE_DATA InitData;
    InitData.pSysMem = offsets;
    InitData.SysMemPitch = dstex.Width * sizeof( *offsets );
    InitData.SysMemSlicePitch = 0; // 1024 * sizeof( float );
    V_RETURN( pd3dDevice->CreateTexture2D( &dstex, &InitData, &m_pTex ) );

    // Create the resource view so that we can bind this
    D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    ZeroMemory( &SRVDesc, sizeof( SRVDesc ) );
    SRVDesc.Format = dstex.Format;
    SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
    SRVDesc.Texture2D.MipLevels = dstex.MipLevels;
    V_RETURN( pd3dDevice->CreateShaderResourceView( m_pTex, &SRVDesc, &m_pSRV) );

    D3D11_SAMPLER_DESC sampDesc;
    ZeroMemory( &sampDesc, sizeof(sampDesc) );
    sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_POINT;
    sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
    sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
    sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
    sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
    sampDesc.MinLOD = 0;
    sampDesc.MaxLOD = 0;
    V_RETURN( pd3dDevice->CreateSamplerState( &sampDesc, &m_pSampler ) );


    SAFE_DELETE_ARRAY(offsets);

    return hr;
}

void OffsetMap::Release()
{
    SAFE_RELEASE(m_pSRV);
    SAFE_RELEASE(m_pTex);
    SAFE_RELEASE(m_pSampler);
}

int OffsetMap::GetLength() const
{
    D3D11_TEXTURE2D_DESC desc;
    m_pTex->GetDesc(&desc);
    
    return desc.Width;
}

D3DXVECTOR2* const OffsetMap::GetRandomOffsets( int length ) const
{
    D3DXVECTOR2* offsets = new D3DXVECTOR2[length];

    for(int i = 0; i < length; ++i)
    {
        //const float x = Halton(i, 2);
        //const float y = Halton(i, 3);
        //offsets[i] = D3DXVECTOR2(x - 0.5f, y - 0.5f);
//        offsets[i] = D3DXVECTOR2(0.5f, 0.5f);
//        offsets[i] = D3DXVECTOR2(float(i) / float(length), 0.0f);
        offsets[i] = D3DXVECTOR2(
            float(double(rand()) / double(RAND_MAX)) - 0.5f,
            float(double(rand()) / double(RAND_MAX)) - 0.5f);
    }
    return offsets;
}

float OffsetMap::Halton(int idx, int base) const
{
    float ret = 0;
    double denom = 1.0 / double(base);
    while(idx > 0)
    {
        ret += float(denom * float(idx % base));
        idx = idx / base;
        denom = denom / base;
    }
    return ret;
}

