

#include <windows.h>
#include <d3dx11.h>
#include <d3dcompiler.h>

#include "DXUT.h"
#include "../Utility/CFlyCamera.h"
#include "../Utility/CompiledShader.h"

#include "DrawBox.h"

#include "GridObject.h" // for GetEnableMSAA()


extern CFlyCamera          g_Camera;               // A model viewing camera

struct BoxConstantBuffer
{
    D3DXMATRIX  localToNDC;
    D3DXVECTOR4 color;
};

struct BoxVertex : public D3DXVECTOR3
{
};

DrawBox::DrawBox()
{
    s_pRasterizerState = NULL;

    s_pVertexShader = NULL;
    s_pPixelShader = NULL;
    s_pVertexLayout = NULL;
    s_pVertexBuffer = NULL;
    s_pIndexBuffer = NULL;
    s_pConstantBuffer = NULL;
}
DrawBox::~DrawBox()
{
    Cleanup();
}

void DrawBox::SetState(ID3D11DeviceContext* pImmediateContext)
{
    pImmediateContext->IASetInputLayout( s_pVertexLayout );

    // Set vertex buffer
    UINT stride = sizeof( BoxVertex );
    UINT offset = 0;
    pImmediateContext->IASetVertexBuffers( 0, 1, &s_pVertexBuffer, &stride, &offset );

    // Set index buffer
    pImmediateContext->IASetIndexBuffer( s_pIndexBuffer, DXGI_FORMAT_R16_UINT, 0 );

    // Set primitive topology
    pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_LINELIST );

	pImmediateContext->RSSetState(s_pRasterizerState);
}

void DrawBox::SetConstants(ID3D11DeviceContext* pImmediateContext, const D3DXMATRIX& l2w)
{
    const D3DXMATRIX worldToCamera = *g_Camera.GetViewMatrix();
    const D3DXMATRIX cameraToNDC = *g_Camera.GetProjMatrix();
    const D3DXMATRIX localToNDC = l2w * worldToCamera * cameraToNDC;

    BoxConstantBuffer cb;
	D3DXMatrixTranspose( &cb.localToNDC, &localToNDC);

    cb.color = D3DXVECTOR4(1.0f, 0.5f, 0.5f, 1.0f);

	pImmediateContext->UpdateSubresource( s_pConstantBuffer, 0, NULL, &cb, 0, 0 );
}

void DrawBox::SetVerts(const D3DXVECTOR3* const pts)
{
 	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
    D3D11_MAPPED_SUBRESOURCE MappedSubresource;
    pImmediateContext->Map( s_pVertexBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedSubresource );

    assert(sizeof(D3DXVECTOR3) == sizeof(BoxVertex));
    memcpy(MappedSubresource.pData, pts, 8 * sizeof(D3DXVECTOR3));

    pImmediateContext->Unmap( s_pVertexBuffer, 0 );
}

void DrawBox::InternalRender(const D3DXVECTOR3* const pts, const D3DXMATRIX& l2w)
{
 	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

    SetVerts(pts);

    SetState(pImmediateContext);

    SetConstants(pImmediateContext, l2w);

	pImmediateContext->VSSetShader( s_pVertexShader, NULL, 0 );
	pImmediateContext->VSSetConstantBuffers( 0, 1, &s_pConstantBuffer );
	pImmediateContext->PSSetShader( s_pPixelShader, NULL, 0 );
	pImmediateContext->DrawIndexed( 24, 0, 0 );
   
}

HRESULT DrawBox::MakeState()
{
    D3D11_RASTERIZER_DESC drd = {
        D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
        D3D11_CULL_NONE,//D3D11_CULL_MODE CullMode;
        FALSE, //BOOL FrontCounterClockwise;
        0, //INT DepthBias;
        0.0f,//FLOAT DepthBiasClamp;
        0.0f,//FLOAT SlopeScaledDepthBias;
        TRUE,//BOOL DepthClipEnable;
        FALSE,//BOOL ScissorEnable;
        GridObject::GetEnableMSAA(),//BOOL MultisampleEnable;
        FALSE//BOOL AntialiasedLineEnable;        
    };

    ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    HRESULT hr = pd3dDevice->CreateRasterizerState(&drd, &s_pRasterizerState);
    assert(SUCCEEDED(hr));

    return hr;
}

HRESULT DrawBox::MakeShaders()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    const WCHAR* fileName = L"Shaders/DrawBox.fx";

    // Create the vertex shader
    CompiledShader vShader(fileName, L"vs_4_0", L"VS_Def");
    HRESULT hr = pd3dDevice->CreateVertexShader( vShader.GetBufferPointer(), vShader.GetBufferSize(), NULL, &s_pVertexShader );
    assert(SUCCEEDED(hr));

    CompiledShader pShader(fileName, L"ps_4_0", L"PS_Def");
    hr = pd3dDevice->CreatePixelShader( pShader.GetBufferPointer(), pShader.GetBufferSize(), NULL, &s_pPixelShader );
    assert(SUCCEEDED(hr));

    // Define the input layout
    D3D11_INPUT_ELEMENT_DESC layout[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
	};
	UINT numElements = ARRAYSIZE( layout );

    // Create the input layout
	hr = pd3dDevice->CreateInputLayout( layout, numElements, vShader.GetBufferPointer(),
                                          vShader.GetBufferSize(), &s_pVertexLayout );
    assert(SUCCEEDED(hr));

    return hr;
}
HRESULT DrawBox::MakeGeometry()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    HRESULT hr = S_OK;

    D3D11_BUFFER_DESC bd;
	ZeroMemory( &bd, sizeof(bd) );
    bd.Usage = D3D11_USAGE_DYNAMIC;
    bd.ByteWidth = sizeof( BoxVertex ) * 8;
    bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &s_pVertexBuffer );
    assert(SUCCEEDED(hr));

    // Create index buffer
    WORD indices[] =
    {
        0, 1,
        0, 2,
        1, 3,
        2, 3,

        0, 4,
        1, 5,
        2, 6,
        3, 7,

        4, 5,
        4, 6,
        5, 7,
        6, 7,
    };
    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof( WORD ) * 24;        // 36 vertices needed for 12 triangles in a triangle list
    bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
	bd.CPUAccessFlags = 0;
    D3D11_SUBRESOURCE_DATA InitData;
	ZeroMemory( &InitData, sizeof(InitData) );
    InitData.pSysMem = indices;
    hr = pd3dDevice->CreateBuffer( &bd, &InitData, &s_pIndexBuffer );
    assert(SUCCEEDED(hr));

    return hr;
}
HRESULT DrawBox::MakeConstants()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;


    bd.ByteWidth = sizeof(BoxConstantBuffer);
    HRESULT hr = pd3dDevice->CreateBuffer( &bd, NULL, &s_pConstantBuffer);
    assert(SUCCEEDED(hr));    
    return hr;
}
HRESULT DrawBox::InternalInit()
{
    HRESULT hr = S_OK;
    V_RETURN(MakeState());
    V_RETURN(MakeShaders());
    V_RETURN(MakeGeometry());
    V_RETURN(MakeConstants());
    return hr;
}

void DrawBox::InternalCleanup()
{
	SAFE_RELEASE(s_pRasterizerState);

    SAFE_RELEASE(s_pVertexShader);
    SAFE_RELEASE(s_pPixelShader);
    SAFE_RELEASE(s_pVertexLayout);
    SAFE_RELEASE(s_pVertexBuffer);
    SAFE_RELEASE(s_pIndexBuffer);
    SAFE_RELEASE(s_pConstantBuffer);
}

