
#include <windows.h>
#include <d3dx11.h>
#include <d3dcompiler.h>

#include <map>

#include "DXUT.h"
#include "../Utility/CFlyCamera.h"
#include "../Utility/CompiledShader.h"

#define _USE_MATH_DEFINES
#include <math.h>

#include <DXUT.h>

#include "GridObject.h"

#ifndef OBB_BOUNDS
#define MAF_TEST_SPHERE
#ifdef MAF_TEST_SPHERE
#include "Sphere.h"
static Sphere* sphere = NULL;
#endif // MAF_TEST_SPHERE

#else // OBB_BOUNDS

#define MAF_TEST_BOX
#ifdef MAF_TEST_BOX
#include "DrawBox.h"
#endif // MAF_TEST_BOX

#endif // OBB_BOUNDS

typedef std::map<const WCHAR*, ID3D11VertexShader*> VertexShaderMap;
typedef VertexShaderMap::iterator VertexShaderIter;
typedef std::pair<const WCHAR*, ID3D11VertexShader*> VertexShaderEntry;

typedef std::map<const WCHAR*, ID3D11PixelShader*> PixelShaderMap;
typedef PixelShaderMap::iterator PixelShaderIter;
typedef std::pair<const WCHAR*, ID3D11PixelShader*> PixelShaderEntry;

static VertexShaderMap vertShaderLUT;
static PixelShaderMap pixelShaderLUT;

extern CFlyCamera          g_Camera;               // A model viewing camera

bool GridObject::bShowBounds = false;
bool GridObject::bWireFrame = false;
bool GridObject::bLights = true;
bool GridObject::bSpecular = true;
bool GridObject::bEnvMap = true;
float GridObject::tessFactor = 32.0f;
bool GridObject::powerOfTwo = true;
bool GridObject::lodByDistance = true;
int GridObject::globalNumTris = 512;
float GridObject::detailScale = 0.01f;
float GridObject::highNumTris = 50000.0f;
float GridObject::lowNumTris = 200.0f;
float GridObject::highDistance = 10.0f;
float GridObject::lowDistance = 100.0f;
int GridObject::msaa = 4;
D3DXVECTOR2 GridObject::windowSize(0.0f, 0.0f);
float GridObject::depthTolerance = 0.01f;
float GridObject::frameDecay = 0.86f;
bool GridObject::reprojEnabled = true;
bool GridObject::useLast = true;

ID3D11RasterizerState*              GridObject::pRasterizerState = NULL;

ID3D11ShaderResourceView*           GridObject::cubeMap = NULL;
ID3D11ShaderResourceView*           GridObject::cubeBlack = NULL;
ID3D11SamplerState*                 GridObject::cubeSampler = NULL;

ID3D11Texture2D*                    GridObject::pPermTexture = NULL;          // random texture for noise lookups
ID3D11ShaderResourceView*           GridObject::pPermTextureSRV = NULL;		
ID3D11Texture1D*                    GridObject::pRandomVecTexture = NULL;     // random vector texture for voronoi noise
ID3D11ShaderResourceView*           GridObject::pRandomVecTextureSRV = NULL;
ID3D11SamplerState*                 GridObject::rndByteSampler = NULL;
ID3D11SamplerState*                 GridObject::rndVecSampler = NULL;

ID3D11Buffer*                         GridObject::everyFrameCB = NULL;

GridObject::GridObject()
	:	vertexShader(NULL),
		pixelShader(NULL),
		everyObjectCB(NULL),
		minNumTris(512),
		fStepTime(0.0f),
		fCurrentTime(1.0f)
{
	D3DXMatrixIdentity(&localToWorld);
    ZeroMemory(&lastLocalToWorld, sizeof(lastLocalToWorld));

	randomOffset = float(rand()) / float(RAND_MAX);
}

void GridObject::RefreshState()
{
    SAFE_RELEASE(pRasterizerState);
}
void GridObject::SetWireFrame(bool b) 
{ 
    bWireFrame = b; 
    RefreshState();
}

void GridObject::SetMSAA(int m)
{
    msaa = max(1, min(8, m));
    RefreshState();
}



HRESULT GridObject::GetVertexShader(const WCHAR* fileName, ID3D11VertexShader*& vShader)
{
	HRESULT hr = S_OK;

	VertexShaderIter iter = vertShaderLUT.find(fileName);
	if(iter != vertShaderLUT.end())
	{
		vShader = iter->second;
		vShader->AddRef();
		return S_OK;
	}

    // Compile the vertex shader
	WCHAR entryPoint[256];
	swprintf_s(entryPoint, L"VS_%s", WideShaderExtention());
	WCHAR shaderModel[256];
	swprintf_s(shaderModel, L"vs_%d_0", ShaderModel());

	CompiledShader shader(fileName, shaderModel, entryPoint);

    // Create the vertex shader
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreateVertexShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &vShader );

	vertShaderLUT.insert(VertexShaderEntry(fileName, vShader));
	return hr;
}

HRESULT GridObject::GetPixelShader(const WCHAR* fileName, ID3D11PixelShader*& pShader)
{
	HRESULT hr = S_OK;

	PixelShaderIter iter = pixelShaderLUT.find(fileName);
	if(iter != pixelShaderLUT.end())
	{
		pShader = iter->second;
		pShader->AddRef();
		return S_OK;
	}

	WCHAR entryPoint[256];
	swprintf_s(entryPoint, L"PS_%s", WideShaderExtention());
	WCHAR shaderModel[256];
	swprintf_s(shaderModel, L"ps_%d_0", ShaderModel());

	CompiledShader shader(fileName, shaderModel, entryPoint);

    // Create the pixel shader
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreatePixelShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &pShader );

	pixelShaderLUT.insert(PixelShaderEntry(fileName, pShader));

	return hr;
}

HRESULT GridObject::GetStates()
{
	HRESULT hr = S_OK;

    if(!pRasterizerState)
    {
        D3D11_RASTERIZER_DESC drd = {
            D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
            D3D11_CULL_BACK,//D3D11_CULL_MODE CullMode;
            FALSE, //BOOL FrontCounterClockwise;
            0, //INT DepthBias;
            0.0f,//FLOAT DepthBiasClamp;
            0.0f,//FLOAT SlopeScaledDepthBias;
            TRUE,//BOOL DepthClipEnable;
            FALSE,//BOOL ScissorEnable;
            GetEnableMSAA(),//BOOL MultisampleEnable;
            FALSE//BOOL AntialiasedLineEnable;        
        };
        if(bWireFrame)
        {
	        drd.FillMode = D3D11_FILL_WIREFRAME;
        }

        ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
        hr = pd3dDevice->CreateRasterizerState(&drd, &pRasterizerState);
	    if( FAILED( hr ) )
		    return hr;
    }

	return hr;
}

HRESULT GridObject::GetConstantBuffers()
{
	HRESULT hr = S_OK;

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    if(!everyFrameCB)
    {
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.CPUAccessFlags = 0;
    
        bd.ByteWidth = sizeof(CBChangesEveryFrame);
        hr = pd3dDevice->CreateBuffer( &bd, NULL, &everyFrameCB);
        if( FAILED( hr ) )
            return hr;
    }
    assert(!everyObjectCB);
    if(!everyObjectCB)
    {
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.CPUAccessFlags = 0;

        bd.ByteWidth = sizeof(CBChangesEveryObject);
        hr = pd3dDevice->CreateBuffer( &bd, NULL, &everyObjectCB);
        if( FAILED( hr ) )
            return hr;
    }

	return hr;
}

HRESULT GridObject::GetTextures()
{
	HRESULT hr = S_OK;
	/// Disable load and set of textures, we aren't using them currently.
#ifdef LOAD_TEXTURES

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Load the Texture
	if(!cubeMap)
	{
		//hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/skybox02.dds", NULL, NULL, &cubeMap, NULL );
        //hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/Hock_Sky.dds", NULL, NULL, &cubeMap, NULL );
        hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/OutputCube.dds", NULL, NULL, &cubeMap, NULL );
        //hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/Black_Cube_128.dds", NULL, NULL, &cubeMap, NULL );
		if( FAILED( hr ) )
			return hr;
	}

    // Load the Texture
	if(!cubeBlack)
	{
		hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/Black_Cube_128.dds", NULL, NULL, &cubeBlack, NULL );
		if( FAILED( hr ) )
			return hr;
	}

    // Create the sample state
	if(!cubeSampler)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &cubeSampler );
		if( FAILED(hr) )
			return hr;
	}

    // Create the sample state
	if(!rndByteSampler)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &rndByteSampler );
		if( FAILED(hr) )
			return hr;
	}

	    // Create the sample state
	if(!rndVecSampler)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &rndVecSampler );
		if( FAILED(hr) )
			return hr;
	}

	hr = CreateRandomPermTexture( pd3dDevice );
	if(FAILED(hr))
		return hr;
	hr = CreateRandomVectorTexture( pd3dDevice );
	if(FAILED(hr))
		return hr;

#endif
	return hr;
}

HRESULT GridObject::Init()
{
	HRESULT hr = S_OK;

	// Compile and create the effect.

	hr = GetVertexShader(ShaderFile(), vertexShader);
    if( FAILED( hr ) )
        return hr;

	hr = GetPixelShader(ShaderFile(), pixelShader);
    if( FAILED( hr ) )
        return hr;

	hr = MakeGeometry();
	if(FAILED(hr))
		return hr;

	hr = GetConstantBuffers();
    if( FAILED( hr ) )
        return hr;

	hr = GetTextures();
    if( FAILED( hr ) )
        return hr;

	return S_OK;
}

HRESULT GridObject::MakeGeometry()
{
	HRESULT hr = S_OK;


	return hr;
}

void GridObject::OnResize()
{
}

void GridObject::Cleanup()
{
	vertShaderLUT.clear();
	pixelShaderLUT.clear();

	SAFE_RELEASE(vertexShader);
	SAFE_RELEASE(pixelShader);

	SAFE_RELEASE(everyFrameCB);
    SAFE_RELEASE(everyObjectCB);

#ifdef LOAD_TEXTURES
	SAFE_RELEASE(cubeSampler);
	SAFE_RELEASE(cubeMap);
	SAFE_RELEASE(cubeBlack);
    SAFE_RELEASE( pPermTextureSRV );
    SAFE_RELEASE( pPermTexture );
    SAFE_RELEASE( pRandomVecTexture );
    SAFE_RELEASE( pRandomVecTextureSRV );
	SAFE_RELEASE(rndByteSampler);
	SAFE_RELEASE(rndVecSampler);
#endif // LOAD_TEXTURES

	SAFE_RELEASE(pRasterizerState);

#ifdef MAF_TEST_SPHERE
    SAFE_DELETE(sphere);
#endif // MAF_TEST_SPHERE
}

HRESULT GridObject::SetEveryFrame()
{
    D3DXMATRIX worldToCamera = *g_Camera.GetViewMatrix();
    D3DXMATRIX cameraToNDC = *g_Camera.GetProjMatrix();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

    if( everyFrameCB )
	{
		CBChangesEveryFrame cb;

		D3DXMatrixTranspose(&cb.mCameraToNDC, &cameraToNDC);

		//RECT rc = DXUTGetWindowClientRect();
		//UINT width = rc.right - rc.left;
		//UINT height = rc.bottom - rc.top;
        const float width = windowSize.x;
        const float height = windowSize.y;
		cb.Details = D3DXVECTOR4(detailScale, 0.0f, FLOAT(width), FLOAT(height));

        const float reprojDecay = reprojEnabled ? frameDecay : 0.0f;
        cb.WindowSize = D3DXVECTOR4(windowSize.x, windowSize.y, depthTolerance, reprojDecay);


			/// minDiffuse, mulSpec, specPower, 1.0/range
		cb.light = D3DXVECTOR4(1.0f, 1.0, 100.0f, 0.025f);
		if(!bLights)
			cb.light.x = 0.0f;
		if(!bSpecular)
			cb.light.y = 0.0f;

		D3DXMatrixTranspose( &cb.mWorldToCamera, &worldToCamera );

		pImmediateContext->UpdateSubresource( everyFrameCB, 0, NULL, &cb, 0, 0 );

        pImmediateContext->VSSetConstantBuffers( 9, 1, &everyFrameCB );
        pImmediateContext->PSSetConstantBuffers( 9, 1, &everyFrameCB );
    }

	return S_OK;
}

HRESULT GridObject::SetEveryObject()
{
    CBChangesEveryObject cb;

    D3DXMatrixTranspose(&cb.mLocalToWorld, &localToWorld);

    cb.ModDivScale = ModDivScale();

    const D3DXMATRIX lastLocalToNDC = useLast 
        ? (lastLocalToWorld * g_Camera.GetLastWorldToNDC())
        : (localToWorld * g_Camera.GetLastWorldToNDC());
    D3DXMatrixTranspose(&cb.mLastLocalToNDC, &lastLocalToNDC);

    D3DXVECTOR3 origin = D3DXVECTOR3(0.0, 0.0, 0.0);
    D3DXMATRIX cameraToWorld;
    D3DXMatrixInverse(&cameraToWorld, NULL, g_Camera.GetViewMatrix());
    D3DXVECTOR4 eyePos;
    D3DXVec3Transform(&eyePos, &origin, &cameraToWorld);
    eyePos.w = fCurrentTime * RandomOffset(0.5f, 1.0f);
    cb.WorldCamPos = eyePos;

    ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

    pImmediateContext->UpdateSubresource( everyObjectCB, 0, NULL, &cb, 0, 0 );

    return S_OK;
}
void GridObject::UpdateConstants()
{
	SetLOD();

	SetEveryObject();
}

void GridObject::SetBuffers()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	// Set the input layout
	pImmediateContext->IASetInputLayout( NULL );
	pImmediateContext->IASetVertexBuffers( 0, 0, NULL, NULL, NULL );

    // Set primitive topology
    pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

}

void GridObject::SetNoiseResources()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
#ifdef LOAD_TEXTURES
    pImmediateContext->PSSetSamplers( 1, 1, &rndByteSampler );
    pImmediateContext->PSSetSamplers( 2, 1, &rndVecSampler );
	pImmediateContext->PSSetShaderResources( 1, 1, &pPermTextureSRV );
	pImmediateContext->PSSetShaderResources( 2, 1, &pRandomVecTextureSRV );
#endif // LOAD_TEXTURES
}

void GridObject::SetResources()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	pImmediateContext->VSSetShader( vertexShader, NULL, 0 );
	pImmediateContext->PSSetShader( pixelShader, NULL, 0 );

    //pImmediateContext->VSSetConstantBuffers( 0, 1, &everyFrameCB );
    //pImmediateContext->PSSetConstantBuffers( 0, 1, &everyFrameCB );

    pImmediateContext->VSSetConstantBuffers( 1, 1, &everyObjectCB );
    pImmediateContext->PSSetConstantBuffers( 1, 1, &everyObjectCB );

    
#ifdef LOAD_TEXTURES
	if(bEnvMap)
	{
		pImmediateContext->PSSetShaderResources( 0, 1, &cubeMap );
	}
	else
	{
	    pImmediateContext->PSSetShaderResources( 0, 1, &cubeBlack );
	}
    pImmediateContext->PSSetSamplers( 0, 1, &cubeSampler );
#endif // LOAD_TEXTURES

}

void GridObject::SetState()
{
    GetStates();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	
	ID3D11RasterizerState* rs = pRasterizerState;
	pImmediateContext->RSSetState(rs);
}

void GridObject::SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi)
{
#ifndef OBB_BOUNDS
    localBounds.Position = (lo + hi) * 0.5f;
    D3DXVECTOR3 diff = hi - localBounds.Position;
    localBounds.Radius = D3DXVec3Length(&diff);
    //sphereBound.Position = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
    //sphereBound.Radius = 1.45f;
#else // OBB_BOUNDS
    localBounds = AABB(lo, hi);
#endif // OBB_BOUNDS
}

void GridObject::SetLocalToWorld(const D3DXMATRIX& l2w)
{ 
    lastLocalToWorld = localToWorld;

    localToWorld = l2w; 
    UpdateBounds(); 
}
void GridObject::UpdateBounds()
{
//#ifndef OBB_BOUNDS
//    localBounds.Radius = 1.45f;
//    //sphereBound.Position = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
//    //sphereBound.Radius = 1.45f;
//#else // OBB_BOUNDS
//    localBounds = OBB(D3DXVECTOR3(-1.45f, -1.45f, -1.45f), D3DXVECTOR3(1.45f, 1.45f, 1.45f));
//#endif // OBB_BOUNDS
    worldBounds = OBB(localBounds);
    worldBounds.Transform(localToWorld);
}

int GridObject::Render()
{
	UpdateConstants();

	SetResources();

    SetNoiseResources();

	SetBuffers();

	SetState();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	const int numTris = NumTris();
	pImmediateContext->Draw( numTris * 3, 0 );

    if(bShowBounds)
    {
#ifdef MAF_TEST_SPHERE
        if(!sphere)
        {
            sphere = new Sphere();
            sphere->Init();
            sphere->SetRadius(localBounds.Radius);
        }
        sphere->OnResize();

        sphere->SetLocalToWorld(GetLocalToWorld());
        sphere->Render();
#endif // MAF_TEST_SPHERE
#ifdef MAF_TEST_BOX
        OBB obb(localBounds);
        DrawBox::Render(obb.Corners, GetLocalToWorld());
#endif // MAF_TEST_BOX
    }

	return numTris;
}

float GridObject::GetDistanceToCamera()
{
    D3DXMATRIX worldToCamera = *g_Camera.GetViewMatrix();
	D3DXVECTOR3 origin = D3DXVECTOR3(0.0, 0.0, 0.0);
	D3DXMATRIX cameraToWorld;
	D3DXMatrixInverse(&cameraToWorld, NULL, &worldToCamera);
	D3DXVECTOR3 eyePos;
	D3DXVec3TransformCoord(&eyePos, &origin, &cameraToWorld);

	D3DXVECTOR3 modelPos = D3DXVECTOR3(localToWorld.m[3][0], localToWorld.m[3][1], localToWorld.m[3][2]);

	D3DXVECTOR3 del = eyePos - modelPos;
	float dist = D3DXVec3Length(&del);
	return dist;
}
void GridObject::SetLOD()
{
	if(lodByDistance)
	{
		const float distance = GetDistanceToCamera();

        if(lowDistance > highDistance)
        {
            const float k1 = (lowNumTris - highNumTris) / (1.0f / (lowDistance * lowDistance) - 1.0f / (highDistance * highDistance) );
            const float k0 = lowNumTris - k1 / (lowDistance * lowDistance);

            float numTris = k0 + k1 / (distance * distance);
		    if(numTris < lowNumTris)
			    numTris = lowNumTris;
		    else if(numTris > highNumTris)
			    numTris = highNumTris;

		    SetMinTris(int(numTris));
        }
        else
        {
            SetMinTris(int(lowNumTris));
        }
	}
	else
	{
		SetMinTris(globalNumTris);
	}
}

//--------------------------------------------------------------------------------------
// This helper function creates a 2D texture full of random bytes
//--------------------------------------------------------------------------------------
HRESULT GridObject::CreateRandomPermTexture( ID3D11Device* pd3dDevice )
{
    HRESULT hr = S_OK;

	if(!pPermTextureSRV)
	{
		// Create our 256 x 256 random number texture up front
		BYTE data[256 * 256];
		for( int i = 0; i < 256 * 256; i++ )
			data[i] = ( BYTE )( rand() % 256 );

		// When we create the texture, pass in the data
		D3D11_TEXTURE2D_DESC desc;
		desc.Width = 256;
		desc.Height = 256;
		desc.MipLevels = 1;
		desc.ArraySize = 1;
		desc.Format = DXGI_FORMAT_R8_TYPELESS; // We can use typeless resource here as long as we make sure that give it an 
		desc.SampleDesc.Count = 1;             // explicit typewhen we create a view on it.
		desc.SampleDesc.Quality = 0;
		desc.Usage = D3D11_USAGE_IMMUTABLE;
		desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
		desc.CPUAccessFlags = 0;
		desc.MiscFlags = 0;
		D3D11_SUBRESOURCE_DATA dataDesc;
		dataDesc.pSysMem = data;
		dataDesc.SysMemPitch = 256;
		V_RETURN( pd3dDevice->CreateTexture2D( &desc, &dataDesc, &pPermTexture ) );

		// Create the shader resource view so we can bind the texture
		D3D11_SHADER_RESOURCE_VIEW_DESC descSRV;
		ZeroMemory( &descSRV, sizeof( descSRV ) );
		descSRV.Format = DXGI_FORMAT_R8_UNORM;
		descSRV.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
		descSRV.Texture2D.MipLevels = 1;
		descSRV.Texture2D.MostDetailedMip = 0;
		V_RETURN( pd3dDevice->CreateShaderResourceView( pPermTexture, &descSRV, &pPermTextureSRV ) );
	}

    return hr;
}

//--------------------------------------------------------------------------------------
// This helper function creates a 1D texture full of random vectors
//--------------------------------------------------------------------------------------
HRESULT GridObject::CreateRandomVectorTexture( ID3D11Device* pd3dDevice )
{
    HRESULT hr = S_OK;

	if(!pRandomVecTextureSRV)
	{
		// Create the random data up front
		float data[4096];
		for( int i = 0; i < 4096; i++ )
			data[i] = float( rand() - ( RAND_MAX / 2 ) ) / ( float )( RAND_MAX / 2 );

		// Create the texture with the data pass in
		D3D11_TEXTURE1D_DESC dstex;
		dstex.Width = 1024;
		dstex.MipLevels = 1;
		dstex.Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		dstex.Usage = D3D11_USAGE_DEFAULT;
		dstex.BindFlags = D3D11_BIND_SHADER_RESOURCE;
		dstex.CPUAccessFlags = 0;
		dstex.MiscFlags = 0;
		dstex.ArraySize = 1;
		D3D11_SUBRESOURCE_DATA InitData;
		InitData.pSysMem = data;
		InitData.SysMemPitch = 1024 * sizeof( float );
		InitData.SysMemSlicePitch = 1024 * sizeof( float );
		V_RETURN( pd3dDevice->CreateTexture1D( &dstex, &InitData, &pRandomVecTexture ) );

		// Create the resource view so that we can bind this
		D3D11_SHADER_RESOURCE_VIEW_DESC SRVDesc;
		ZeroMemory( &SRVDesc, sizeof( SRVDesc ) );
		SRVDesc.Format = dstex.Format;
		SRVDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE1D;
		SRVDesc.Texture2D.MipLevels = dstex.MipLevels;
		V_RETURN( pd3dDevice->CreateShaderResourceView( pRandomVecTexture, &SRVDesc, &pRandomVecTextureSRV ) );
	}

    return hr;
}
