
#ifndef GRID_OBJECT_H
#define GRID_OBJECT_H

#define LOAD_TEXTURES
#define OBB_BOUNDS

#include "Object.h"
#include "../Utility/Bounds.h"

class GridObject : public Object
{
private:
	ID3D11VertexShader*		vertexShader;
	ID3D11PixelShader*		pixelShader;

protected:
	static ID3D11Buffer*			everyFrameCB;
    ID3D11Buffer*                   everyObjectCB;

	static ID3D11ShaderResourceView*           cubeMap;
	static ID3D11ShaderResourceView*           cubeBlack;
	static ID3D11SamplerState*                 cubeSampler;

	static ID3D11SamplerState*                 rndByteSampler;
	static ID3D11SamplerState*                 rndVecSampler;

	static ID3D11Texture2D*                    pPermTexture;          // random texture for noise lookups
	static ID3D11ShaderResourceView*           pPermTextureSRV;		
	static ID3D11Texture1D*                    pRandomVecTexture;     // random vector texture for voronoi noise
	static ID3D11ShaderResourceView*           pRandomVecTextureSRV;


	static ID3D11RasterizerState*              pRasterizerState;

    static bool bShowBounds;
	static bool bWireFrame;
	static bool bLights;
	static bool bSpecular;
	static bool bEnvMap;
	static float tessFactor;
	static int globalNumTris;
	static bool powerOfTwo;
	static bool lodByDistance;
	static float detailScale;
    static float lowNumTris;
    static float highNumTris;
    static float lowDistance;
    static float highDistance;
    static int msaa;
    static D3DXVECTOR2 windowSize;
    static float depthTolerance;
    static float frameDecay;
    static bool reprojEnabled;
    static bool useLast;

	int minNumTris;
	D3DXMATRIX localToWorld;
	float randomOffset;

    D3DXMATRIX lastLocalToWorld;

	float fStepTime;
	float fCurrentTime;

#ifndef OBB_BOUNDS
    SphereBounds localBounds;
    SphereBounds worldBounds;
#else // OBB_BOUNDS
    AABB localBounds;
    OBB worldBounds;
#endif // OBB_BOUNDS

	struct CBChangesEveryFrame
	{
		D3DXMATRIX mCameraToNDC;
		D3DXVECTOR4 Details;
        D3DXVECTOR4 WindowSize;

		/// minDiffuse, mulSpec, specPower, 1.0/range
		D3DXVECTOR4 light; 

		D3DXMATRIX mWorldToCamera;
	};

    struct CBChangesEveryObject
    {
        D3DXMATRIX mLocalToWorld;

        D3DXVECTOR4 ModDivScale;

        D3DXMATRIX mLastLocalToNDC;

        D3DXVECTOR4 WorldCamPos;
    };

public:

	GridObject();

	virtual HRESULT Init();
	virtual void OnResize();
	virtual void Cleanup();

	virtual int Render();
    static  HRESULT SetEveryFrame();

    static bool GetShowBounds() { return bShowBounds; }
    static void SetShowBounds(bool b) { bShowBounds = b; }

	static bool GetWireFrame() { return bWireFrame; }
	static void SetWireFrame(bool b);

    static bool GetEnableMSAA() { return msaa > 1; }
    static int GetMSAA() { return msaa; }
    static void SetMSAA(int m);

	static bool GetLights() { return bLights; }
	static void SetLights(bool b) { bLights = b; }

	static bool GetSpecular() { return bSpecular; }
	static void SetSpecular(bool b) { bSpecular = b; }

	static bool GetEnvMap() { return bEnvMap; }
	static void SetEnvMap(bool b) { bEnvMap = b; }

	static float GetTessFactor() { return tessFactor; }
	static void SetTessFactor(float v) { tessFactor = v; }

	static bool GetPowerOfTwo() { return powerOfTwo; }
	static void SetPowerOfTwo(bool v) { powerOfTwo = v; }

	static int GetGlobalNumTris() { return globalNumTris; }
	static void SetGlobalNumTris(int v) { globalNumTris = v; }

	static bool GetLODByDistance() { return lodByDistance; }
	static void SetLODByDistance(bool b) { lodByDistance = b; }

	static float GetDetailScale() { return detailScale; }
	static void SetDetailScale(float v) { detailScale = v; }

	static float GetHighNumTris() { return highNumTris; }
	static void SetHighNumTris(float n) { highNumTris = n; }

    static float GetHighDistance() { return highDistance; }
    static void SetHighDistance(float f) { highDistance = f; }

    static float GetLowNumTris() { return lowNumTris; }
    static void SetLowNumTris(float n) { lowNumTris = n; }

    static float GetLowDistance() { return lowDistance; }
    static void SetLowDistance(float f) { lowDistance = f; }

    static D3DXVECTOR2 GetWindowSize() { return windowSize; }
    static void SetWindowSize(const D3DXVECTOR2& vec) { windowSize = vec; }

    static float GetDepthTolerance() { return depthTolerance; }
    static void SetDepthTolerance(float f) { depthTolerance = f; }

    static float GetFrameDecay() { return frameDecay; }
    static void SetFrameDecay(float f) { frameDecay = f; }

    static bool GetReprojEnabled() { return reprojEnabled; }
    static void SetReprojEnabled(bool b) { reprojEnabled = b; }

    static bool GetUseLast() { return useLast; }
    static void SetUseLast(bool b) { useLast = b; }

    static ID3D11ShaderResourceView* GetEnvironmentMap() { return cubeMap; }

	virtual const D3DXMATRIX& GetLocalToWorld() const { return localToWorld; }
	virtual void SetLocalToWorld(const D3DXMATRIX& l2w);

    virtual void SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi);
    virtual const Bounds& GetLocalBounds() const { return localBounds; }
    virtual const Bounds& GetWorldBounds() const { return worldBounds; }

	virtual void IncrementTime(float fAdvanceTime) { fStepTime = fAdvanceTime; fCurrentTime += fAdvanceTime; }

	int GetMinTris() const { return minNumTris; }
	void SetMinTris(int v) { minNumTris = v; }

	float RandomOffset(float lo = 0.0f, float hi = 1.0f) const
	{
		return lo + (hi - lo) * randomOffset;
	}

protected:
	virtual const WCHAR* ShaderFile() const = 0;
	virtual int ShaderModel() const = 0;
	virtual const CHAR* ShaderExtention() const = 0;
	virtual const WCHAR* WideShaderExtention() const = 0;

	virtual D3DXVECTOR4 ModDivScale() const = 0;

	virtual int NumU() const = 0;
	virtual int NumV() const = 0;

	int NumQuads() const { 	return (NumU() - 1) * (NumV() - 1); }
	int NumTris() const { return NumQuads() * 2; }
	int NumIndices() const { return NumTris() * 3; }

	float GetDistanceToCamera();
	void SetLOD();

	virtual HRESULT GetConstantBuffers();
	virtual HRESULT MakeGeometry();

    virtual void UpdateConstants();
    virtual HRESULT SetEveryObject();
	virtual void SetBuffers();
    virtual void SetResources();
	virtual void SetState();

    static void RefreshState();

    void UpdateBounds();

	virtual HRESULT GetVertexShader(const WCHAR* fileName, ID3D11VertexShader*& shader);
	virtual HRESULT GetPixelShader(const WCHAR* fileName, ID3D11PixelShader*& pixelShader);

public:
    /// TODO: Move this into a shared Noise setup utility class.
	static void SetNoiseResources();

	static HRESULT CreateRandomPermTexture( ID3D11Device* pd3dDevice );
	static HRESULT CreateRandomVectorTexture( ID3D11Device* pd3dDevice );

private:
	HRESULT GetStates();
	HRESULT GetTextures();

};
#endif // GRID_OBJECT_H
