
#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <D3DX11.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include <map>

#include <DXUT.h>

#include "PatchSurface.h"

#include "../Utility/CompiledShader.h"

typedef std::map<const WCHAR*, ID3D11HullShader*> HullShaderMap;
typedef HullShaderMap::iterator HullShaderIter;
typedef std::pair<const WCHAR*, ID3D11HullShader*> HullShaderEntry;

typedef std::map<const WCHAR*, ID3D11DomainShader*> DomainShaderMap;
typedef DomainShaderMap::iterator DomainShaderIter;
typedef std::pair<const WCHAR*, ID3D11DomainShader*> DomainShaderEntry;

static HullShaderMap hullShaderLUT;
static DomainShaderMap domainShaderLUT;


PatchSurface::PatchSurface(const WCHAR* shaderFile = L"Shaders/toy_fir_tree3.3dmodel.fx")
	: GridObject(),
		shaderFile(shaderFile)
{
}


HRESULT PatchSurface::GetHullShader(const WCHAR* fileName, ID3D11HullShader*& pShader)
{
	HRESULT hr = S_OK;

	HullShaderIter iter = hullShaderLUT.find(fileName);
	if(iter != hullShaderLUT.end())
	{
		pShader = iter->second;
		pShader->AddRef();
		return S_OK;
	}

	WCHAR entryPoint[256];
	swprintf_s(entryPoint, L"HS_%s", WideShaderExtention());
	WCHAR shaderModel[256];
	swprintf_s(shaderModel, L"hs_%d_0", ShaderModel());

	CompiledShader shader(fileName, shaderModel, entryPoint);

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreateHullShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &pShader );

	hullShaderLUT.insert(HullShaderEntry(fileName, pShader));

	return hr;
}

HRESULT PatchSurface::GetDomainShader(const WCHAR* fileName, ID3D11DomainShader*& pShader)
{
	HRESULT hr = S_OK;

	DomainShaderIter iter = domainShaderLUT.find(fileName);
	if(iter != domainShaderLUT.end())
	{
		pShader = iter->second;
		pShader->AddRef();
		return S_OK;
	}

	WCHAR entryPoint[256];
	swprintf_s(entryPoint, L"DS_%s", WideShaderExtention());
	WCHAR shaderModel[256];
	swprintf_s(shaderModel, L"ds_%d_0", ShaderModel());

	CompiledShader shader(fileName, shaderModel, entryPoint);

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreateDomainShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &pShader );

	domainShaderLUT.insert(DomainShaderEntry(fileName, pShader));

	return hr;
}

HRESULT PatchSurface::Init()
{
	HRESULT hr = GridObject::Init();

	if(FAILED(hr))
		return hr;

	/// init hull and domain shaders
	hr = GetHullShader(ShaderFile(), hullShader);
	if(FAILED(hr))
		return hr;

	hr = GetDomainShader(ShaderFile(), domainShader);
	if(FAILED(hr))
		return hr;

	return S_OK;
}

void PatchSurface::Cleanup()
{
	hullShaderLUT.clear();
	domainShaderLUT.clear();

	GridObject::Cleanup();

	/// Release domain and hull shaders
	SAFE_RELEASE(hullShader);
	SAFE_RELEASE(domainShader);

	SAFE_RELEASE(materialCB);

}

HRESULT PatchSurface::GetConstantBuffers()
{
	HRESULT hr = S_OK;

	hr = GridObject::GetConstantBuffers();
    if( FAILED( hr ) )
        return hr;

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;

    ZeroMemory( &bd, sizeof(bd) );
    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(MaterialCB);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;

    hr = pd3dDevice->CreateBuffer( &bd, NULL, &materialCB);
    if( FAILED( hr ) )
        return hr;

	return hr;
}

void PatchSurface::UpdateConstants()
{
	GridObject::UpdateConstants();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	// Moving the materials setup to be consts embedded in the .fx file,
	// so that the .fx file completely determines the appearance of the model.
	MaterialCB mCb;
	D3DXVECTOR4 vec(0.8, 0.7, -1.0, 0.0);
	D3DXVec4Normalize(&mCb.LightVector, &vec);

	mCb.TessFactor = D3DXVECTOR4(tessFactor, tessFactor, tessFactor, tessFactor);

    pImmediateContext->UpdateSubresource( materialCB, 0, NULL, &mCb, 0, 0 );

    pImmediateContext->VSSetConstantBuffers( 3, 1, &materialCB);
    pImmediateContext->HSSetConstantBuffers( 3, 1, &materialCB);
    pImmediateContext->DSSetConstantBuffers( 3, 1, &materialCB);
    pImmediateContext->PSSetConstantBuffers( 3, 1, &materialCB);
}

HRESULT PatchSurface::MakeGeometry()
{
	/// No-op function to prevent the unnecessary creation of an index buffer by base class.
	HRESULT hr = S_OK;

	return hr;
}


static float overTess = -1.0f;
int PatchSurface::NumU() const 
{ 

	float minTris = float(GetMinTris());
	float minQuads = minTris * 0.5f;
	float vertsPerSide = sqrt(minQuads) + 1.0f;

	const float kMaxTessFactor = 64.0f;
	float patchPerSide = ceil(vertsPerSide / (kMaxTessFactor + 1));

	tessFactor = vertsPerSide / patchPerSide - 1.0f;
	tessFactor = max(1, tessFactor);
	assert(tessFactor > 0);

	if(powerOfTwo)
	{
		float logTess = float(log(tessFactor) / M_LN2);
		logTess += 0.5f;
		int intLogTess = int(logTess);
		float fracLogTess = logTess - intLogTess;
		logTess = float(intLogTess);
		tessFactor = pow(2.0f, logTess) - 1.0f + fracLogTess;
	}

	if(overTess > 0)
		tessFactor = overTess;

	return int(patchPerSide);
}

int PatchSurface::NumV() const 
{ 
	return NumU();
}

void PatchSurface::SetResources()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	GridObject::SetResources();

	/// set hull and domain shaders
    pImmediateContext->HSSetShader( hullShader, NULL, 0 );

    pImmediateContext->DSSetShader( domainShader, NULL, 0 );
    pImmediateContext->GSSetShader( NULL, NULL, 0 );

    pImmediateContext->HSSetConstantBuffers( 0, 1, &everyFrameCB );
    
    pImmediateContext->DSSetConstantBuffers( 0, 1, &everyFrameCB );
}

void PatchSurface::SetBuffers()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	pImmediateContext->IASetInputLayout( NULL );
	pImmediateContext->IASetVertexBuffers( 0, 0, NULL, NULL, NULL );

    // Set primitive topology
    pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_4_CONTROL_POINT_PATCHLIST );
}

int PatchSurface::Render()
{
	UpdateConstants();

	SetResources();

	SetBuffers();

	SetState();

	const int CTL_PTS_PER_SIDE = 2;
	const int CTL_PTS_PER_QUAD = CTL_PTS_PER_SIDE * CTL_PTS_PER_SIDE;

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	int numU = NumU();
	int numV = NumV();
	pImmediateContext->Draw( numU * numV * CTL_PTS_PER_QUAD, 0);

	return GetMinTris();
}

D3DXVECTOR4 PatchSurface::ModDivScale() const
{
	int numU = NumU();
	int numV = NumV();
	D3DXVECTOR4 modDivScale(float(numU), float(numU), 1.0f / float(numU), 1.0f / float(numV));
	return modDivScale;
}
