
#define _USE_MATH_DEFINES

#include <windows.h>
#include <d3dx11.h>
#include <d3dcompiler.h>

#include "DXUT.h"
#include "../Utility/CFlyCamera.h"
#include "../Utility/CompiledShader.h"

#include "SimpleScape.h"

#include "GridObject.h" // for MSAA setting

extern CFlyCamera          g_Camera;               // A model viewing camera

struct ScapeConstantBuffer
{
    D3DXMATRIX  localToNDC;
    D3DXMATRIX  localToWorld;
    D3DXVECTOR4 color;
    D3DXMATRIX  lastLocalToNDC;
    D3DXVECTOR4 texScale;
    D3DXVECTOR4 light;
    D3DXVECTOR4 windowSize;
    D3DXVECTOR4 chessBoard;
};

struct ScapeVertex
{
    D3DXVECTOR3 Position;
    D3DXVECTOR3 Normal;
    D3DXVECTOR2 TexCoord;
};

SimpleScape::SimpleScape() :
    m_pRasterizerState(NULL),

    m_pVertexShader(NULL),
    m_pPixelShader(NULL),
    m_pVertexLayout(NULL),
    m_pVertexBuffer(NULL),
    m_pIndexBuffer(NULL),
    m_pConstantBuffer(NULL),

    m_pDetailTop(NULL),
    m_pDetailBot(NULL),
    m_pMask(NULL),
    m_pSampler(NULL),
    m_pSamplePrevious(NULL),

    m_pVerts(NULL),

    numU(100),
    numV(100),
    lo(-100.0f, -100.0f, 0.0f),
    hi(100.0f, 100.0f, 0.0f),

    m_bWireframe(false)
{
    ZeroMemory(&localToWorld, sizeof(localToWorld));
    localToWorld.m[0][0] = 1.0f;
    localToWorld.m[1][2] = -1.0f;
    localToWorld.m[2][1] = 1.0f;
    localToWorld.m[3][3] = 1.0f;
    lastLocalToWorld = localToWorld;
    D3DXMatrixInverse(&worldToLocal, NULL, &localToWorld);
}

HRESULT SimpleScape::Init()
{
    HRESULT hr = S_OK;
    V_RETURN(MakeState());
    V_RETURN(MakeTextures());
    V_RETURN(MakeShaders());
    V_RETURN(MakeGeometry());
    V_RETURN(MakeConstants());
    return hr;
}

void SimpleScape::Cleanup()
{
	SAFE_RELEASE(m_pRasterizerState);

    SAFE_RELEASE(m_pVertexShader);
    SAFE_RELEASE(m_pPixelShader);
    SAFE_RELEASE(m_pVertexLayout);
    SAFE_RELEASE(m_pVertexBuffer);
    SAFE_RELEASE(m_pIndexBuffer);
    SAFE_RELEASE(m_pConstantBuffer);

	SAFE_RELEASE(m_pDetailTop);
	SAFE_RELEASE(m_pDetailBot);
	SAFE_RELEASE(m_pMask);
	SAFE_RELEASE(m_pSampler);
    SAFE_RELEASE(m_pSamplePrevious);

    SAFE_DELETE(m_pVerts);
}

int SimpleScape::Render()
{
    if(GetEnableDraw())
    {
 	    ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

        SetState(pImmediateContext);
        SetTextures(pImmediateContext);

        SetConstants(pImmediateContext, localToWorld);

	    pImmediateContext->VSSetShader( m_pVertexShader, NULL, 0 );
	    pImmediateContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
	    pImmediateContext->PSSetShader( m_pPixelShader, NULL, 0 );
	    pImmediateContext->PSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
	    pImmediateContext->DrawIndexed( NumIndices(), 0, 0 );

        return NumTris();
    }
    return 0;
}

bool SimpleScape::QueryHeight(const D3DXVECTOR2& worldXZLoc, float& height, D3DXVECTOR3& normal)
{
    D3DXVECTOR3 pos(worldXZLoc.x, 0.0f, worldXZLoc.y);
    D3DXVECTOR4 xyLoc;
    D3DXVec3Transform(&xyLoc, &pos, &worldToLocal);
    float x = (xyLoc.x - lo.x) / (hi.x - lo.x) * NumU();
    float y = (xyLoc.y - lo.y) / (hi.y - lo.y) * NumV();
    int iX = int(x);
    int iY = int(y);
    int iXPlus = iX + 1;
    int iYPlus = iY + 1;
    float fracX = x - iX;
    float fracY = y - iY;
    if(iX < 0)
    {
        iX = 0;
        fracX = 0;
        iXPlus = 0;
    }
    else if( iXPlus >= NumU())
    {
        iX = NumU() - 1;
        fracX = 0;
        iXPlus = iX;
    }
    if(iY < 0)
    {
        iY = 0;
        fracY = 0;
        iYPlus = 0;
    }
    else if(iYPlus >= NumV())
    {
        iY = NumV() - 1;
        fracY = 0;
        iYPlus = iY;
    }

    height = m_pVerts[iX + iY * NumU()].Position.z * (1.0f - fracX) * (1.0f - fracY)
        + m_pVerts[iXPlus + iY * NumU()].Position.z * fracX * (1.0f - fracY)
        + m_pVerts[iX + iYPlus * NumU()].Position.z * (1.0f - fracX) * fracY
        + m_pVerts[iXPlus + iYPlus * NumU()].Position.z * fracX * fracY;

    normal = m_pVerts[iX + iY * NumU()].Normal * (1.0f - fracX) * (1.0f - fracY)
        + m_pVerts[iXPlus + iY * NumU()].Normal * fracX * (1.0f - fracY)
        + m_pVerts[iX + iYPlus * NumU()].Normal * (1.0f - fracX) * fracY
        + m_pVerts[iXPlus + iYPlus * NumU()].Normal * fracX * fracY;

    D3DXVec3Normalize(&normal, &normal);

    return xyLoc.x >= lo.x && xyLoc.y >= lo.y && xyLoc.x <= hi.x && xyLoc.y <= hi.y;
}

static D3DXVECTOR3 DiffNorm(ScapeVertex* verts, int idx0, int idx1)
{
    D3DXVECTOR3 diff = verts[idx1].Position - verts[idx0].Position;
    D3DXVECTOR3 binormal(diff.y, -diff.x, 0.0f);
    D3DXVECTOR3 norm;
    D3DXVec3Cross(&norm, &binormal, &diff);
    D3DXVec3Normalize(&norm, &norm);
    return norm;
}
void SimpleScape::CalcNormals(ScapeVertex* verts)
{
    for(int j = 0; j < NumV(); ++j)
    {
        for(int i = 0; i < NumU(); ++i)
        {
            D3DXVECTOR3 norm(0.0f, 0.0f, 0.0f);
            int idx = i + j * NumU();
            if(i > 0)
                norm += DiffNorm(verts, idx, idx - 1);
            if(i < NumU() - 1)
                norm += DiffNorm(verts, idx, idx + 1);
            if(j > 0)
                norm += DiffNorm(verts, idx, idx - NumU());
            if(j < NumV() - 1)
                norm += DiffNorm(verts, idx, idx + NumU());

            D3DXVec3Normalize(&verts[idx].Normal, &norm);
        }
    }
}

void SimpleScape::SetLocalToWorld(const D3DXMATRIX& l2w)
{
    // Ignore any requests to set the transform, the terrain is fixed in space
    // in order to allow height queries to remain valid.

    //lastLocalToWorld = localToWorld;

    //localToWorld = l2w;
    //RefreshTransform();
}

void SimpleScape::RefreshTransform()
{
    localBounds = OBB(lo, hi);
    worldBounds = localBounds;
    worldBounds.Transform(localToWorld);
}

void SimpleScape::SetTextures(ID3D11DeviceContext* pImmediateContext)
{
	pImmediateContext->PSSetShaderResources( 0, 1, &m_pDetailTop );
	pImmediateContext->PSSetShaderResources( 1, 1, &m_pDetailBot );
	pImmediateContext->PSSetShaderResources( 2, 1, &m_pMask );
    pImmediateContext->PSSetSamplers( 0, 1, &m_pSampler);
    pImmediateContext->PSSetSamplers( 4, 1, &m_pSamplePrevious);
}

void SimpleScape::SetState(ID3D11DeviceContext* pImmediateContext)
{
    CheckState();

    pImmediateContext->IASetInputLayout( m_pVertexLayout );

    // Set vertex buffer
    const UINT stride = sizeof( ScapeVertex );
    const UINT offset = 0;
    pImmediateContext->IASetVertexBuffers( 0, 1, &m_pVertexBuffer, &stride, &offset );

    // Set index buffer
    pImmediateContext->IASetIndexBuffer( m_pIndexBuffer, DXGI_FORMAT_R16_UINT, 0 );

    // Set primitive topology 
    pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );

	pImmediateContext->RSSetState(m_pRasterizerState);
}

void SimpleScape::SetConstants(ID3D11DeviceContext* pImmediateContext, const D3DXMATRIX& l2w)
{
    const D3DXMATRIX worldToCamera = *g_Camera.GetViewMatrix();
    const D3DXMATRIX cameraToNDC = *g_Camera.GetProjMatrix();
    const D3DXMATRIX localToNDC = l2w * worldToCamera * cameraToNDC;

    ScapeConstantBuffer cb;
	D3DXMatrixTranspose( &cb.localToNDC, &localToNDC);
    D3DXMatrixTranspose( &cb.localToWorld, &localToWorld );

    cb.color = D3DXVECTOR4(1.0f, 0.5f, 0.5f, 1.0f);

    D3DXMATRIX lastLocalToNDC = lastLocalToWorld * g_Camera.GetLastWorldToNDC();
    D3DXMatrixTranspose(&cb.lastLocalToNDC, &lastLocalToNDC);

    static D3DXVECTOR4 s_TexScale(40.0f, 40.0f, 57.0f, 57.0f);
    cb.texScale = s_TexScale;

    D3DXVECTOR4 light(1.0f, 1.0f, 0.0f, 0.0f);
    if(!GridObject::GetLights())
        light.x = 0;
    cb.light = light;

    const float frameDecay = GridObject::GetFrameDecay();
    const float reprojDecay = GridObject::GetReprojEnabled() ? frameDecay : 0.0f;
    const D3DXVECTOR2 windowSize = GridObject::GetWindowSize();
    const float depthTolerance = GridObject::GetDepthTolerance();
    cb.windowSize = D3DXVECTOR4(windowSize.x, windowSize.y, depthTolerance, reprojDecay);

    if(m_SquareSize > 0)
    {
        cb.chessBoard = D3DXVECTOR4(
            1.0f,
            0.5f / m_SquareSize,
            0.7f,
            0.0f);
    }
    else
    {
        cb.chessBoard = D3DXVECTOR4(0.0f, 0.0f, 0.0f, 0.0f);
    }

    pImmediateContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &cb, 0, 0 );
}

void SimpleScape::SetVerts(const ScapeVertex* const pts, int count)
{
 	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
    D3D11_MAPPED_SUBRESOURCE MappedSubresource;
    pImmediateContext->Map( m_pVertexBuffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedSubresource );

    memcpy(MappedSubresource.pData, pts, count * sizeof(ScapeVertex));

    pImmediateContext->Unmap( m_pVertexBuffer, 0 );
}

static float SineHeight(const float u, const float v)
{
    const float freqU = 2.1f;
    const float freqV = 1.7f;
    float sineU = float(sin(u * M_PI * 2.0 * freqU));
    float sineV = float(sin(v * M_PI * 2.0f * freqV));

    return sineU * sineV * 2.0f;
}

ScapeVertex* SimpleScape::GenVerts()
{
    ScapeVertex* verts = new ScapeVertex[NumVerts()];

    lo.z = FLT_MAX;
    hi.z = -FLT_MAX;
    for(int j = 0; j < numV; ++j)
    {
        const float v = float(j) / float(numV - 1);
        for(int i = 0; i < numU; ++i)
        {
            ScapeVertex& vert = verts[i + j * numU];
            const float u = float(i) / float(numU - 1);

#if 0
            const float height = 0.0f;
#elif 1
            const float height = SineHeight(u, v);
#else
            const float height = (u + v) * 20.0f;
#endif
            lo.z = min(lo.z, height);
            hi.z = max(hi.z, height);

            vert.Position = D3DXVECTOR3(
                lo.x + u * (hi.x - lo.x),
                lo.y + v * (hi.y - lo.y),
                height);
            vert.Normal = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
            vert.TexCoord = D3DXVECTOR2(u, v);
        }
    }
    CalcNormals(verts);
    RefreshTransform();

    return verts;
}

WORD* SimpleScape::GenIndices()
{
    WORD* indices = new WORD[NumIndices()];

    int iIndex = 0;
    const int endJ = NumV() - 1;
    const int endI = NumU() - 1;
    const int stride = NumU();
    for(int j = 0; j < endJ; ++j)
    {
        for(int i = 0; i < endI; ++i)
        {
            indices[iIndex++] = WORD(i + j * stride);
            indices[iIndex++] = WORD(i + 1 + j * stride);
            indices[iIndex++] = WORD(i + stride + j * stride);

            indices[iIndex++] = WORD(i + stride + j * stride);
            indices[iIndex++] = WORD(i + 1 + j * stride);
            indices[iIndex++] = WORD(i + 1 + stride + j * stride);
        }
    }

    assert(iIndex == NumIndices());

    return indices;
}

void SimpleScape::SetResolution(int numU, int numV)
{
    this->numU = numU;
    this->numV = numV;

    MakeGeometry();
}

void SimpleScape::CheckState()
{
    if(m_bWireframe != GridObject::GetWireFrame())
    {
        SAFE_RELEASE(m_pRasterizerState);
        m_bWireframe = GridObject::GetWireFrame();
        MakeState();
    }
}
HRESULT SimpleScape::MakeState()
{
    D3D11_RASTERIZER_DESC drd = {
        D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
        D3D11_CULL_NONE,//D3D11_CULL_MODE CullMode;
        FALSE, //BOOL FrontCounterClockwise;
        0, //INT DepthBias;
        0.0f,//FLOAT DepthBiasClamp;
        0.0f,//FLOAT SlopeScaledDepthBias;
        TRUE,//BOOL DepthClipEnable;
        FALSE,//BOOL ScissorEnable;
        GridObject::GetEnableMSAA(),//BOOL MultisampleEnable;
        FALSE//BOOL AntialiasedLineEnable;        
    };
    if(m_bWireframe)
    {
	    drd.FillMode = D3D11_FILL_WIREFRAME;
    }

    ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    HRESULT hr = pd3dDevice->CreateRasterizerState(&drd, &m_pRasterizerState);
    assert(SUCCEEDED(hr));

    return hr;
}

HRESULT SimpleScape::MakeTextures()
{
	HRESULT hr = S_OK;
	/// Disable load and set of textures, we aren't using them currently.

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

	if(!m_pSampler)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
#if 0
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
#else
        sampDesc.Filter = D3D11_FILTER_ANISOTROPIC;
        sampDesc.MaxAnisotropy = 16;
#endif
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &m_pSampler );
		if( FAILED(hr) )
			return hr;
	}
	if(!m_pSamplePrevious)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
		sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &m_pSamplePrevious );
		if( FAILED(hr) )
			return hr;
	}


    // Load the Texture
	if(!m_pDetailTop)
	{
		hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/terrain_01.jpg", NULL, NULL, &m_pDetailTop, NULL );
		if( FAILED( hr ) )
			return hr;
	}
	if(!m_pDetailBot)
	{
		hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/terrain_00.png", NULL, NULL, &m_pDetailBot, NULL );
		if( FAILED( hr ) )
			return hr;
	}
	if(!m_pMask)
	{
		hr = D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/GrassMask.png", NULL, NULL, &m_pMask, NULL );
		if( FAILED( hr ) )
			return hr;
	}

    return hr;
}

HRESULT SimpleScape::MakeShaders()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    const WCHAR* fileName = L"Shaders/DrawTerrain.fx";

    // Create the vertex shader
    CompiledShader vShader(fileName, L"vs_5_0", L"VS_Def");
    HRESULT hr = pd3dDevice->CreateVertexShader( vShader.GetBufferPointer(), vShader.GetBufferSize(), NULL, &m_pVertexShader );
    assert(SUCCEEDED(hr));

    CompiledShader pShader(fileName, L"ps_5_0", L"PS_Def");
    hr = pd3dDevice->CreatePixelShader( pShader.GetBufferPointer(), pShader.GetBufferSize(), NULL, &m_pPixelShader );
    assert(SUCCEEDED(hr));

    // Define the input layout
    D3D11_INPUT_ELEMENT_DESC layout[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D11_INPUT_PER_VERTEX_DATA, 0 },
	};
	UINT numElements = ARRAYSIZE( layout );

    // Create the input layout
	hr = pd3dDevice->CreateInputLayout( layout, numElements, vShader.GetBufferPointer(),
                                          vShader.GetBufferSize(), &m_pVertexLayout );
    assert(SUCCEEDED(hr));

    return hr;
}
HRESULT SimpleScape::MakeGeometry()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    HRESULT hr = S_OK;

    {
        SAFE_DELETE(m_pVerts);
        m_pVerts = GenVerts();

        D3D11_BUFFER_DESC bd;
	    ZeroMemory( &bd, sizeof(bd) );
        bd.Usage = D3D11_USAGE_DYNAMIC;
        bd.ByteWidth = sizeof( ScapeVertex ) * NumVerts();
        bd.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	    bd.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
        D3D11_SUBRESOURCE_DATA InitData;
	    ZeroMemory( &InitData, sizeof(InitData) );
        InitData.pSysMem = m_pVerts;
        hr = pd3dDevice->CreateBuffer( &bd, &InitData, &m_pVertexBuffer );
        assert(SUCCEEDED(hr));
    }

    {
        D3D11_BUFFER_DESC bd;
	    ZeroMemory( &bd, sizeof(bd) );
        // Create index buffer
        WORD* indices = GenIndices();
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.ByteWidth = sizeof( WORD ) * NumIndices();        
        bd.BindFlags = D3D11_BIND_INDEX_BUFFER;
	    bd.CPUAccessFlags = 0;
        D3D11_SUBRESOURCE_DATA InitData;
	    ZeroMemory( &InitData, sizeof(InitData) );
        InitData.pSysMem = indices;
        hr = pd3dDevice->CreateBuffer( &bd, &InitData, &m_pIndexBuffer );
        assert(SUCCEEDED(hr));
        SAFE_DELETE_ARRAY(indices);
    }

    return hr;
}
HRESULT SimpleScape::MakeConstants()
{
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;


    bd.ByteWidth = sizeof(ScapeConstantBuffer);
    HRESULT hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer);
    assert(SUCCEEDED(hr));    
    return hr;
}

void SimpleScape::SetChessBoard( float squareSize )
{
    m_SquareSize = squareSize;
}


