

#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <D3DX11.h>

#include <map>

#include "DXUT.h"
#include "../Utility/CFlyCamera.h"
#include "../Utility/CompiledShader.h"

#define _USE_MATH_DEFINES
#include <math.h>

#include "Sphere.h"

#include "GridObject.h" // for settings like wireframe

extern CFlyCamera g_Camera;               // A model viewing camera

Sphere::Sphere()
	: m_pVertexShader(NULL),
		m_pPixelShader(NULL),
		m_pEveryFrameCB(NULL),
		m_pRasterizerStateSolid(NULL),
		m_pRasterizerStateWireframe(NULL),
		m_fTime(0)
{
	D3DXMatrixIdentity(&m_LocalToWorld);
}

Sphere::~Sphere()
{
}

HRESULT Sphere::MakeVertexShader()
{
	HRESULT hr = S_OK;

	CompiledShader shader(L"Shaders\\Sphere.fx", L"vs_4_0", L"VS_Main");
    
	// Create the vertex shader
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreateVertexShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &m_pVertexShader );

	return hr;
}

HRESULT Sphere::MakePixelShader()
{
	HRESULT hr = S_OK;

	CompiledShader shader(L"Shaders\\Sphere.fx", L"ps_4_0", L"PS_Main");

    // Create the pixel shader
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreatePixelShader( shader.GetBufferPointer(), shader.GetBufferSize(), NULL, &m_pPixelShader );

	return hr;
}

HRESULT Sphere::MakeConstantBuffers()
{
	HRESULT hr = S_OK;

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;

    bd.ByteWidth = sizeof(CBChangesEveryFrame);
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pEveryFrameCB);
    if( FAILED( hr ) )
        return hr;


	return hr;
}

HRESULT Sphere::MakeRasterizerStates()
{
	HRESULT hr = S_OK;

    D3D11_RASTERIZER_DESC drd = {
        D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
        D3D11_CULL_NONE,//D3D11_CULL_BACK,//        D3D11_CULL_MODE CullMode;
        FALSE, //BOOL FrontCounterClockwise;
        0, //INT DepthBias;
        0.0f,//FLOAT DepthBiasClamp;
        0.0f,//FLOAT SlopeScaledDepthBias;
        TRUE,//BOOL DepthClipEnable;
        FALSE,//BOOL ScissorEnable;
        TRUE,//BOOL MultisampleEnable;
        FALSE//BOOL AntialiasedLineEnable;        
    };
	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();
    hr = pd3dDevice->CreateRasterizerState(&drd, &m_pRasterizerStateSolid);
	if( FAILED( hr ) )
		return hr;

	drd.FillMode = D3D11_FILL_WIREFRAME;
    hr = pd3dDevice->CreateRasterizerState(&drd, &m_pRasterizerStateWireframe);
	if( FAILED( hr ) )
		return hr;

	return hr;
}

HRESULT Sphere::Init()
{
	HRESULT hr = S_OK;

	hr = MakeVertexShader();
	if( FAILED(hr) )
		return hr;

	hr = MakePixelShader();
	if( FAILED(hr) )
		return hr;

	hr = MakeConstantBuffers();
	if( FAILED(hr) )
		return hr;

	hr = MakeRasterizerStates();
	if( FAILED(hr) )
		return hr;

	return hr;
}

void Sphere::OnResize()
{
	SetOnResize();
}

void Sphere::Cleanup()
{
	SAFE_RELEASE(m_pVertexShader);
	SAFE_RELEASE(m_pPixelShader);

	SAFE_RELEASE(m_pEveryFrameCB);

	SAFE_RELEASE(m_pRasterizerStateSolid);
	SAFE_RELEASE(m_pRasterizerStateWireframe);
}

int Sphere::Render()
{
	UpdateConstants();

	SetResources();

	SetBuffers();

	SetState();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	int numTris = NumTris();
	pImmediateContext->Draw( numTris * 3, 0 );

	return numTris;
}

void Sphere::SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi)
{
    localBounds.Position = (lo + hi) * 0.5f;
    D3DXVECTOR3 diff = hi - localBounds.Position;
    localBounds.Radius = D3DXVec3Length(&diff);
}

void Sphere::SetRadius(FLOAT f) 
{ 
    m_Radius = f; 
    localBounds = SphereBounds(D3DXVECTOR3(0.0f, 0.0f, 0.0f), m_Radius);
    worldBounds = localBounds;
    worldBounds.Transform(GetLocalToWorld());
}

int Sphere::NumU() const
{
	return 100;
}
int Sphere::NumV() const
{
	return 100;
}
int Sphere::NumTris() const
{
	return NumU() * NumV() * 2;
}

void Sphere::UpdateConstants()
{
    D3DXMATRIX worldToCamera = *g_Camera.GetViewMatrix();
    D3DXMATRIX cameraToNDC = *g_Camera.GetProjMatrix();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	{
		CBChangesEveryFrame cb;

		D3DXMatrixTranspose(&cb.mCameraToNDC, &cameraToNDC);

        D3DXMatrixTranspose(&cb.m_LocalToWorld, &m_LocalToWorld);

		D3DXVECTOR4 modDivScale(float(NumU()), float(NumU()), 1.0f / float(NumU()), 1.0f / float(NumV()));
		cb.m_ModDivScale = modDivScale;

		D3DXMatrixTranspose( &cb.m_WorldToCamera, &worldToCamera );

		D3DXVECTOR3 origin = D3DXVECTOR3(0.0, 0.0, 0.0);
		D3DXMATRIX cameraToWorld;
		D3DXMatrixInverse(&cameraToWorld, NULL, &worldToCamera);
		D3DXVECTOR4 eyePos;
		D3DXVec3Transform(&eyePos, &origin, &cameraToWorld);
		//eyePos.w = m_fTime;
        eyePos.w = m_Radius;
		cb.m_WorldCamPos = eyePos;
		pImmediateContext->UpdateSubresource( m_pEveryFrameCB, 0, NULL, &cb, 0, 0 );
	}

}

void Sphere::SetResources()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	pImmediateContext->VSSetShader( m_pVertexShader, NULL, 0 );
	pImmediateContext->PSSetShader( m_pPixelShader, NULL, 0 );

    pImmediateContext->VSSetConstantBuffers( 0, 1, &m_pEveryFrameCB );

    pImmediateContext->PSSetConstantBuffers( 0, 1, &m_pEveryFrameCB );
}

void Sphere::SetBuffers()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	// Set the input layout
	pImmediateContext->IASetInputLayout( NULL );
	pImmediateContext->IASetVertexBuffers( 0, 0, NULL, NULL, NULL );

    // Set primitive topology
    pImmediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
}

void Sphere::SetState()
{
	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();
	
	ID3D11RasterizerState* rs = GridObject::GetWireFrame() ? m_pRasterizerStateWireframe : m_pRasterizerStateSolid;
	pImmediateContext->RSSetState(rs);
}

const D3DXMATRIX& Sphere::GetLocalToWorld() const
{
	return m_LocalToWorld;
}

void Sphere::SetLocalToWorld(const D3DXMATRIX& l2w)
{
	m_LocalToWorld = l2w;
}

void Sphere::IncrementTime(float fAdvanceTime)
{
	m_fTime += fAdvanceTime;
}

void Sphere::SetOnResize()
{
}