
#ifndef NOISE_FX
#define NOISE_FX
//--------------------------------------------------------------------------------------
// File: Noise.fx
//
// The noise effect file for the ProceduralMaterials11 sample.  This file implements
// both Simplex noise in 3D and 4D and voronoi noise.
//
// For a good explaination of the simplex noise algorithm see
// http://staffwww.itn.liu.se/~stegu/simplexnoise/simplexnoise.pdf
// 
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
Texture2D<float>   g_txRandomByte	: register( t1 );   // permutation texture
Texture1D<float4>  g_txRandVector	: register( t2 );	 // random 1D texture of random vectors used for voronoi noise
SamplerState rndByteSampler : register( s1 );
SamplerState rndVecSampler : register( s2 );

//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
SamplerState g_samWrap
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};

#define MAF_UNROLL_VORONOI

/// Some helpers for cleaning up dealing with derivatives.
///
/// Extract partials of a w.r.t. b, where b is a float3.
#define EXTRACT_DERIV3(a, b) float4(a`(b.x), a`(b.y), a`(b.z), 0.0f)

/// Set the partials of a w.r.t. b to be zero, where b is a float3. Essentially marks a as constant as 
/// far as b is concerned.
#define ZERO_DERIV3(a, b) (a`(b.x) = a`(b.y) = a`(b.z) = 0.0f)

/// Set the derivative of a w.r.t. s = 1, as we generally want where a = frac(s);
/// If s=func(t), and t is the parameter we will eventually be differentiating w.r.t.,
/// then we want to set d(a)/d(t) = d(s)/d(t).
#define FRAC_DERIV3(a, s, t) (a`(t.x) = s`(t.x), a`(t.y) = s`(t.y), a`(t.z) = s`(t.z))

/// Explicitly set the derivative of a w.r.t. p to be d, where p and d are float3's.
#define SET_DERIV3(a, p, d) (a`((p).x) = (d).x, a`((p).y) = (d).y, a`((p).z) = (d).z)

//--------------------------------------------------------------------------------------
// Find which simplex we are in by magnitude sorting (3D case)
// This fast code for finding the simplex comes from the Perlin Fire sample from NVidia Corporation.  
// For more information about it, visit NVidia's developer website.
// http://developer.download.nvidia.com/whitepapers/2007/SDK10/PerlinFire.pdf
//--------------------------------------------------------------------------------------
/// Reworked to take individual float3's for output. Symbolic differentiation compiler
/// currently doesn't support arrays or array elements. 
void Simplex3D( const in float3 P, out float3 simplex0, out float3 simplex1, out float3 simplex2, out float3 simplex3, const float3 param )
{
    float3 T = P.xzy >= P.yxz;
    simplex0 = 0;
    simplex1 = T.xzy > T.yxz;
    simplex2 = T.yxz <= T.xzy;
    simplex3 = 1;
	ZERO_DERIV3(simplex0, param);
	ZERO_DERIV3(simplex1, param);
	ZERO_DERIV3(simplex2, param);
	ZERO_DERIV3(simplex3, param);
}

//--------------------------------------------------------------------------------------
// Find which simplex we are in by magnitude sorting (4D case)
// This fast code for finding the simplex comes from the Perlin Fire sample from NVidia Corporation.  
// For more information about it, visit NVidia's developer website.
//--------------------------------------------------------------------------------------
/// Reworked to take individual float3's for output. Symbolic differentiation compiler
/// currently doesn't support arrays or array elements. 
void Simplex4D( const in float4 P, 
	out float4 simplex0,
	out float4 simplex1,
	out float4 simplex2,
	out float4 simplex3,
	out float4 simplex4,
	const float3 param
	 )
{
    float4 offset0;

    float3 isX = step( P.yzw, P.xxx );
    offset0.x = dot( isX, float3(1, 1, 1) );
    offset0.yzw = 1 - isX;

    float2 isY = step( P.zw, P.yy );
    offset0.y += dot( isY, float2(1, 1) );
    offset0.zw += 1 - isY;

    float isZ = step( P.w, P.z );
    offset0.z += isZ;
    offset0.w += 1 - isZ;

    // offset0 now contains the unique values 0,1,2,3 in each channel
    simplex4 = 1;
    simplex3 = saturate (   offset0);
    simplex2 = saturate (-- offset0);
    simplex1 = saturate (-- offset0);
    simplex0 = 0;

	ZERO_DERIV3(simplex4, param);
	ZERO_DERIV3(simplex3, param);
	ZERO_DERIV3(simplex2, param);
	ZERO_DERIV3(simplex1, param);
	ZERO_DERIV3(simplex0, param);
}


//--------------------------------------------------------------------------------------
// 3D gradients
// These are used as a sort of gradient lookup table when doing noise evaluation.
//--------------------------------------------------------------------------------------
/// Not sure how this was working at all, some subtlety beyond me. But the compiler
/// was warning that these globals were being initialized, but a const float at the
/// global scope must be set from the app. Since these are never accessed from the
/// app, I made them static const's to quiet the compiler warnings, without changing
/// function. 
//cbuffer cbCubeGradients
//{
    static const float g_TexDims = 256.0f;
    static const float g_TexRandomDims = 1024.0f;
    
    // Gradients for the 3D cube that the 6 tetrahedra live in
    static const float3 g_grad3[16] = 
    {
        float3(1,1,0),float3(-1,1,0),float3(1,-1,0),float3(-1,-1,0),
        float3(1,0,1),float3(-1,0,1),float3(1,0,-1),float3(-1,0,-1),
        float3(0,1,1),float3(0,-1,1),float3(0,1,-1),float3(0,-1,-1),
        
        //Wrap around to let us use an AND instead of a MOD
        float3(1,1,0),float3(-1,1,0),float3(1,-1,0),float3(-1,-1,0)
    };

    // Gradients for the 4D hypercube that the 24 tetrahetra live in
    static const float4 g_grad4[32] = 
    {
        float4(0,1,1,1), float4(0,1,1,-1), float4(0,1,-1,1), float4(0,1,-1,-1),
        float4(0,-1,1,1), float4(0,-1,1,-1), float4(0,-1,-1,1), float4(0,-1,-1,-1),
        float4(1,0,1,1), float4(1,0,1,-1), float4(1,0,-1,1), float4(1,0,-1,-1),
        float4(-1,0,1,1), float4(-1,0,1,-1), float4(-1,0,-1,1), float4(-1,0,-1,-1),
        float4(1,1,0,1), float4(1,1,0,-1), float4(1,-1,0,1), float4(1,-1,0,-1),
        float4(-1,1,0,1), float4(-1,1,0,-1), float4(-1,-1,0,1), float4(-1,-1,0,-1),
        float4(1,1,1,0), float4(1,1,-1,0), float4(1,-1,1,0), float4(1,-1,-1,0),
        float4(-1,1,1,0), float4(-1,1,-1,0), float4(-1,-1,1,0), float4(-1,-1,-1,0),
    };
//};

#define F3 0.333333333333    // 1 / 3: magic number for 3D space noise
#define G3 0.166666666667    // 1 / 6: magic number for 3D space noise

/// Small test function to test out compiler. Performs no useful work. 
float4 XXX(float3 P : TEXCOORD0) : SV_Target
{

	float u = dot(P, F3);
	float3 Pi = floor(P + u);
	Pi`(P.x) = Pi`(P.y) = Pi`(P.z) = 0.0f;

	float3 Psq = Pi + P * P;

//	float3 Psq_dP = Psq`(P);
	float4 Psq_dP = EXTRACT_DERIV3(Psq.x, P);

	return float4(Psq_dP.xyz, 1.0f);
}

/// Since we need to manually unroll the loop to avoid using arrays,
/// pull the interior of the loop out into a function. 
float AddPointNoise3(float3 Pi, float3 corn, float3 pp)
{
    float3 tex = Pi + corn;
    tex /= g_TexDims;
        
    // We want an index, but the random values are normalized... so mul by 255
    uint index = g_txRandomByte.SampleLevel( rndByteSampler, tex.xy + tex.z, 0 ) * 255;
    index &= 15; // Same thing as %16
        
    float t = saturate( 0.6f - dot( pp, pp ) );
    t *= t; 
    return t * t * dot( g_grad3[index], pp );
}

//--------------------------------------------------------------------------------------
// 3D simplex noise
//--------------------------------------------------------------------------------------
float SimplexNoise3D( float3 P, const float3 param )
{
    
    // Skew input space to determine which simplex we're in
    float u = dot( P, F3 );     // ( P.x + P.y + P.z ) * F3;
    float3 Pi = floor( P + u );
    float v = dot( Pi, G3 );    // ( Pi.x + Pi.y + Pi.z ) * G3;
    float3 PO = Pi - v;         // Unskew the cell origin back to < x, y, z > space

	/// Symbolic Diff compiler currently doesn't support arrays or array elements.
	/// Converted float3 p[4] to pp0-pp3. 

    float3 pp0 = P - PO;              // The x, y, z distances from the cell origin

	/// Pi has no interesting derivative, it's essentially a constant here.
	/// Explicitly set partials to zero, so compiler can get around it. 
	ZERO_DERIV3(Pi, param);
    
    // For the 3D case, the simplex shape is a tetrahedron.
    // Find the 4 corners of the simplex
	/// Converted corner from array to individual variables to appease compiler. 
    float3 corn0;
	float3 corn1;
	float3 corn2;
	float3 corn3;
    Simplex3D( pp0, corn0, corn1, corn2, corn3, param );

	/// Corners have no interesting derivative, they are essentiall constants here.
	/// Explicitly set partials to zero so compiler doesn't worry about them. 
	ZERO_DERIV3(corn0, param);
	ZERO_DERIV3(corn1, param);
	ZERO_DERIV3(corn2, param);
	ZERO_DERIV3(corn3, param);

    //p[0] = p[0] - corner[0] + 0*G3;  // This line is commented out because, while it's complete to include it,
                                       // the value of p[0] turns out to be p[0].
    float3 pp1 = pp0 - corn1 +   G3;
    float3 pp2 = pp0 - corn2 + 2*G3;
    float3 pp3 = pp0 - corn3 + 3*G3;

    float totalNoise = 0;

	/// Having eliminated arrays, manually unroll loop. No extra work, since it was
	/// already being unrolled. Just longer code. 
	totalNoise += AddPointNoise3(Pi, corn0, pp0);
	totalNoise += AddPointNoise3(Pi, corn1, pp1);
	totalNoise += AddPointNoise3(Pi, corn2, pp2);
	totalNoise += AddPointNoise3(Pi, corn3, pp3);

    // Scale by 32 to get us into unit range
	float output = 32.0f * totalNoise;

	return output;
}

float4 YYY(float3 P : TEXCOORD0) : SV_Target
{
	float output = SimplexNoise3D( P, P );

//	float3 dOutput = float3(output`(P.x), output`(P.y), output`(P.z));
	float4 dOutput = EXTRACT_DERIV3(output, P);

	return float4(dOutput.xyz, 1.0f);
}

/// Since we need to manually unroll the loop to avoid using arrays,
/// pull the interior of the loop out into a function. 
float AddPointNoise4(float4 Pi, float4 corn, float4 pp)
{
    float4 tex = Pi + corn;
    tex /= g_TexDims;
        
    // We want an index, but the random values are normalized... so mul by 255
    uint index = g_txRandomByte.SampleLevel( rndByteSampler, tex.xy + tex.z * tex.w, 0 ) * 255;
    index &= 31; // Same thing as %32
        
    float t = saturate( 0.6f - dot( pp, pp ) );
    t *= t; 
    return t * t * dot( g_grad4[index], pp );
}
//--------------------------------------------------------------------------------------
// 4D simplex noise
//--------------------------------------------------------------------------------------
float SimplexNoise4D( float4 P, const float3 param )
{
    #define F4 0.309016994375f // ( sqrt( 5 ) - 1 ) / 4
    #define G4 0.138196601125f // ( 5 - sqrt( 5 ) ) / 20

    // Skew the (x,y,z) space to determine which cell of 24 simplices we're in
    float u = dot( P, F4 );		// ( P.x + P.y + P.z + P.w ) * F4;
    float4 Pi = floor( P + u );
    float v = dot( Pi, G4 );	// ( Pi.x + Pi.y + Pi.z + Pi.w ) * G4;
    float4 P0 = Pi - v;			// Unskew the cell origin back to < x, y, z, w > space

	/// Symbolic Diff compiler currently doesn't support arrays or array elements.
	/// Converted float3 p[4] to pp0-pp3. 
    float4 pp0 = P - P0;				// The x, y, z, w distances from the cell origin

	/// Pi has no interesting derivative, it's essentially a constant here.
	/// Explicitly set partials to zero, so compiler can get around it. 
	ZERO_DERIV3(Pi, param);

    // To find out which of the 24 possible simplexes we're in
    float4 corn0, corn1, corn2, corn3, corn4;
    Simplex4D( pp0, corn0, corn1, corn2, corn3, corn4, param );

	/// Corners have no interesting derivative, they are essentiall constants here.
	/// Explicitly set partials to zero so compiler doesn't worry about them. 
	ZERO_DERIV3(corn0, param);
	ZERO_DERIV3(corn1, param);
	ZERO_DERIV3(corn2, param);
	ZERO_DERIV3(corn3, param);
	ZERO_DERIV3(corn4, param);

    //p[0] = p[0] - corner[0] + 0*G4;  // This line is commented out because, while it's complete to include it,
                                       // the value of p[0] turns out to be p[0].
    float4 pp1 = pp0 - corn1 +   G4;
    float4 pp2 = pp0 - corn2 + 2*G4;
    float4 pp3 = pp0 - corn3 + 3*G4;
    float4 pp4 = pp0 - corn4 + 4*G4;
    
    float totalNoise = 0;

	totalNoise += AddPointNoise4(Pi, corn0, pp0);
	totalNoise += AddPointNoise4(Pi, corn1, pp1);
	totalNoise += AddPointNoise4(Pi, corn2, pp2);
	totalNoise += AddPointNoise4(Pi, corn3, pp3);
	totalNoise += AddPointNoise4(Pi, corn4, pp4);


    // Scale by 32 to get us into unit range
//    return 27.0f * totalNoise;
	float output = 27.0f * totalNoise;

	return output;
}

float4 YYY4(float4 P : TEXCOORD0) : SV_Target
{
	float output = SimplexNoise4D( P, P.xyz );

//	float3 dOutput = float3(output`(P.x), output`(P.y), output`(P.z));
	float4 dOutput = EXTRACT_DERIV3(output, P);

	return float4(dOutput.xyz, 1.0f);
}


// Noise types
#define NT_SIMPLEX3D            0    // Standard 3D simplex noise
#define NT_SIMPLEX4D            1    // 4D simplex noise; good for animating
#define NT_SIMPLEX3D_TURB       2    // 3D simplex noise with turbulence
#define NT_SIMPLEX4D_TURB       3    // 4D simplex noise with turulence
#define NT_VORONOI3             4    // Voronoi noise evaluated over a 3x3x3 grid
#define NT_VORONOI3_2           5    // Voronoi noise over a 3x3x3 grid, but using the second distance function
#define NT_VORONOI2             6    // Voronoi noise evaluated over a 2x2x2 grid (faster, but some artifacts)
#define NT_VORONOI2_2           7    // Voronoi noise over a 2x2x2 grid, but using the second distance function


//--------------------------------------------------------------------------------------
// Voronoi noise overview
// For each sample position, we are trying to determine the distance to the closest point in
// a collection of randomly placed points in 3D space.  Because iterating over a series
// of points in space would take a very LONG time, we divide space up into a regular grid.
// The coordinates of each cell of this grid are used to create a hash that indexes into
// a set of random vectors stored in g_txRandVector.  This random vector corresponds to
// the location of a point inside that grid cell.  For each point in space, we examine
// all 9 surrounding grid cells and find the closest point for all 9 cells.  The methods
// that implement this routine are the 3 methods (for 3x3x3 surrounding cells).
//
// We also use a slightly faster method labelled the 2 method.  This method only looks at 4
// surrounding cells.  This means that we could miss the closest point in some cases.
// To increase the odds that we'll find the closest point, each cell in the 2X2 method
// contains 3 random points.  It's still not foolproof, but it's faster than the 3X3 method.
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Voronoi cell lookup - This gets a random point for a particular grid cell in 3D space.
// W is a random variation to the lookup.  Changing W with time will animate the 
// Voronoi noise.
//--------------------------------------------------------------------------------------
float3 GetCellPoint( float3 Cell, float W )
{
    float3 NewCell = Cell;
    
    // This is the hash for which random index corresponds to which cell
    float hashIndex = ( ( NewCell.z / NewCell.y ) * NewCell.x ) + ( NewCell.x * NewCell.y );
    hashIndex += W;
    hashIndex /= g_TexRandomDims;
    float3 randPt = g_txRandVector.SampleLevel( rndVecSampler, hashIndex, 0 ).xyz;
    
    // Translate the random sampled point (randPt) to the center of the cell
    // and make sure it stays within the cell boundaries
    return Cell + float3( 0.5, 0.5, 0.5 ) + randPt * 0.5;
}

float TestCellPoint( const float3 P, const float3 cell, const float3 step, const float W, const float3 param)
{
    float3 delta = GetCellPoint( cell + step, W ) - P;
	ZERO_DERIV3(delta, param);
    float testDist = dot( delta, delta );
	testDist`(param.x) = testDist * 2.0f * delta.x;
	testDist`(param.y) = testDist * 2.0f * delta.y;
	testDist`(param.z) = testDist * 2.0f * delta.z;

	return testDist;
}
float Voronoi3_Unroll( float3 P, float W, const float3 param )
{
    float3 Cell = floor( P );
	ZERO_DERIV3(Cell, param);

	float test;
	// Z = -1
	// Y = -1
	// X=[-1..1]
	float output = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, -1.0f), W, param);

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, -1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, -1.0f), W, param);
	output = min(output, test);

	// Y = 0
	// X=[-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, -1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, -1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, -1.0f), W, param);
	output = min(output, test);

	// Y = 1
	// X=[-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, -1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, -1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, -1.0f), W, param);
	output = min(output, test);

	// Z = 0
	// Y = -1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, 0.0f), W, param);
	output = min(output, test);

	// Y = 0
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, 0.0f), W, param);
	output = min(output, test);

	// Y = 1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, 0.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, 0.0f), W, param);
	output = min(output, test);

	
	// Z = 1
	// Y = -1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, 1.0f), W, param);
	output = min(output, test);

	// Y = 0
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, 1.0f), W, param);
	output = min(output, test);

	// Y = 1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, 1.0f), W, param);
	output = min(output, test);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, 1.0f), W, param);
	output = min(output, test);


	return output;
}
void CheckSecondMin(float test, inout float minDistSq, inout float output)
{
	if(test < minDistSq)
	{
		output = minDistSq;
		minDistSq = test;
	}
	else
	{
		output = min(output, test);
	}
}
float Voronoi3_2_Unroll( float3 P, float W, const float3 param )
{
    float3 Cell = floor( P );
	ZERO_DERIV3(Cell, param);

	float test;
	// Z = -1
	// Y = -1
	// X=[-1..1]
	float minDistSq = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, -1.0f), W, param);
	float output = 1000.0f;

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 0
	// X=[-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 1
	// X=[-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, -1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Z = 0
	// Y = -1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 0
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, 0.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	
	// Z = 1
	// Y = -1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, -1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, -1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, -1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 0
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 0.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 0.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 0.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	// Y = 1
	// X = [-1..1]
	test = TestCellPoint(P, Cell, float3(-1.0f, 1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(0.0f, 1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);

	test = TestCellPoint(P, Cell, float3(1.0f, 1.0f, 1.0f), W, param);
	CheckSecondMin(test, minDistSq, output);


	return output;
}
//--------------------------------------------------------------------------------------
// Voronoi noise 3x3x3 block.
// W is a random variation to the lookup.  Changing W with time will animate the 
// Voronoi noise.
//--------------------------------------------------------------------------------------
float Voronoi3_Loop( float3 P, float W, const float3 param )
{

    float3 Cell = floor( P );
	ZERO_DERIV3(Cell, param);

    float3 Iter;

	float output = 100.0f; // Set mindistSq to a high number
	float4 deriv = float4(0.0f, 0.0f, 0.0f, 0.0f);
    
    for( Iter.x = Cell.x - 1; Iter.x < Cell.x + 2; Iter.x += 1 )
    {
        for( Iter.y = Cell.y - 1; Iter.y < Cell.y + 2; Iter.y += 1 )
        {
            for( Iter.z = Cell.z - 1; Iter.z < Cell.z + 2; Iter.z += 1 )
            {
                // Get the squared distance to the random sample point.
                float3 delta = GetCellPoint( Iter, W ) - P;
				ZERO_DERIV3(delta, param);
                float testDist = dot( delta, delta );
				if(output > testDist)
				{
					output = testDist;
#ifndef MAF_SIMPLE_FAILURE
					/// Compiler currently having trouble with d testDist / d delta partials.
					/// Trivial derivative, set manually. 
					deriv = float4(
						testDist * 2.0f * delta.x,
						testDist * 2.0f * delta.y,
						testDist * 2.0f * delta.z,
						0.0f);
#else // MAF_SIMPLE_FAILURE
					deriv = float4(
						testDist`(delta.x),
						testDist`(delta.y),
						testDist`(delta.z),
						0.0f);
#endif // MAF_SIMPLE_FAILURE
				}
            }
        }
    }

	output`(param.x) = deriv.x;    
	output`(param.y) = deriv.y;    
	output`(param.z) = deriv.z;    
    
	return output;
}

//--------------------------------------------------------------------------------------
// Voronoi noise 3x3x3 block second distance.
// W is a random variation to the lookup.  Changing W with time will animate the 
// Voronoi noise.
//--------------------------------------------------------------------------------------
float Voronoi3_2_Loop( float3 P, float W, const float3 param )
{
    float3 Cell = floor( P );
	ZERO_DERIV3(Cell, param);

    float mindistSq = 100.0f;   // Set mindistSq to a high number
	float4 mindistSqDeriv = float4(0.0f, 0.0f, 0.0f, 0.0f);
    float3 Iter;

	float output = 100.0f;
	float4 deriv = float4(0.0f, 0.0f, 0.0f, 0.0f);
    
    for( Iter.x = Cell.x - 1; Iter.x < Cell.x + 2 ; Iter.x += 1 )
    {
        for( Iter.y = Cell.y - 1; Iter.y < Cell.y + 2; Iter.y += 1 )
        {
            for( Iter.z = Cell.z - 1; Iter.z < Cell.z + 2; Iter.z += 1 )
            {
                // Get the squared distance to the random sample point.
                float3 delta = GetCellPoint( Iter, W ) - P;
				ZERO_DERIV3(delta, param);
                float testDist = dot( delta, delta );
                if( testDist < mindistSq )
                {
                    output = mindistSq;
					deriv = mindistSqDeriv;
                    mindistSq = testDist;
#ifndef MAF_SIMPLE_FAILURE
					/// Compiler currently having trouble with d testDist / d delta partials.
					/// Trivial derivative, set manually. 
					mindistSqDeriv = float4(
						testDist * 2.0f * delta.x,
						testDist * 2.0f * delta.y,
						testDist * 2.0f * delta.z,
						0.0f);
#else // MAF_SIMPLE_FAILURE
					mindistSqDeriv = float4(
						testDist`(delta.x),
						testDist`(delta.y),
						testDist`(delta.z),
						0.0f);
#endif //  MAF_SIMPLE_FAILURE
                }
                else
                {
					if(testDist < output)
					{
						output = testDist;
#ifndef MAF_SIMPLE_FAILURE
						/// Compiler currently having trouble with d testDist / d delta partials.
						/// Trivial derivative, set manually. 
						deriv = float4(
							testDist * 2.0f * delta.x,
							testDist * 2.0f * delta.y,
							testDist * 2.0f * delta.z,
							0.0f);
#else // MAF_SIMPLE_FAILURE
						deriv = float4(
							testDist`(delta.x),
							testDist`(delta.y),
							testDist`(delta.z),
							0.0f);
#endif // MAF_SIMPLE_FAILURE
					}
                }
            }
        }
    }
	output`(param.x) = deriv.x;    
	output`(param.y) = deriv.y;    
	output`(param.z) = deriv.z;    
    
	return output;
}

float Voronoi2_Unroll( float3 P, float W, const float3 param )
{
	P += float3( 0.5f, 0.5f, 0.5f );
	float output = Voronoi3_Unroll( P, W, param);

	float test = Voronoi3_Unroll( P, W + 1 * 4, param );
	output = min(test, output);

	test = Voronoi3_Unroll( P, W + 2 * 4, param );
	output = min(test, output);

	return output;
}
/// $BUG - This isn't quite right. We should be passing the current minDistSq and output
/// down into the Voronoi3_2 test so that they can find the second min across all W shifts.
/// This just finds the min of the 3 second minimums. But we aren't using this function 
/// right now anyway. So not sure which is really more "correct". Just noting that this doesn't
/// match the loop version exactly.
float Voronoi2_2_Unroll( float3 P, float W, const float3 param )
{
	P += float3( 0.5f, 0.5f, 0.5f );

	float output = Voronoi3_2_Unroll( P, W, param);

	float test = Voronoi3_2_Unroll( P, W + 1 * 4, param );
	output = min(output, test);

	test = Voronoi3_2_Unroll( P, W + 2 * 4, param );
	output = min(output, test);

	return output;
}
//--------------------------------------------------------------------------------------
// Voronoi noise 2x2x2 block approximation to the 3x3x3 version.
// W is a random variation to the lookup.  Changing W with time will animate the 
// Voronoi noise.
//--------------------------------------------------------------------------------------
float Voronoi2_Loop( float3 P, float W, const float3 param )
{
    float3 Cell = floor( P + float3( 0.5, 0.5, 0.5 ) );  // Bias to the position between the 8 neighboring cells
	ZERO_DERIV3(Cell, param);

    float3 Iter;
    
	float output = 100.0f;
	float4 deriv = float4(0.0f, 0.0f, 0.0f, 0.0f);

    for( Iter.x = Cell.x - 1; Iter.x < Cell.x + 1; Iter.x += 1 )
    {
        for( Iter.y = Cell.y - 1; Iter.y < Cell.y + 1; Iter.y += 1 )
        {
            for( Iter.z = Cell.z - 1; Iter.z < Cell.z + 1; Iter.z += 1 )
            {
                // Sample 3 random points by shifting W
                for( int WShift = 0; WShift < 3; WShift++ )
                {
                    // Sample 3 points and shift W by WShift * 4.  Why 4?  Why not.
                    // Then get the squared distance to the random sample point.
                    float3 delta = GetCellPoint( Iter, W + WShift * 4 ) - P;
					ZERO_DERIV3(delta, param);
                    float testDist = dot( delta, delta );
					if(testDist < output)
					{
						output = testDist;
#ifndef MAF_SIMPLE_FAILURE
						/// Compiler currently having trouble with d testDist / d delta partials.
						/// Trivial derivative, set manually. 
						deriv = float4(
							testDist * 2.0f * delta.x,
							testDist * 2.0f * delta.y,
							testDist * 2.0f * delta.z,
							0.0f);
#else // MAF_SIMPLE_FAILURE
						deriv = float4(
							testDist`(delta.x),
							testDist`(delta.y),
							testDist`(delta.z),
							0.0f);
#endif // MAF_SIMPLE_FAILURE
					}
                }
            }
        }
    }
	deriv *= 0.75f;

	output`(param.x) = deriv.x;    
	output`(param.y) = deriv.y;    
	output`(param.z) = deriv.z;    
    
    return output;
}

//--------------------------------------------------------------------------------------
// Voronoi noise 2x2x2 block second distance approximation to the 3x3x3 version.
// W is a random variation to the lookup.  Changing W with time will animate the 
// Voronoi noise.
//--------------------------------------------------------------------------------------
float Voronoi2_2_Loop( float3 P, float W, const float3 param )
{
    float3 Cell = floor( P + float3(0.5,0.5,0.5) );  // Bias to the position between the 8 neighboring cells
	ZERO_DERIV3(Cell, param);

    float mindistSq = 100.0f;  // Set mindistSq to a high number
	float4 mindistSqDeriv = float4(0.0f, 0.0f, 0.0f, 0.0f);
    float3 Iter;

	float output = 100.0f;
	float4 deriv = float4(0.0f, 0.0f, 0.0f, 0.0f);
    
    for( Iter.x = Cell.x - 1; Iter.x < Cell.x + 1; Iter.x += 1 )
    {
        for( Iter.y = Cell.y - 1; Iter.y < Cell.y + 1; Iter.y += 1 )
        {
            for( Iter.z = Cell.z - 1; Iter.z < Cell.z + 1; Iter.z += 1 )
            {
                // Sample 3 random points by shifting W
                for( int WShift = 0; WShift < 3; WShift++ )
                {
                    // Sample 3 points and shift W by WShift * 4.  Why 4?  Why not.
                    // Then get the squared distance to the random sample point.
                    float3 delta = GetCellPoint( Iter, W + WShift * 4 ) - P;
					ZERO_DERIV3(delta, param);
                    float testDist = dot( delta, delta );
                    if( testDist < mindistSq )
                    {
                        output = mindistSq;
						deriv = mindistSqDeriv;
                        mindistSq = testDist;
#ifndef MAF_SIMPLE_FAILURE
						/// Compiler currently having trouble with d testDist / d delta partials.
						/// Trivial derivative, set manually. 
						mindistSqDeriv = float4(
							testDist * 2.0f * delta.x,
							testDist * 2.0f * delta.y,
							testDist * 2.0f * delta.z,
							0.0f);
#else // MAF_SIMPLE_FAILURE
						mindistSqDeriv = float4(
							testDist`(delta.x),
							testDist`(delta.y),
							testDist`(delta.z),
							0.0f);
#endif // MAF_SIMPLE_FAILURE
                    }
                    else
                    {
						if(testDist < output)
						{
							output = testDist;
#ifndef MAF_SIMPLE_FAILURE
							/// Compiler currently having trouble with d testDist / d delta partials.
							/// Trivial derivative, set manually. 
							deriv = float4(
								testDist * 2.0f * delta.x,
								testDist * 2.0f * delta.y,
								testDist * 2.0f * delta.z,
								0.0f);
#else // MAF_SIMPLE_FAILURE
							deriv = float4(
								testDist`(delta.x),
								testDist`(delta.y),
								testDist`(delta.z),
								0.0f);
#endif // MAF_SIMPLE_FAILURE
						}
                    }
                }
            }
        }
    }
    deriv *= 1.5f;

	output`(param.x) = deriv.x;    
	output`(param.y) = deriv.y;    
	output`(param.z) = deriv.z;    
    
    return output;
}

//--------------------------------------------------------------------------------------
// Select different Voronoi noises based upon the input
//--------------------------------------------------------------------------------------
/// Modified to return not just the noise value but it's partial derivatives with
/// respect to the input P. 
float VoronoiNoise( float3 P, float W, uniform uint NoiseType, const float3 param )
{
    float output;
    
	[flatten]switch(NoiseType)
    {
		case NT_VORONOI3:
#ifdef MAF_UNROLL_VORONOI
			output = Voronoi3_Unroll( P, W, param );
#else // MAF_UNROLL_VORONOI
			output = Voronoi3_Loop( P, W, param );
#endif // MAF_UNROLL_VORONOI
			break;

		case NT_VORONOI3_2:
#ifdef MAF_UNROLL_VORONOI
			output = Voronoi3_2_Unroll( P, W, param );
#else // MAF_UNROLL_VORONOI
			output = Voronoi3_2_Loop( P, W, param );
#endif // MAF_UNROLL_VORONOI
			break;

		case NT_VORONOI2:
#ifdef MAF_UNROLL_VORONOI
			output = Voronoi2_Unroll( P, W, param );
#else // MAF_UNROLL_VORONOI
			output = Voronoi2_Loop( P, W, param );
#endif // MAF_UNROLL_VORONOI
			break;
		
		default: // if( NoiseType == NT_VORONOI2_2 )
#ifdef MAF_UNROLL_VORONOI
			output = Voronoi2_2_Unroll( P, W, param );
#else // MAF_UNROLL_VORONOI
			output = Voronoi2_2_Loop( P, W, param );
#endif // MAF_UNROLL_VORONOI
			break;
	}
        
    return output;
}

//--------------------------------------------------------------------------------------
// Sum octaves of noise at different frequencies.  For simplex noise, this produces the
// traditional "cloud" look.
//--------------------------------------------------------------------------------------
/// Modified to return not just the noise value but it's partial derivatives with respect
/// to the input vInputCoords. Noise and derivatives are linearly accumulated over specified
/// number of octaves. 
float fBm( float4 vInputCoords, uniform uint nNumOctaves, uniform float fLacunarity, 
           uniform float fGain, uniform uint NoiseType, const float3 param )
{
	/// Zero out accumulator. 
	float noiseSum = 0.0f;
	/// Need to accumulate derivative as well as signal. 
	float3 derivSum = float3(0.0f, 0.0f, 0.0f);

    float fAmplitude = 1;
    float fAmplitudeSum = 0;
	float fTotalVertexScale = 1.0f;
    
    float4 vSampleCoords = vInputCoords;
    
    [unroll]
    for( uint i=0; i<nNumOctaves; i++ )
    {
		float noise = 0.0f;
		float3 deriv = float3(0.0f, 0.0f, 0.0f);
        if( NoiseType == NT_SIMPLEX3D )
		{
			noise = SimplexNoise3D( vSampleCoords.xyz, param.xyz );
			deriv = EXTRACT_DERIV3(noise, param).xyz;
		}
        else if( NoiseType == NT_SIMPLEX4D )
		{
            noise = SimplexNoise4D( vSampleCoords, param.xyz );
			deriv = EXTRACT_DERIV3(noise, param).xyz;
		}    
        else if( NoiseType == NT_SIMPLEX3D_TURB )
		{
			noise = SimplexNoise3D( vSampleCoords.xyz, param.xyz );
			deriv = EXTRACT_DERIV3(noise, param).xyz;
			if(noise < 0)
			{
				/// Rather than noise.output = abs(noise.output), if
				/// the output is negative, we negate it and its derivative. 
				noise = -noise;
				deriv = -deriv;
			}
		}
        else if( NoiseType == NT_SIMPLEX4D_TURB )
		{
            noise = SimplexNoise4D( vSampleCoords, param.xyz );
			deriv = EXTRACT_DERIV3(noise, param).xyz;
			if(noise < 0)
			{
				/// Rather than noise.output = abs(noise.output), if
				/// the output is negative, we negate it and its derivative. 
				noise = -noise;
				deriv = -deriv;
			}
		}
        else if( NoiseType >= NT_VORONOI3 )
		{
            noise = VoronoiNoise( vSampleCoords.xyz, vSampleCoords.w, NoiseType, param.xyz );
			deriv = EXTRACT_DERIV3(noise, param).xyz;
		}
		
		/// Derivative must be scaled by the total scale applied to the sample coords
		/// so far.
		deriv *= fTotalVertexScale;
		
		noiseSum += fAmplitude * noise;
		derivSum += fAmplitude * deriv;

		/// Accumulate both noise signal and derivative, both modulated by amplitude. 
        fAmplitudeSum += fAmplitude;
        fAmplitude *= fGain;
        vSampleCoords *= fLacunarity;
		fTotalVertexScale *= fLacunarity;
    }
    
	/// Normalize both signal and derivative by summed amplitude. 
    noiseSum /= fAmplitudeSum;
	derivSum /= fAmplitudeSum;

	noiseSum`(param.x) = derivSum.x;
	noiseSum`(param.y) = derivSum.y;
	noiseSum`(param.z) = derivSum.z;
    return noiseSum;
}

float4 ZZZ(float3 P : TEXCOORD0) : SV_Target
{
	float output = fBm( float4(P, 0.0f), 8, 2.0, 0.5, NT_SIMPLEX3D, P );

	float3 dOutput = float3(output`(P.x), output`(P.y), output`(P.z));

	return float4(dOutput.xyz, 1.0f);
}

#endif // #ifndef NOISE_FX