
#ifndef PARAMETRIC_3D_PATCH_FX
#define PARAMETRIC_3D_PATCH_FX

#include "Parametric3DSurface.fx"

#define CONTROL_PTS_PER_SIDE	(2)
#define INPUT_PATCH_SIZE	(CONTROL_PTS_PER_SIDE * CONTROL_PTS_PER_SIDE)
#define OUTPUT_PATCH_SIZE	(INPUT_PATCH_SIZE)

struct HS_INPUT
{
	float2 vDomain			: DOMAINPOS;
};

struct HS_CONSTANT_DATA_OUTPUT
{
    float Edges[4]          : SV_TessFactor;
    float Inside[2]         : SV_InsideTessFactor;
	uint PatchID			: PATCHID;
};

struct HS_OUTPUT
{
    float2 vDomain          : DOMAINPOS;
};

HS_INPUT VS_Patch( const VSHARDWARE_INPUT input )
{
	HS_INPUT output;
	output.vDomain = DomainPoint( input );
	return output;
}


// This constant hull shader is executed once per patch. In a more complex scene,
// you might calculate a variable tessellation factor based on the camera's distance.

HS_CONSTANT_DATA_OUTPUT HS_PatchConstant( InputPatch<HS_INPUT, INPUT_PATCH_SIZE> ip,
                                          uint PatchID : SV_PrimitiveID )
{    
    HS_CONSTANT_DATA_OUTPUT Output;

    float TessAmount = TessFactor.x;

    Output.Edges[0] = Output.Edges[1] = Output.Edges[2] = Output.Edges[3] = TessAmount;
    Output.Inside[0] = Output.Inside[1] = TessAmount;

	Output.PatchID = PatchID;

    return Output;
}

// The hull shader is called once per output control point, which is specified with
// outputcontrolpoints.  For this sample, we take the control points from the vertex
// shader and pass them directly off to the domain shader.  In a more complex scene,
// you might perform a basis conversion from the input control points into a Bezier
// patch, such as the SubD11 Sample.

// The input to the hull shader comes from the vertex shader

// The output from the hull shader will go to the domain shader.
// The tessellation factor, topology, and partition mode will go to the fixed function
// tessellator stage to calculate the UVW and domain points.

[domain("quad")]
[partitioning("pow2")]
[outputtopology("triangle_cw")]
[outputcontrolpoints(OUTPUT_PATCH_SIZE)]
[patchconstantfunc("HS_PatchConstant")]
HS_OUTPUT HS_Patch( InputPatch<HS_INPUT, INPUT_PATCH_SIZE> p, 
                    uint i : SV_OutputControlPointID,
                    uint PatchID : SV_PrimitiveID )
{
    HS_OUTPUT Output;
    Output.vDomain = p[i].vDomain;
    return Output;
}

//--------------------------------------------------------------------------------------
// Patch evaluation domain shader section
//--------------------------------------------------------------------------------------
// The domain shader is run once per vertex and calculates the final vertex's position
// and attributes.  It receives the UVW from the fixed function tessellator and the
// control point outputs from the hull shader.  Since we are using the DirectX 11
// Tessellation pipeline, it is the domain shader's responsibility to calculate the
// final SV_POSITION for each vertex.  In this sample, we evaluate the vertex's
// position using a Bernstein polynomial and the normal is calculated as the cross
// product of the U and V derivatives.

// The input SV_DomainLocation to the domain shader comes from fixed function
// tessellator.  And the OutputPatch comes from the hull shader.  From these, you
// must calculate the final vertex position, color, texcoords, and other attributes.

// The output from the domain shader will be a vertex that will go to the video card's
// rasterization pipeline and get drawn to the screen.

float2 DomainPoint(const HS_CONSTANT_DATA_OUTPUT input, const float2 UV)
{
	float2 domainPoint;
	domainPoint.x = (UV.x + (input.PatchID % (int)ModDivScale.x)) * ModDivScale.z;
	domainPoint.y = (UV.y + (input.PatchID / (int)ModDivScale.y)) * ModDivScale.w;

	return domainPoint;
}



PS_INPUT_HARDWARE DS_BasicPatch( const PARAMETRICSURFACE3D_OUTPUT surface, const float2 UV )
{
	float3 localPos = surface.Position;

	float3 worldPos = mul(float4(localPos, 1), LocalToWorld).xyz;
	float3 viewPos = mul(float4(worldPos, 1), WorldToCamera).xyz;

    PS_INPUT_HARDWARE Output;
    Output.Position = mul( float4(viewPos,1), CameraToNDC);
	Output.LocalPos = localPos;

#ifdef NORMAL_IN_VS
	float3 dF_du = surface.Position`(UV.x);
	float3 dF_dv = surface.Position`(UV.y);
	float3 norm = cross(dF_du, dF_dv);
	Output.LocalNormal = NORMALIZE_CHECK(norm);
#endif // NORMAL_IN_VS

	Output.DomainPos = UV;

	return Output;
}

PS_INPUT_HARDWARE DS_ProfileProduct( HS_CONSTANT_DATA_OUTPUT input, 
                    float2 UV : SV_DomainLocation,
                    const OutputPatch<HS_OUTPUT, OUTPUT_PATCH_SIZE> patch )
{
	UV = DomainPoint(input, UV);
	const PARAMETRICSURFACE3D_OUTPUT surface = ProfileProduct_EvaluateInterp( UV );
	
    return DS_BasicPatch( surface, UV );    
}

PS_INPUT_HARDWARE DS_SurfaceOfRevolution( HS_CONSTANT_DATA_OUTPUT input, 
                    float2 UV : SV_DomainLocation,
                    const OutputPatch<HS_OUTPUT, OUTPUT_PATCH_SIZE> patch )
{
	UV = DomainPoint(input, UV);
	const PARAMETRICSURFACE3D_OUTPUT surface = SurfaceOfRevolution_EvaluateInterp( UV );
	
    return DS_BasicPatch( surface, UV );    
}

PS_INPUT_HARDWARE DS_WireProduct( HS_CONSTANT_DATA_OUTPUT input, 
                    float2 UV : SV_DomainLocation,
                    const OutputPatch<HS_OUTPUT, OUTPUT_PATCH_SIZE> patch )
{
	UV = DomainPoint(input, UV);
	const PARAMETRICSURFACE3D_OUTPUT surface = WireProduct_EvaluateInterp( UV );
	
    return DS_BasicPatch( surface, UV );    
}

float4 PS_Patch( const PS_INPUT_HARDWARE input ) : SV_Target
{
	return PS_Grid(input);
}

#endif // PARAMETRIC_3D_PATCH_FX