
#ifndef PARAMETRIC_3D_SURFACE_FX
#define PARAMETRIC_3D_SURFACE_FX

//#define NORMAL_IN_VS
#define NEIGHBOR_INTERP
#define MAF_ENABLE_PROC_MATERIALS

//------------------------------------------------------------------------------
//
// Useful constants
//
//------------------------------------------------------------------------------
static const float    PI = 3.141592654;
static const float TWOPI = 6.283185307;

//------------------------------------------------------------------------------
//
// 3D parametric surface definitions
//
//------------------------------------------------------------------------------

struct PARAMETRICSURFACE3D_OUTPUT
{
    float3 Position     : POSITION;         // 3D position on the plane
};


//------------------------------------------------------------------------------
//
// Common parameters for shading and transformation.
//
//------------------------------------------------------------------------------

cbuffer cbMaterial : register(b3)
{
	float4 LightVector = normalize(float4(.8, .7, -1, 0.0));
	float4 TessFactor;
};

#include "BaseBuffers.fx"

#include "ProcSurface.fx"

#include "Reproj.fx"

//------------------------------------------------------------------------------
//
// 3D parametric surface methods
//
//------------------------------------------------------------------------------

bool2 IsOdd(const float2 uv)
{
	uint numVerts = ceil(TessFactor.x);
	float2 fIntUV =  uv.xy * numVerts.xx + 0.5f.xx;
	uint2 intUV = uint2(fIntUV.x, fIntUV.y);
	bool2 odd = (intUV % uint2(2,2)) != uint2(0,0);

	return odd;
}

float2 GetStepToNeighbors(bool2 odd)
{
	return 1.0f.xx / ceil(TessFactor.xx);
}

PARAMETRICSURFACE3D_OUTPUT ProfileProduct_EvaluateInterp( const float2 input )
{   
#ifdef NEIGHBOR_INTERP
	const bool2 odd = IsOdd(input);
	const float2 step = GetStepToNeighbors(odd);
#endif // NEIGHBOR_INTERP

    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate_FuncU( input.x );
    profileCurve.Position *= .004;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate_FuncV( input.y );
    crossSectionCurve.Position *= .01;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	profileCurve.Position`(input.y) = 0;
	crossSectionCurve.Position`(input.x) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = crossSectionCurve.Position.x * profileCurve.Position.x;
    output.Position.y = crossSectionCurve.Position.y * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    

#ifdef NEIGHBOR_INTERP
	float2 midProfile = profileCurve.Position;
	if(odd.x)
	{
		float lo = max(0.0f, input.x - step.x);
		PARAMETRICCURVE2D_OUTPUT profileCurveLo = BSpline_Evaluate_FuncU( lo );
		float2 positionLo = profileCurveLo.Position * 0.004f;

		float hi = min(1.0f, input.x + step.x);
		PARAMETRICCURVE2D_OUTPUT profileCurveHi = BSpline_Evaluate_FuncU( hi );
		float2 positionHi = profileCurveHi.Position * 0.004f;

		float2 norm = (positionHi - positionLo).yx;
		norm.y = -norm.y;
		float t = dot(positionLo, norm) - dot(profileCurve.Position, norm);
		t /= dot(norm, norm);
		
		midProfile += t * norm;
	}

	float2 midCrossSection = crossSectionCurve.Position;
	if(odd.y)
	{
		float lo = max(0.0f, input.y - step.y);
		PARAMETRICCURVE2D_OUTPUT crossSectionCurveLo = BSpline_Evaluate_FuncV( lo );
		float2 positionLo = crossSectionCurveLo.Position * 0.01f;

		float hi = min(1.0f, input.y + step.y);
		PARAMETRICCURVE2D_OUTPUT crossSectionCurveHi = BSpline_Evaluate_FuncV( hi );
		float2 positionHi = crossSectionCurveHi.Position * 0.01f;

		float2 norm = (positionHi - positionLo).yx;
		norm.y = -norm.y;
		float t = dot(positionLo, norm) - dot(crossSectionCurve.Position, norm);
		t /= dot(norm, norm);

		midCrossSection += t * norm;
	}
	float3 midPosition;
	midPosition.xy = midProfile.xx * midCrossSection.xy;
	midPosition.z = 0.7 - midProfile.y;

	output.Position = lerp(midPosition, output.Position, min(1.0f, frac(TessFactor.xxx) * 1.0f));
#endif // NEIGHBOR_INTERP
    
    return output;
}

PARAMETRICSURFACE3D_OUTPUT ProfileProduct_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate_FuncU( input.x );
    profileCurve.Position *= .004;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate_FuncV( input.y );
    crossSectionCurve.Position *= .01;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	profileCurve.Position`(input.y) = 0;
	crossSectionCurve.Position`(input.x) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = crossSectionCurve.Position.x * profileCurve.Position.x;
    output.Position.y = crossSectionCurve.Position.y * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    
    
    return output;
}

float2 CosSin_Evaluate(const float theta)
{
	return float2(cos(theta), -sin(theta));
}
PARAMETRICSURFACE3D_OUTPUT SurfaceOfRevolution_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate_FuncU( input.x );
    profileCurve.Position *= .004;
  
    float theta = TWOPI * input.y;

	float2 crossSectionCurve = CosSin_Evaluate(theta);

	/// Explicitly tell the compiler there is no interdependence on parameters.
	crossSectionCurve`(input.x) = 0;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	profileCurve.Position`(input.y) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = crossSectionCurve.x * profileCurve.Position.x;
    output.Position.y = crossSectionCurve.y * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    
    
    return output;
}

PARAMETRICSURFACE3D_OUTPUT SurfaceOfRevolution_EvaluateInterp( const float2 input )
{    
#ifdef NEIGHBOR_INTERP
	bool2 odd = IsOdd(input.xy);
	float2 step = GetStepToNeighbors(odd);
#endif // NEIGHBOR_INTERP

    PARAMETRICCURVE2D_OUTPUT profileCurve = BSpline_Evaluate_FuncU( input.x );
    profileCurve.Position *= .004;
  
    float theta = TWOPI * input.y;

	float2 crossSectionCurve = CosSin_Evaluate(theta);

	/// Explicitly tell the compiler there is no interdependence on parameters.
	crossSectionCurve`(input.x) = 0;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	profileCurve.Position`(input.y) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = crossSectionCurve.x * profileCurve.Position.x;
    output.Position.y = crossSectionCurve.y * profileCurve.Position.x;
    output.Position.z = .7 - profileCurve.Position.y;
    
    
#ifdef NEIGHBOR_INTERP
	float2 midProfile = profileCurve.Position;
	if(odd.x)
	{
		float lo = max(0.0f, input.x - step.x);
		PARAMETRICCURVE2D_OUTPUT profileCurveLo = BSpline_Evaluate_FuncU( lo );
		float2 positionLo = profileCurveLo.Position * 0.004f;

		float hi = min(1.0f, input.x + step.x);
		PARAMETRICCURVE2D_OUTPUT profileCurveHi = BSpline_Evaluate_FuncU( hi );
		float2 positionHi = profileCurveHi.Position * 0.004f;

		float2 norm = (positionHi - positionLo).yx;
		norm.y = -norm.y;
		float t = dot(positionLo, norm) - dot(profileCurve.Position, norm);
		t /= dot(norm, norm);
		
		midProfile += t * norm;
	}

	float2 midCrossSection = crossSectionCurve;
	if(odd.y)
	{
		float lo = theta - step.x * TWOPI;
		float2 positionLo = CosSin_Evaluate(lo);

		float hi = theta + step.x * TWOPI;
		float2 positionHi = CosSin_Evaluate(hi);

		midCrossSection = (positionHi + positionLo) * 0.5f;
	}
	float3 midPosition;
	midPosition.xy = midProfile.xx * midCrossSection.xy;
	midPosition.z = 0.7 - midProfile.y;

	output.Position = lerp(midPosition, output.Position, min(1.0f, frac(TessFactor.xxx) * 1.0f));
#endif // NEIGHBOR_INTERP

    return output;
}


PARAMETRICSURFACE3D_OUTPUT WireProduct_Evaluate( const float2 input )
{    
    PARAMETRICCURVE2D_OUTPUT wireCurve = BSpline_Evaluate_FuncU( input.x );
    wireCurve.Position *= .005;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate_FuncV( input.y );
    crossSectionCurve.Position *= .0005;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	wireCurve.Position`(input.y) = 0;
	crossSectionCurve.Position`(input.x) = 0;
    
	float2 normVector = wireCurve.Position`(input.x);
	normVector.xy = NORMALIZE_CHECK(-normVector.yx);
	normVector`(input.y) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = wireCurve.Position.x + normVector.x * crossSectionCurve.Position.x;
    output.Position.y =-wireCurve.Position.y + normVector.y * crossSectionCurve.Position.x;
    output.Position.z =-crossSectionCurve.Position.y;
        
    return output;
}

PARAMETRICSURFACE3D_OUTPUT WireProduct_EvaluateInterp( const float2 input )
{   
#ifdef NEIGHBOR_INTERP 
	const bool2 odd = IsOdd(input);
	const float2 step = GetStepToNeighbors(odd);
#endif // NEIGHBOR_INTERP

    PARAMETRICCURVE2D_OUTPUT wireCurve = BSpline_Evaluate_FuncU( input.x );
    wireCurve.Position *= .005;
  
    PARAMETRICCURVE2D_OUTPUT crossSectionCurve = BSpline_Evaluate_FuncV( input.y );
    crossSectionCurve.Position *= .0005;

	/// Explicitly tell the compiler there is no interdependence on parameters.
	wireCurve.Position`(input.y) = 0;
	crossSectionCurve.Position`(input.x) = 0;
    
	float2 normVector = wireCurve.Position`(input.x);
	normVector.xy = NORMALIZE_CHECK(-normVector.yx);
	normVector`(input.y) = 0;
    
    PARAMETRICSURFACE3D_OUTPUT output;
    
    //
    // Compute position
    //    
    output.Position.x = wireCurve.Position.x + normVector.x * crossSectionCurve.Position.x;
    output.Position.y =-wireCurve.Position.y + normVector.y * crossSectionCurve.Position.x;
    output.Position.z =-crossSectionCurve.Position.y;
    
    
#ifdef NEIGHBOR_INTERP
	float2 midWire = wireCurve.Position;
	float2 midNorm = normVector;
	if(odd.x)
	{
		float lo = max(0.0f, input.x - step.x);
		PARAMETRICCURVE2D_OUTPUT wireCurveLo = BSpline_Evaluate_FuncU( lo );
		float2 positionLo = wireCurveLo.Position * 0.005f;
		float2 normLo = wireCurveLo.Position`(lo);
		normLo = NORMALIZE_CHECK(-normLo.yx);

		float hi = min(1.0f, input.x + step.x);
		PARAMETRICCURVE2D_OUTPUT wireCurveHi = BSpline_Evaluate_FuncU( hi );
		float2 positionHi = wireCurveHi.Position * 0.005f;
		float2 normHi = wireCurveHi.Position`(hi);
		normHi = NORMALIZE_CHECK(-normHi.yx);
		midNorm = NORMALIZE_CHECK(normLo + normHi);

		float2 norm = (positionHi - positionLo).yx;
		norm.y = -norm.y;
		float t = dot(positionLo, norm) - dot(wireCurve.Position, norm);
		t /= dot(norm, norm);
		
		midWire += t * norm;
	}

	float2 midCrossSection = crossSectionCurve.Position;
	if(odd.y)
	{
		float lo = max(0.0f, input.y - step.y);
		PARAMETRICCURVE2D_OUTPUT crossSectionCurveLo = BSpline_Evaluate_FuncV( lo );
		float2 positionLo = crossSectionCurveLo.Position * 0.0005f;

		float hi = min(1.0f, input.y + step.y);
		PARAMETRICCURVE2D_OUTPUT crossSectionCurveHi = BSpline_Evaluate_FuncV( hi );
		float2 positionHi = crossSectionCurveHi.Position * 0.0005f;

		float2 norm = (positionHi - positionLo).yx;
		norm.y = -norm.y;
		float t = dot(positionLo, norm) - dot(crossSectionCurve.Position, norm);
		t /= dot(norm, norm);

		midCrossSection += t * norm;
	}
	float3 midPosition;
	midPosition.x = midWire.x + midNorm.x * midCrossSection.x;
	midPosition.y =-midWire.y + midNorm.y * midCrossSection.x;
	midPosition.z = -midCrossSection.y;

	output.Position = lerp(midPosition, output.Position, min(1.0f, frac(TessFactor.xxx) * 1.0f));
#endif // NEIGHBOR_INTERP
    
    return output;
}

//------------------------------------------------------------------------------
//
// Generic shaders that call the above surface functions follow. 
//
//------------------------------------------------------------------------------
struct VSHARDWARE_INPUT
{
	uint Index	: SV_VertexID;
};

struct PS_INPUT_HARDWARE
{
    float4 Position		: SV_Position;        // Screen space
	float3 LocalPos		: POSITION;
#ifdef NORMAL_IN_VS
	float3 LocalNormal	: NORMAL;
#endif // NORMAL_IN_VS
    float2 DomainPos	: TEXCOORD1;        // Parametric space
	float4 LastPos		: TEXCOORD2;
};

//------------------------------------------------------------------------------
//
// The generic "put the point to screen and set up for the pixel shader" vertex shader.
// This is never called directly, but does work common to all vertex shaders.
//
//------------------------------------------------------------------------------
PS_INPUT_HARDWARE VS_Hardware(const PARAMETRICSURFACE3D_OUTPUT surfacePoint, float2 domainPoint)
{
    float4 positionWorld = mul( float4(surfacePoint.Position, 1), LocalToWorld );
    float4 positionView  = mul( positionWorld, WorldToCamera );
    float4 positionScreen= mul( positionView, CameraToNDC );

//    positionScreen = VS_Perturb(positionScreen);

//positionScreen.xy = domainPoint * 0.5f.xx - 0.5f.xx;
//positionScreen.zw = float2(0.5f, 1.0f);
    
    PS_INPUT_HARDWARE output;
    output.Position  = positionScreen;
//output.Position = float4(domainPoint.x * 2.0f - 1.0f, -(domainPoint.y * 2.0f - 1.0f), 0.5f, 1.0f);
    output.DomainPos = domainPoint;
	output.LocalPos = surfacePoint.Position;

#ifdef NORMAL_IN_VS
	float3 dF_du = surfacePoint.Position`(domainPoint.x);
	float3 dF_dv = surfacePoint.Position`(domainPoint.y);
	float3 norm = cross(dF_du, dF_dv);
	output.LocalNormal = NORMALIZE_CHECK(norm);
#endif // NORMAL_IN_VS

//	output.LastPos = mul( float4(surfacePoint.Position, 1), LastLocalToNDC );
	output.LastPos = VS_Reproject(surfacePoint.Position, LastLocalToNDC);
    
    return output;
}

//------------------------------------------------------------------------------
//
// A shader variant for each type of surface.
//
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//
// First vertex shaders

float2 DomainPoint(const VSHARDWARE_INPUT input)
{
	float2 domainPoint;
#if 0
	domainPoint.x = (input.Index % (int)ModDivScale.x) * ModDivScale.z;
	domainPoint.y = (input.Index / (int)ModDivScale.y) * ModDivScale.w;

#else
	const uint iTri = input.Index / 3;
	const uint iQuad = iTri / 2;
	const bool odd = (iTri % 2) != 0;

	const float base_U = iQuad % int(ModDivScale.x);
	const float base_V = iQuad / int(ModDivScale.y);

	const uint iTriVert = input.Index % 3;

	const float tri_U = base_U + 
		( odd
			? (iTriVert + 1) / 2
			: iTriVert % 2 );

	const float tri_V = base_V +
		( odd
			? (iTriVert + 1) % 2
			: iTriVert / 2 );

	domainPoint = float2(tri_U, tri_V) * ModDivScale.zw;
#endif


	return domainPoint;
}

//------------------------------------------------------------------------------
//
// Now pixel shaders

/// Some defines to control what lighting effects to apply.
#define MAF_APPLY_SPECULAR
#define MAF_APPLY_ENV_MAP
//#define MAF_APPLY_FRACTAL_VORONOI
//#define MAF_APPLY_CELL_NOISE

#ifdef MAF_ENABLE_PROC_MATERIALS
#ifdef MAF_APPLY_FRACTAL_VORONOI
#define PROC_EFFECT(a,b,c)	FractalVoronoi( a, b, c)
#elif defined(MAF_APPLY_FRACTAL_VORONOI_SMALL)
#define PROC_EFFECT(a,b,c)	FractalVoronoiSmall( a, b, c)
#elif defined(MAF_APPLY_CELL_NOISE)
#define PROC_EFFECT(a,b,c)	CellNoise( a, b, c)
#elif defined(MAF_APPLY_CELL_NOISE_SMALL)
#define PROC_EFFECT(a,b,c)	CellNoiseSmall( a, b, c)
#elif defined(MAF_APPLY_LUMPY_NOISE)
#define PROC_EFFECT(a,b,c) LumpyNoise( a, b, c )
#elif defined(MAF_APPLY_LUMPY_NOISE_FINE)
#define PROC_EFFECT(a,b,c) LumpyNoiseFine( a, b, c )
#elif defined(MAF_APPLY_NOISE_ISLANDS)
#define PROC_EFFECT(a,b,c) NoiseIslands( a, b, c )
#elif defined(MAF_APPLY_DEFINED_BLOBS)
#define PROC_EFFECT(a,b,c)  DefinedBlobs( a, b, c)
#endif
#endif // MAF_ENABLE_PROC_MATERIALS

float4 PS_Hardware(const PS_INPUT_HARDWARE input, const PARAMETRICSURFACE3D_OUTPUT surfacePoint) 
{
//return float4(input.LocalPos, 1.0f);
#ifdef NORMAL_IN_VS
	float3 localNorm = NORMALIZE_CHECK(input.LocalNormal);
    float3 norm = NORMALIZE_CHECK(mul( localNorm, (float3x3)LocalToWorld ));
#else // NORMAL_IN_VS
	float3 dF_du = surfacePoint.Position`(input.DomainPos.x);
	float3 dF_dv = surfacePoint.Position`(input.DomainPos.y);
	float3 localNorm = NORMALIZE_CHECK(cross(dF_du, dF_dv));
    float3 norm = NORMALIZE_CHECK(mul( localNorm, (float3x3)LocalToWorld ));
#endif // NORMAL_IN_VS

	const float3 localPos = input.LocalPos;
//	const float3 localPos = surfacePoint.Position.xyz;
	const float3 worldPosition = mul(float4(localPos, 1.0f), LocalToWorld).xyz;

    float3 eyeVector = NORMALIZE_CHECK( WorldCamPos.xyz - worldPosition );

    const float attenDetail = 1.0f;
    const float attenSpec = 1.0f;

	float detail = attenDetail;
#ifdef PROC_EFFECT
	float3 dirt = float3( 1.0f, 1.0f, 1.0f );
	float3 localTan = MakeTangent( localNorm );
	float3 localBin = cross( localTan, localNorm );

	const float aafactor = UnitLocalToScreen( localPos, localTan, localBin );
//return float4(aafactor.xxx * DetailParams.x, 1.0f);
	detail *= saturate( aafactor * DetailParams.x );

	float3 bumpedNorm = PROC_EFFECT( localPos, localNorm, localTan );

	bumpedNorm = NORMALIZE_CHECK( mul( bumpedNorm, (float3x3)LocalToWorld ) );

//#define MAF_PRESERVE_DIRT
#ifndef MAF_PRESERVE_DIRT
	bumpedNorm = NORMALIZE_CHECK( lerp( norm, bumpedNorm, detail ) );
#endif // MAF_PRESERVE_DIRT
	
	float bumpDotNorm = dot( bumpedNorm, norm );
	dirt *= float3( bumpDotNorm, bumpDotNorm, bumpDotNorm * bumpDotNorm );

#ifdef MAF_PRESERVE_DIRT
	bumpedNorm = NORMALIZE_CHECK( lerp( norm, bumpedNorm, detail ) );
#endif // MAF_PRESERVE_DIRT
	norm = bumpedNorm;
#endif // PROC_EFFECT

//return float4(norm.xyz, 1.0f);
//return  float4(localNorm.xyz, 1.0f);
    float fDiffuse = saturate(.5f + .5f * dot(norm, LightVector.xyz ) ) * Light.x;
//return float4(fDiffuse.xxx, 1.0f);

    float fTotalIntensity = AmbientFactor
                          + DiffuseFactor  * fDiffuse;
    
	float4 materialColor = MaterialColor;
	float reflectivity = Reflectivity;

#ifdef MAF_APPLY_SPECULAR
    float3 halfwayVector = NORMALIZE_CHECK( eyeVector + LightVector.xyz );
    

	float fSpecular = pow(saturate(dot(halfwayVector, norm)), SpecularPower) * Light.y;

#endif // MAF_APPLY_SPECULAR
    
#ifdef MAF_APPLY_PIANO_WOOD

	const float3 localTan = MakeTangent(localNorm);
	const float3 localBin = (cross(localTan, localNorm));
	const float aafactor = UnitLocalToScreen(localPos, localTan, localBin);
	const float fadeIn = saturate(aafactor * DetailParams.x * 0.4f);

//	fSpecular *= pow(fadeIn, DetailParams.y);

	const float4 specularComponent = saturate(float4((SpecularFactor*fSpecular).xxx, 0.0f)) * attenSpec.xxxx;

	float vert = frac(localPos.z * 30.0f);

	vert += (cos(localPos.x * 10.0f) + cos(localPos.y * 15.0f)) * cos(localPos.z * 20.0f) * 0.1f;

	float vertFreq = vert > 0.5f ? 1.0f : 0.5f;

	float hori = frac(localPos.x * 15.0f);
	hori += cos(localPos.y * 10.0f) * 0.1f;

	float horiFreq = hori > 0.5f ? 1.0f : 0.5f;

	// float combo = abs(localNorm.z);
	float combo = pow(abs(localNorm.z), 5.0f);
	float comboFreq = lerp(vertFreq, horiFreq, combo);
	comboFreq = lerp(0.75f, comboFreq, fadeIn);

    float4 outputColor = (MaterialColor * comboFreq * fTotalIntensity +  specularComponent);

#else // MAF_APPLY_PIANO_WOOD

#ifdef MAF_APPLY_LAMP_SHADE
	if(input.DomainPos.x > 0.5f)
	{
		materialColor = float4(0.9f, 0.9f, 0.85f, 1.0f);
		fSpecular = 0;
		fTotalIntensity = 0.5f + saturate(dot(eyeVector, norm)) * 0.5f;
		reflectivity = 0.0f;
	}
#endif // MAF_LAMP_SHADE

	fTotalIntensity += saturate(SpecularFactor * fSpecular) * attenSpec;

    float4 outputColor = materialColor * fTotalIntensity;

#endif // MAF_APPLY_PIANO_WOOD
	
#ifdef MAF_APPLY_ENV_MAP
	eyeVector.xyz = reflect(-eyeVector.xyz, norm);
	float4 env = cubeMap.Sample( cubeSampler, eyeVector.xyz );

	outputColor = lerp(outputColor, env, reflectivity);
#endif // MAF_APPLY_ENV_MAP

#ifdef PROC_EFFECT
	outputColor.xyz *= dirt;
#endif // PROC_EFFECT

#if 0
	const float2 depthCoord = input.Position.xy;
	const float prevDepth = previousDepth.Load(int2(depthCoord.x, depthCoord.y), 0);
//	outputColor.xyz = lerp(outputColor.xyz, prevDepth.xxx, 0.5f);
	outputColor.xyz = prevDepth.xxx;
#endif
#if 0
	float2 colorCoord = input.Position.xy / 256.0f.xx;
	float4 prevColor = previousColor.Sample(cubeSampler, colorCoord);
	outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, 0.95f);
#endif
#if 0
	float4 lastPos = input.LastPos / input.LastPos.w;
	float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;
	float4 prevColor = previousColor.Sample(cubeSampler, colorCoord);
	outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, 0.8f);
#endif
#if 0
	float4 lastPos = input.LastPos / input.LastPos.w;
	float2 depthCoord = (lastPos.xy * float2(0.5f, -0.5f) + 0.5f) * WindowSize.xy;
	const float prevDepth = previousDepth.Load(int2(depthCoord.x, depthCoord.y), 0);
//	outputColor.xyz = prevDepth.xxx;
//	outputColor.xyz = input.Position.zzz;
	float diff = (prevDepth.x - input.Position.z) / max(prevDepth, input.Position.z);
	outputColor.xyz = diff > WindowSize.z
		? float3(1.0f, 0.0f, 0.0f) 
		: (diff < -WindowSize.z 
			? float3(0.0f, 1.0f, 0.0f)
			: float3(0.0f, 0.0f, 1.0f));
#endif
#if 0
	const float4 lastPos = input.LastPos / input.LastPos.w;
	if(lastPos.x * lastPos.x < 1.0f && lastPos.y * lastPos.y < 1.0f)
	{
		const float2 fDepthCoord = (lastPos.xy * float2(0.5f, -0.5f) + 0.5f) * WindowSize.xy - 0.5f.xx;
		const int2 iDepthCoord = int2(fDepthCoord.x, fDepthCoord.y);
		//if(iDepthCoord.x >= 0 && iDepthCoord.x < WindowSize.x && iDepthCoord.y >= 0 && iDepthCoord.y < WindowSize.y)
		{
			const float prevDepth = previousDepth.Load(iDepthCoord, 0);
			const float diff = abs(prevDepth.x - input.Position.z) / max(prevDepth, input.Position.z);
			if(diff < WindowSize.z)
			{
					const float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;
					const float4 prevColor = previousColor.Sample(cubeSampler, colorCoord);
					outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, WindowSize.w);
			}
			else if(WindowSize.w <= 0)
			{
				outputColor.xyz = float3(1.0f, 0.0f, 0.0f);
			}
		}
		//else
		//{
			//outputColor.xyz = float3(0.0f, 0.0f, 1.0f);
		//}
	}
#endif
#if 0
	const float4 lastPos = input.LastPos / input.LastPos.w;
	if(lastPos.x * lastPos.x < 1.0f && lastPos.y * lastPos.y < 1.0f)
	{
		const float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;
		const float4 prevColor = previousColor.Sample(cubeSampler, colorCoord);
		outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, WindowSize.w);
	}
#endif
#if 1
	outputColor = PS_Reproject(input.Position, input.LastPos, outputColor);
#endif
    
    return outputColor;
}

float4 PS_Hardware_ProfileProduct(const PS_INPUT_HARDWARE input) 
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = ProfileProduct_Evaluate( input.DomainPos );

    return PS_Hardware(input, surfacePoint);
}

float4 PS_Hardware_SurfaceOfRevolution(const PS_INPUT_HARDWARE input) 
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = SurfaceOfRevolution_Evaluate( input.DomainPos );

    return PS_Hardware(input, surfacePoint);
}

float4 PS_Hardware_WireProduct(const PS_INPUT_HARDWARE input) 
{
    PARAMETRICSURFACE3D_OUTPUT surfacePoint = WireProduct_Evaluate( input.DomainPos );

    return PS_Hardware(input, surfacePoint);
}



#endif // PARAMETRIC_3D_SURFACE_FX
