
#ifndef PROC_SURFACE_FX
#define PROC_SURFACE_FX

#include "SVD.fx"

#include "ProceduralHelpers.fx"
#include "NormalizeCheck.fx"

float3 MakeTangent(const float3 localNorm)
{
	float3 localTan = float3(localNorm.z, 0.0f, -localNorm.x);
	if(dot(localTan, localTan) < 1.0e-2f)
		localTan = float3(0.0f, localNorm.z, -localNorm.y);

	return NORMALIZE_CHECK(localTan);
}

float2x3 ProjectLocalUV(const float3 localPos)
{
	float4x4 localToNDC = mul(mul(LocalToWorld, WorldToCamera), CameraToNDC);
	float4 PSc = mul(float4(localPos, 1.0f), localToNDC);
	PSc.xy *= DetailParams.zw;
	PSc.xy /= PSc.w;
	float3 dPSc_x_dLP = float3(PSc.x`(localPos.x), PSc.x`(localPos.y), PSc.x`(localPos.z));
	float3 dPSc_y_dLP = float3(PSc.y`(localPos.x), PSc.y`(localPos.y), PSc.y`(localPos.z));
	
	return float2x3(dPSc_x_dLP, dPSc_y_dLP);
}

float UnitLocalToScreen(const float3 localPos, const float3 localTan, const float3 localBin)
{
	float2x3 proj = ProjectLocalUV(localPos);

	float2x2 mat;
	mat[0] = mul(proj, localBin);
	mat[1] = mul(proj, localTan);

//return min(length(mat[0]), length(mat[1]));

#if 0
	float len = min(length(mat[0]), length(mat[1]));
#else
	float2 sigma;
	Jacobi2x2SVD(mat, sigma);
	float len = sigma.y;
#endif

	return len;
}

//--------------------------------------------------------------------------------------
// Adding octaves of cellular noise
//--------------------------------------------------------------------------------------
float3 FractalVoronoi( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    // Scale the position to make the noise look smaller, and shift it away from the origin
//    float3 NoiseIn = localPos * 10 + float3( 100, 100, 100 );
	float3 NoiseIn = localPos * 10 + float3( 100, 100, 100 );
	
	/// Get the noise signal and its derivatives. 
    float VoronoiNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ) , 4, 2, 0.5, NT_VORONOI3, localPos);
    
    // Shift and clamp the noise to make it look like like a bunch of small cracks
	VoronoiNoise += 0.3f;
	VoronoiNoise = min(VoronoiNoise, 1.0f);
    VoronoiNoise = pow( VoronoiNoise, 8.0f );

    float BumpAmount = 2.2f;
    float3 BumpedPos = localPos + localNorm * VoronoiNoise * BumpAmount;
        
	/// Use the symbolic derivative computed in and returned from fBm() to
	/// generate a local normal. 
	float3 binormal = NORMALIZE_CHECK(cross(localTan, localNorm));
	float3 tangent = NORMALIZE_CHECK(cross(localNorm, binormal));

	float3 deriv = EXTRACT_DERIV3(VoronoiNoise, localPos).xyz * 3.35f;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm 
		- tangent * dot(tangent, deriv.xyz)
		- binormal * dot(binormal, deriv.xyz));

	return bumpNorm;
}

//--------------------------------------------------------------------------------------
// Adding octaves of cellular noise
//--------------------------------------------------------------------------------------
float3 FractalVoronoiSmall( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    // Scale the position to make the noise look smaller, and shift it away from the origin
//    float3 NoiseIn = localPos * 10 + float3( 100, 100, 100 );
	float3 NoiseIn = localPos * 50 + float3( 100, 100, 100 );
	
	/// Get the noise signal and its derivatives. 
    float VoronoiNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ) , 4, 2, 0.5, NT_VORONOI3, localPos);
    
    // Shift and clamp the noise to make it look like like a bunch of small cracks
	VoronoiNoise += 0.3f;
	VoronoiNoise = min(VoronoiNoise, 1.0f);
    VoronoiNoise = pow( VoronoiNoise, 8.0f );

    float BumpAmount = 2.2f;
    float3 BumpedPos = localPos + localNorm * VoronoiNoise * BumpAmount;
        
	/// Use the symbolic derivative computed in and returned from fBm() to
	/// generate a local normal. 
	float3 binormal = NORMALIZE_CHECK(cross(localTan, localNorm));
	float3 tangent = NORMALIZE_CHECK(cross(localNorm, binormal));

	float3 deriv = EXTRACT_DERIV3(VoronoiNoise, localPos).xyz * 3.35f;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm 
		- tangent * dot(tangent, deriv.xyz)
		- binormal * dot(binormal, deriv.xyz));

	return bumpNorm;
}

//--------------------------------------------------------------------------------------
// Cellular Noise uses voronoi noise to generate nice cellular or lumpy textures
//--------------------------------------------------------------------------------------
float3 CellNoise( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    // Scale the position to make the noise look smaller, and shift it away from the origin
    float3 NoiseIn = localPos * 10 + float3( 100, 100, 100 );
	/// Get the noise signal and its derivatives. 
    float VNoise = VoronoiNoise( NoiseIn, WorldCamPos.w * 1.1, NT_VORONOI3_2, localPos );
    
    float BumpAmount = 0.02;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;
        
	/// Use the symbolic derivatives computed in and returned from VoronoiNoise() to
	/// generate a local normal. 
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz * -0.42f;		

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}

//--------------------------------------------------------------------------------------
// Cellular Noise uses voronoi noise to generate nice cellular or lumpy textures
//--------------------------------------------------------------------------------------
float3 CellNoiseSmall( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    // Scale the position to make the noise look smaller, and shift it away from the origin
    float3 NoiseIn = localPos * 50 + float3( 100, 100, 100 );
	/// Get the noise signal and its derivatives. 
    float VNoise = VoronoiNoise( NoiseIn, WorldCamPos.w * 1.1, NT_VORONOI3_2, localPos );
    
    float BumpAmount = 0.02;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;
        
	/// Use the symbolic derivatives computed in and returned from VoronoiNoise() to
	/// generate a local normal. 
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz * -0.42f;		

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}

float3 LumpyNoise( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    float3 NoiseIn = localPos * 0.1 + float3( 100, 100, 100 );
    float VNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ), 
		8, // numOctaves
		2, // lacunarity
		0.5, // gain
		NT_SIMPLEX3D,	// Noise type
		localPos); 

    float BumpAmount = 0.01;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;
        
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz * BumpAmount;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}

float3 LumpyNoiseFine( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    float3 NoiseIn = localPos * 10.0 + float3( 100, 100, 100 );
    float VNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ), 
		8, // numOctaves
		1.25, // lacunarity
		0.75, // gain
		NT_SIMPLEX3D,	// Noise type
		localPos); 

    float BumpAmount = 0.01;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;
        
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz * BumpAmount;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}

float3 NoiseIslands( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    float3 NoiseIn = localPos * 10.0 + float3( 100, 100, 100 );
    float VNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ), 
		8, // numOctaves
		1.25, // lacunarity
		0.75, // gain
		NT_SIMPLEX3D,	// Noise type
		localPos); 

    float BumpAmount = 0.01;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;

	VNoise *= BumpAmount;
	VNoise = saturate(VNoise);
        
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}

float3 DefinedBlobs( const float3 localPos, const float3 localNorm, const float3 localTan)
{
    float3 NoiseIn = localPos * 5.0 + float3( 100, 100, 100 );
    float VNoise = fBm( float4( NoiseIn, WorldCamPos.w * 1.1 ), 
		8, // numOctaves
		1.25, // lacunarity
		0.75, // gain
		NT_SIMPLEX3D,	// Noise type
		localPos); 

    float BumpAmount = 1.0e-4f;
    float3 BumpedPos = localPos + localNorm * VNoise * BumpAmount;

	VNoise *= BumpAmount;
//	VNoise = saturate(abs(VNoise));
	VNoise = saturate(pow(abs(VNoise), 0.6f));
        
	float3 binormal = cross(localTan, localNorm);

	float3 deriv = EXTRACT_DERIV3(VNoise, localPos).xyz;

	float3 bumpNorm = NORMALIZE_CHECK(localNorm
		- localTan * dot(localTan, deriv)
		- binormal * dot(binormal, deriv));

	return bumpNorm;
}


#endif // PROC_SURFACE_FX