


//--------------------------------------------------------------------------------------
// Constant Buffer Variables
//--------------------------------------------------------------------------------------
cbuffer cbChangesEveryFrame : register( b0 )
{
    matrix CameraToNDC;

    matrix LocalToWorld;

	float4 ModDivScale;

    matrix WorldToCamera;
	float4 WorldCamPos;
};


//--------------------------------------------------------------------------------------
// Shader input and output definitions
//--------------------------------------------------------------------------------------
struct VS_INPUT
{
	uint Index	: SV_VertexID;
};

struct PS_INPUT
{
    float4 Position		: SV_Position;      // Screen space
	float3 LocalPos		: POSITION;			// Object space
    float2 DomainPos	: TEXCOORD1;        // Parametric space
};

float2 DomainPoint( uint vertexId )
{
	const uint iTri = vertexId / 3;
	const uint iQuad = iTri / 2;
	const bool odd = (iTri % 2) != 0;

	const float base_U = iQuad % int(ModDivScale.x);
	const float base_V = iQuad / int(ModDivScale.y);

	const uint iTriVert = vertexId % 3;

	const float tri_U = base_U + 
		( odd
			? (iTriVert + 1) / 2
			: iTriVert % 2 );

	const float tri_V = base_V +
		( odd
			? (iTriVert + 1) % 2
			: iTriVert / 2 );

	return float2(tri_U, tri_V) * ModDivScale.zw;
}

float3 SphereWarp(const float2 domainPoint, const float3 localPos)
{
	static const float TWOPI = 6.283185307;
	static const float PI = 3.141592654;
	static const float OFFSET = 0.05f;

	float offsetU = sin(domainPoint.y * PI * 10 + WorldCamPos.w) * OFFSET;
	float offsetV = sin(domainPoint.x * TWOPI * 3 + WorldCamPos.w) * OFFSET;

	float scaleU = sin(domainPoint.y * TWOPI);
	scaleU *= scaleU;
	return localPos * (1.0f + (offsetU + offsetV) * scaleU);
}

float3 SpherePosition(const float2 domainPoint)
{
	static const float TWOPI = 6.283185307;
	static const float    PI = 3.141592654;

	float sinPhi = sin(domainPoint.y * PI);
	float cosPhi = cos(domainPoint.y * PI);

	float sinTheta = sin(domainPoint.x * TWOPI);
	float cosTheta = cos(domainPoint.x * TWOPI);

	float3 localPos = float3(cosTheta * sinPhi,
							sinTheta * sinPhi,
							cosPhi);

//#define MAF_WARP_SPHERE
#ifdef MAF_WARP_SPHERE
	localPos = SphereWarp(domainPoint, localPos);
#else // MAF_WARP_SPHERE
	localPos *= WorldCamPos.w;
#endif // MAF_WARP_SPHERE

	return localPos;
}

PS_INPUT VS_Main( const VS_INPUT input )
{
	const float2 domainPoint = DomainPoint(input.Index);

	PS_INPUT output;
	output.LocalPos = SpherePosition(domainPoint);

	float4 worldPos = mul(float4(output.LocalPos, 1.0f), LocalToWorld);
	float4 cameraPos = mul(worldPos, WorldToCamera);
	output.Position = mul(cameraPos, CameraToNDC);

//output.Position.xy = domainPoint.xy * 0.5f.xx - 0.5f.xx;
//output.Position.zw = float2(0.5f, 1.0f);
	output.DomainPos = domainPoint;

	return output;
}

float4 Illuminate(const float3 localPosition, const float3 localNormal)
{
	const float AmbientFactor = 0.1f;
	const float DiffuseFactor = 0.7f;
	const float SpecularFactor = 4.31;
	const float SpecularPower = 64.0f;
	const float4 MaterialBrite = float4(0.9921569, 0.7450981, 0, 1.0);
	const float4 MaterialDark = float4(0.1, 0.2, 0.1, 1.0);
	const float3 LightVector = normalize(float3(.8, .7, 1));

	float3 worldPosition = mul(float4(localPosition, 1.0f), LocalToWorld).xyz;
	float3 worldNormal = normalize(mul(localNormal, (float3x3)LocalToWorld));

    float fDiffuse = saturate(dot(worldNormal, LightVector.xyz ));

    float fTotalIntensity = AmbientFactor
                          + DiffuseFactor  * fDiffuse;
    
    float3 eyeVector = normalize(WorldCamPos.xyz - worldPosition);
    float3 halfwayVector = normalize(eyeVector + LightVector.xyz);
    
    float4 outputColor = lerp(MaterialDark, MaterialBrite, fTotalIntensity.xxxx);

    float fSpecular = pow(saturate(dot(halfwayVector, worldNormal)), SpecularPower);

	outputColor.xyz += (SpecularFactor * fSpecular).xxx;

	return outputColor;
}

float4 PS_Main( const PS_INPUT input ) : SV_Target
{
	float3 surfacePosition = SpherePosition(input.DomainPos);

#if 1
	float3 tangent = surfacePosition`(input.DomainPos.x);
	float3 binormal = surfacePosition`(input.DomainPos.y);
	float3 localNormal = normalize(cross(binormal, tangent));
#else
	float3 localNormal = normalize(surfacePosition);
#endif


	float4 outputColor = Illuminate(surfacePosition, localNormal);

	return outputColor;
}
