
#include <windows.h>
#include <d3dx11.h>
#include <D3DX10math.h>

#include "Bounds.h"
#include "CullVolume.h"


Bounds::CullStatus SphereBounds::CullTest(const CullVolume& culler) const
{
    const float dist = culler.CheckDist(Position);
    if(dist + Radius < 0)
        return kAllOut;
    
    if(dist - Radius >= 0)
        return kAllIn;
    
    return kSplit;
}

static float Scale(const D3DXMATRIX& xfm)
{
    const D3DXVECTOR3 row0(xfm.m[0][0], xfm.m[0][1], xfm.m[0][2]);
    const D3DXVECTOR3 row1(xfm.m[1][0], xfm.m[1][1], xfm.m[1][2]);
    const D3DXVECTOR3 row2(xfm.m[2][0], xfm.m[2][1], xfm.m[2][2]);
    const float lenSq0 = D3DXVec3LengthSq(&row0);
    const float lenSq1 = D3DXVec3LengthSq(&row1);
    const float lenSq2 = D3DXVec3LengthSq(&row2);
    const float maxLenSq = max(lenSq0, max(lenSq1, lenSq2));

    return sqrt(maxLenSq);
}
Bounds& SphereBounds::Transform(const D3DXMATRIX& l2w)
{
    /// Assumes a pure scale/rotate/translation transform
    const D3DXVECTOR3 pos = Position;
    const float radius = Radius;
    D3DXVec3TransformCoord(&Position, &pos, &l2w);
    Radius = radius * Scale(l2w);

    return *this;
}

Bounds::CullStatus AABB::CullTest(const CullVolume& culler) const
{
    return OBB(*this).CullTest(culler);
}

Bounds& AABB::Transform(const D3DXMATRIX& localToWorld)
{
    OBB obb(*this);
    obb.Transform(localToWorld);
    lo = obb.Min();
    hi = obb.Max();
    return *this;
}

OBB::OBB(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi)
{
    Init(lo, hi);
}

OBB::OBB(const AABB& aabb)
{
    Init(aabb.Min(), aabb.Max());
}

void OBB::Init(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi)
{
    D3DXVECTOR3* const pts = Corners;
    int iPt = 0;
    for(int k = 0; k < 2; ++k)
    {
        for(int j = 0; j < 2; ++j)
        {
            for(int i = 0; i < 2; ++i)
            {
                pts[iPt].x = i & 1 ? lo.x : hi.x;
                pts[iPt].y = j & 1 ? lo.y : hi.y;
                pts[iPt].z = k & 1 ? lo.z : hi.z;
                ++iPt;
            }
        }
    }
}

Bounds::CullStatus OBB::CullTest(const CullVolume& culler) const
{
    D3DXVECTOR2 dists = culler.CheckDist(Corners, 8);

    if(dists.y < 0)
        return kAllOut;

    if(dists.x < 0)
        return kSplit;

    return kAllIn;
}

Bounds& OBB::Transform(const D3DXMATRIX& localToWorld)
{
    for(int i = 0; i < 8; ++i)
    {
        D3DXVec3TransformCoord(Corners + i, Corners + i, &localToWorld);
    }
    return *this;
}

D3DXVECTOR3 OBB::Min() const
{
    D3DXVECTOR3 lo(FLT_MAX, FLT_MAX, FLT_MAX);
    for(int i = 0; i < 8; ++i)
    {
        lo.x = min(lo.x, Corners[i].x);
        lo.y = min(lo.y, Corners[i].y);
        lo.z = min(lo.z, Corners[i].z);
    }
    return lo;
}
D3DXVECTOR3 OBB::Max() const
{
    D3DXVECTOR3 hi(-FLT_MAX, -FLT_MAX, -FLT_MAX);
    for(int i = 0; i < 8; ++i)
    {
        hi.x = max(hi.x, Corners[i].x);
        hi.y = max(hi.y, Corners[i].y);
        hi.z = max(hi.z, Corners[i].z);
    }
    return hi;
}
