#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"
#include "DXUTcamera.h"
#include "DXUTres.h"
#include "CFlyCamera.h"
#undef min // use __min instead
#undef max // use __max instead

#include "CameraModeSelect.h"
#include "CameraMode.h"

//--------------------------------------------------------------------------------------
// Constructor 
//--------------------------------------------------------------------------------------
CFlyCamera::CFlyCamera()
{
    m_eCameraMode = CameraModeSelect::kRandom;
    m_pCameraMode = NULL; // will be allocated in ResetGoals();

    m_bEnablePositionMovement = false;

    ZeroMemory(&lastWorldToNDC, sizeof(lastWorldToNDC));

	ResetGoals();
}

CFlyCamera::~CFlyCamera()
{
    SAFE_DELETE(m_pCameraMode);
}

void CFlyCamera::SetSpeedGoal(float f)
{
    m_pCameraMode->SetSpeedGoal(f);
}
float CFlyCamera::GetSpeedGoal() const
{
    return m_pCameraMode->GetSpeedGoal();
}

VOID CFlyCamera::ResetGoals()
{
	D3DXVECTOR2 mousePos = D3DXVECTOR2(0.0f, 0.0f);
	D3DXVECTOR2 lastPos = D3DXVECTOR2(0.0f, 0.0f);

    ResetMode();
}

void CFlyCamera::SetFlyMode(CameraModeSelect::eCameraModes mode) 
{ 
    if(m_eCameraMode != mode)
    {
        m_eCameraMode = mode; 
        ResetMode();
    }
}

VOID CFlyCamera::ResetMode()
{
    float speed = m_pCameraMode ? m_pCameraMode->GetSpeedGoal() : 20.0f / 32.0f;
    SAFE_DELETE(m_pCameraMode);
    m_pCameraMode = CameraModeSelect::Create(GetFlyMode());
    m_pCameraMode->Init(*this);
    m_pCameraMode->SetSpeedGoal(speed);
}

//--------------------------------------------------------------------------------------
// Update the view matrix & the model's world matrix based 
//       on user input & elapsed time
//--------------------------------------------------------------------------------------
VOID CFlyCamera::FrameMove( FLOAT fElapsedTime )
{
    if( IsKeyDown( m_aKeys[CAM_RESET] ) )
	{
        Reset();
	}

    const float kMaxFrameDelta = 0.1f;
    fElapsedTime = __min(fElapsedTime, kMaxFrameDelta);
    m_pCameraMode->Advance(*this, fElapsedTime);
    D3DXVECTOR3 eye = m_pCameraMode->GetEyePt();
    D3DXVECTOR3 lookAt = m_pCameraMode->GetLookAtPt();

    SetViewParams(&eye, &lookAt);

}

VOID CFlyCamera::SetMatrix()
{
	D3DXVECTOR3 direction = m_vLookAt - m_vEye;

	D3DXVECTOR3 up = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	D3DXVECTOR3 bin;
	D3DXVec3Cross(&bin, &direction, &up);
	D3DXVec3Normalize(&bin, &bin);
	D3DXVec3Cross(&up, &bin, &direction);
	D3DXVec3Normalize(&up, &up);
	
	D3DXVECTOR3 lookDown = m_vLookAt;
	// look down from the direction of travel.
//	lookDown.y -= 2.0f;

    // Update the view matrix
    D3DXMatrixLookAtLH( &m_mView, &m_vEye, &lookDown, &up);
}


VOID CFlyCamera::SetDragRect( RECT& rc )
{
    CBaseCamera::SetDragRect( rc );

    SetWindow( rc.right - rc.left, rc.bottom - rc.top );
}


//--------------------------------------------------------------------------------------
// Reset the camera's position back to the default
//--------------------------------------------------------------------------------------
VOID CFlyCamera::Reset()
{
    CBaseCamera::Reset();

	ResetGoals();
}


//--------------------------------------------------------------------------------------
// Override for setting the view parameters
//--------------------------------------------------------------------------------------
VOID CFlyCamera::SetViewParams( D3DXVECTOR3* pvEyePt, D3DXVECTOR3* pvLookatPt )
{
    CBaseCamera::SetViewParams( pvEyePt, pvLookatPt );

    SetMatrix();
}

VOID CFlyCamera::SetOffProjParams( FLOAT left, FLOAT right, FLOAT bot, FLOAT top, FLOAT fNearPlane, FLOAT fFarPlane )
{
	m_fFOV = 0;
	m_fAspect = (bot - top) / (right - left);
    m_fNearPlane = fNearPlane;
    m_fFarPlane = fFarPlane;

   D3DXMatrixPerspectiveOffCenterLH( &m_mProj, left, right, bot, top, fNearPlane, fFarPlane );
}
VOID CFlyCamera::SetProjection(const D3DXMATRIX* proj)
{
    m_mProj = *proj;
}


//--------------------------------------------------------------------------------------
// Call this from your message proc so this class can handle window messages
//--------------------------------------------------------------------------------------
LRESULT CFlyCamera::HandleMessages( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    CBaseCamera::HandleMessages( hWnd, uMsg, wParam, lParam );

    if( uMsg == WM_LBUTTONDOWN )
    {
		lastPos = mousePos;
    }

    if( uMsg == WM_MOUSEMOVE )
    {
        int iMouseX = ( short )LOWORD( lParam );
        int iMouseY = ( short )HIWORD( lParam );

		mousePos.x = FLOAT(iMouseX) / FLOAT(m_nWidth) * 2.0f - 1.0f;
		mousePos.y = FLOAT(iMouseY) / FLOAT(m_nHeight) * 2.0f - 1.0f;
    }

    return FALSE;
}

bool CFlyCamera::ToggleBuildLog(const WCHAR* const fileName)
{
    bool retVal = false;
    if(m_pCameraMode->GetLogMode() == CameraMode::kLogBuild)
    {
        retVal = m_pCameraMode->EndBuildLog(fileName);
    }
    else
    {
        retVal = m_pCameraMode->BeginBuildLog();
    }
    return retVal;
}
bool CFlyCamera::ToggleRunFromLog(const WCHAR* const fileName)
{
    bool retVal = false;
    if(m_pCameraMode->GetLogMode() != CameraMode::kLogFrom)
    {
        retVal = m_pCameraMode->BeginFromLog(fileName);
    }
    else
    {
        retVal = m_pCameraMode->EndFromLog();
    }
    return retVal;
}
