
#define _USE_MATH_DEFINES
#include <math.h>
#include <vector>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "CameraDrive.h"
#include "CFlyCamera.h"
#include "../Objects/Scape.h"

static float kMaxSpeed = 10.0f;

CameraDrive::CameraDrive()
    :   m_vEyePt(0.0f, 0.0f, 0.0f),
        m_vLookAtPt(1.0f, 0.0f, 0.0f),
        m_vUpDir(0.0f, 1.0f, 0.0f),
        m_LogMode(kLogNone)
{
    ResetGoals();
}

VOID CameraDrive::SetSpeedGoal(float f)
{
    speedGoal = f;
}
float CameraDrive::GetSpeedGoal() const
{ 
    return speedGoal; 
}
VOID CameraDrive::ResetGoals()
{
//	speedGoal = kMaxSpeed * 2.0f / (1 << 5);
    speedGoal = 0.0f;
	speedCurr = 0.0f;
	speedAcc = 10.0f;

	yawGoal = 0.0f;
	yawCurr = 0.0f;
	yawAcc = 5.5f;

	pitchGoal = 0.0f;
	pitchCurr = 0.0f;
	pitchAcc = 5.5f;

}

void CameraDrive::Reset()
{
    ResetGoals();
}

void CameraDrive::Init(const CFlyCamera& camera)
{
    m_vEyePt = *camera.GetEyePt();
    m_vLookAtPt = *camera.GetLookAtPt();
}

static FLOAT AdvanceTo(FLOAT goal, FLOAT curr, FLOAT acc, FLOAT dt)
{
	if(goal > curr)
	{
		if((curr += acc * dt) > goal)
			curr = goal;
	}
	else if(goal < curr)
	{
		if((curr -= acc * dt) < goal)
			curr = goal;
	}
	return curr;
}

VOID CameraDrive::UpdateGoals(const D3DXVECTOR2& mousePos, const D3DXVECTOR2& lastPos, const FLOAT fElapsedTime)
{
	static FLOAT speedInc = 10.0f;
    if(!DXUTIsKeyDown(VK_CONTROL) && !DXUTIsKeyDown(VK_SHIFT))
    {
	    if(DXUTIsKeyDown('1'))
		    speedGoal += -speedInc * fElapsedTime;
	    if(DXUTIsKeyDown('2'))
		    speedGoal += speedInc * fElapsedTime;
	    if(DXUTIsKeyDown('3') || DXUTIsKeyDown('0'))
		    speedGoal = 0.0f;
    }

    const bool inDrag = DXUTIsMouseButtonDown(VK_LBUTTON);
	if(inDrag)
	{
		static FLOAT pitchTrigger = 0.01f;
		FLOAT delPitch = mousePos.y - lastPos.y;
		if(delPitch > pitchTrigger)
		{
			pitchGoal = -(delPitch - pitchTrigger) / (1.0f - pitchTrigger);
		}
		else if(delPitch < -pitchTrigger)
		{
			pitchGoal = -(delPitch + pitchTrigger) / (1.0f - pitchTrigger);
		}
		else
		{
			pitchGoal = 0.0f;
		}

		static FLOAT yawTrigger = 0.01f;
		FLOAT delYaw = mousePos.x - lastPos.x;
		if(delYaw > yawTrigger)
		{
			yawGoal = -(delYaw - yawTrigger) / (1.0f - yawTrigger);
		}
		else if(delYaw < -yawTrigger)
		{
			yawGoal = -(delYaw + yawTrigger) / (1.0f - yawTrigger);
		}
		else
		{
			yawGoal = 0.0f;
		}
	}
	else
	{
		pitchGoal = yawGoal = 0.0f;
	}

    if(!DXUTIsKeyDown(VK_CONTROL))
    {
        const int kMinSpeedKey = '5';
        const int kMaxSpeedFKey = VK_F10;
        const int kMinSpeedFKey = VK_F5;
        const int numSpeeds = kMaxSpeedFKey - kMinSpeedFKey + 1;
        for(int i = 0; i < numSpeeds; ++i)
        {
            if(DXUTIsKeyDown(BYTE(kMinSpeedKey + i)) ) // mafinch - disable function key speedup || DXUTIsKeyDown(BYTE(kMinSpeedFKey + i)))
            {
                speedGoal = float(1 << i) * kMaxSpeed * 2.0f / float(1 << (numSpeeds-1));
            }
        }
    }
}

VOID CameraDrive::AdvanceToGoals(const FLOAT fElapsedTime)
{
    speedCurr = AdvanceTo(speedGoal, speedCurr, speedAcc, fElapsedTime);
    pitchCurr = AdvanceTo(pitchGoal, pitchCurr, pitchAcc, fElapsedTime);
    yawCurr = AdvanceTo(yawGoal, yawCurr, yawAcc, fElapsedTime);

    if(speedGoal > kMaxSpeed)
        speedGoal = kMaxSpeed;
    if(speedGoal < -kMaxSpeed * 0.5f)
        speedGoal = -kMaxSpeed * 0.5f;
}

D3DXVECTOR3 CameraDrive::UpdateManuals(const D3DXVECTOR3& eye, const D3DXVECTOR3& dir, const D3DXVECTOR3& bin, const FLOAT fElapsedTime)
{
    D3DXVECTOR3 ret = eye;
	static FLOAT kAbsoluteSpeed = 2.0f;
    float speed = kAbsoluteSpeed * fElapsedTime;
    if(DXUTIsKeyDown(VK_SHIFT))
        speed *= 5.0f;
	if(DXUTIsKeyDown(VK_UP))
		ret += dir * speed;
	if(DXUTIsKeyDown(VK_DOWN))
		ret -= dir * speed;
	if(DXUTIsKeyDown(VK_LEFT))
		ret += bin * speed;
	if(DXUTIsKeyDown(VK_RIGHT))
		ret -= bin * speed;
    return ret;
}

float CameraDrive::QueryTerrainHeight(const D3DXVECTOR3& pos) const
{
    float height = -FLT_MAX;
   extern Scape* _scape;
    if(_scape)
    {
        D3DXVECTOR2 xzLoc(pos.x, pos.z);
        D3DXVECTOR3 norm;
        _scape->QueryHeight(xzLoc, height, norm);
    }
    return height;
}
float CameraDrive::FitToGround(D3DXVECTOR3& pos, const float radius) const
{
    float height = pos.y;
    extern Scape* _scape;
    if(_scape)
    {
        D3DXVECTOR2 xzLoc(pos.x, pos.z);
        D3DXVECTOR3 norm;
        _scape->QueryHeight(xzLoc, height, norm);
    }
    return height;
}

FLOAT CameraDrive::GetYaw(const D3DXVECTOR3& lookAtDir) const
{
	FLOAT yaw = atan2(lookAtDir.z, lookAtDir.x);
    return yaw;
}

D3DXVECTOR3 CameraDrive::SetYaw(const FLOAT yaw, const FLOAT up) const
{
    const FLOAT s = sin(yaw);
    const FLOAT c = cos(yaw);

    return D3DXVECTOR3(c, up, s);
}

FLOAT CameraDrive::BlendYaw(FLOAT lookAtYaw, const FLOAT travelYaw, const FLOAT fSecs, const FLOAT yawRate) const
{
    const FLOAT maxYaw = yawRate * fSecs;
    float diff = travelYaw - lookAtYaw;
    if(diff > M_PI)
        diff -= 2.0f * M_PI;
    else if(diff < -M_PI)
        diff += 2.0f * M_PI;
    if(diff > maxYaw)
    {
        lookAtYaw += maxYaw;
    }
    else if(diff < -maxYaw)
    {
        lookAtYaw -= maxYaw;
    }
    else
    {
        lookAtYaw = travelYaw;
    }
    return lookAtYaw;
}

D3DXVECTOR3 CameraDrive::RotateYaw(const D3DXVECTOR3& dir, const FLOAT rads) const
{
    const FLOAT s = sin(rads);
    const FLOAT c = cos(rads);

    D3DXVECTOR3 newDir(
        dir.x * c + -dir.z * s,
        dir.y,
        dir.x * s + dir.z * c);

    return newDir;
}

FLOAT CameraDrive::GetPitch(const D3DXVECTOR3& lookAtDir) const
{
	return asin(lookAtDir.y);
}
FLOAT CameraDrive::SetPitch(const FLOAT pitch) const
{
    return sin(pitch);
}
FLOAT CameraDrive::BlendPitch(FLOAT lookAtPitch, const FLOAT travelPitch, const FLOAT fSecs, const FLOAT pitchRate) const
{
    const FLOAT maxPitch = pitchRate * fSecs;
    float diff = travelPitch - lookAtPitch;
    if(diff > maxPitch)
        lookAtPitch += maxPitch;
    else if(diff < -maxPitch)
        lookAtPitch -= maxPitch;
    else
        lookAtPitch = travelPitch;
    return lookAtPitch;
}

D3DXVECTOR3 CameraDrive::BlendToGoal(const D3DXVECTOR3& curr, const D3DXVECTOR3& goal, const FLOAT fElapsed, const D3DXVECTOR2& rate) const
{
    const float goalYaw = GetYaw(goal);
    const float currYaw = GetYaw(curr);

    const float newYaw = BlendYaw(currYaw, goalYaw, fElapsed, rate.x);

    const float goalPitch = GetPitch(goal);
    const float currPitch = GetPitch(curr);
    const float newPitch = BlendPitch(currPitch, goalPitch, fElapsed, rate.y);

    D3DXVECTOR3 newDir = SetYaw(newYaw, SetPitch(newPitch));

    D3DXVec3Normalize(&newDir, &newDir);

    return newDir;
}
void CameraDrive::BlendLookAtToTravel(const FLOAT fSecs)
{
    if(!DXUTIsMouseButtonDown(VK_LBUTTON))
    {
        const D3DXVECTOR3 travel = GetTravelDirection(m_vEyePt, m_vLookAtPt);
        const D3DXVECTOR3 lookAtDir = GetLookAtDirection(m_vEyePt, m_vLookAtPt);

        const D3DXVECTOR3 newDir = lookAtDir + (travel - lookAtDir) * (1.0f - pow(0.5f, fSecs));
    
        m_vLookAtPt = m_vEyePt + newDir * GetFollowDistance(m_vEyePt, m_vLookAtPt);
    }
}


D3DXVECTOR3 CameraDrive::GetTravelDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
	D3DXVECTOR3 direction = lookAt - eye;

    D3DXVec3Normalize(&direction, &direction);

    return direction;
}

FLOAT CameraDrive::GetFollowDistance(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
    return 5.0f;
}

D3DXVECTOR3 CameraDrive::GetLookAtDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
	D3DXVECTOR3 direction = lookAt - eye;

    D3DXVec3Normalize(&direction, &direction);

    return direction;
}

static FLOAT UpdatePitch(D3DXVECTOR3& lookAtDir, const FLOAT pitchSpeed, const FLOAT fElapsedTime)
{
	FLOAT pitch = asin(lookAtDir.y);
	pitch += pitchSpeed * fElapsedTime;
	const FLOAT SMALL = 1.0e-2f;
	if(pitch > M_PI_2 - SMALL)
		pitch = FLOAT(M_PI_2) - SMALL;
	else if(pitch < -M_PI_2 + SMALL)
		pitch = -FLOAT(M_PI_2) + SMALL;

	lookAtDir.y = sin(pitch);

    return pitch;
}

static FLOAT UpdateYaw(D3DXVECTOR3& lookAtDir, const FLOAT yawSpeed, const FLOAT fElapsedTime)
{
	FLOAT newXZLen = sqrt(1.0f - lookAtDir.y * lookAtDir.y);

	FLOAT yaw = atan2(lookAtDir.z, lookAtDir.x);
	yaw += yawSpeed * fElapsedTime;

	lookAtDir.x = cos(yaw) * newXZLen;
	lookAtDir.z = sin(yaw) * newXZLen;

    return yaw;
}
VOID CameraDrive::Advance(const CFlyCamera& camera, const FLOAT fElapsedTime)
{
    if(SetFromLog())
    {
        return;
    }

    const D3DXVECTOR2& mousePos = camera.GetMousePosition();
    const D3DXVECTOR2& lastPos = camera.GetLastMousePosition();

    extern bool DisableCameraKeyboard();
    if(!DisableCameraKeyboard())
    {
        UpdateGoals(mousePos, lastPos, fElapsedTime);
    }
    AdvanceToGoals(fElapsedTime);

    D3DXVECTOR3 lookAt = *camera.GetLookAtPt();
    D3DXVECTOR3 eye = *camera.GetEyePt();

    const FLOAT dist = GetFollowDistance(eye, lookAt);

    D3DXVECTOR3 lookAtDir = GetLookAtDirection(eye, lookAt);

    const D3DXVECTOR3 travel = GetTravelDirection(eye, lookAt);

    eye += travel * fElapsedTime * speedCurr;

    UpdatePitch(lookAtDir, pitchCurr, fElapsedTime);

    UpdateYaw(lookAtDir, yawCurr, fElapsedTime);

	D3DXVECTOR3 up = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	D3DXVECTOR3 bin;
	D3DXVec3Cross(&bin, &lookAtDir, &up);
	D3DXVec3Normalize(&bin, &bin);
	D3DXVec3Cross(&up, &bin, &lookAtDir);
	D3DXVec3Normalize(&up, &up);

    if(!DisableCameraKeyboard())
    {
        eye = UpdateManuals(eye, lookAtDir, bin, fElapsedTime);
    }

	lookAt = eye + dist * lookAtDir;

    m_vEyePt = eye;
    m_vLookAtPt = lookAt;
    m_vUpDir = up;

    AddToLog();
}


typedef std::vector<Object*>		ObjectList;
typedef ObjectList::iterator		ObjectIter;
typedef ObjectList::const_iterator	ConstObjectIter;

float CameraDrive::CheckCollision(const D3DXVECTOR3& eyePt, const D3DXVECTOR3& travel, const FLOAT elapsed) const
{
    const float kCameraWidth = 0.25f;
    const float kCameraLength = 0.5f;
    const float kLookAheadSecs = 0.5f;
    const float lookAhead = kLookAheadSecs * speedCurr + kCameraLength;
    const D3DXVECTOR3 axis = D3DXVECTOR3(travel.x, 0.0f, travel.z);
    const D3DXVECTOR3 extent = D3DXVECTOR3(-travel.z, 0, travel.x);

    D3DXVECTOR2 max(-FLT_MAX, -FLT_MAX);
    D3DXVECTOR2 min(FLT_MAX, FLT_MAX);

    for(int i = 0; i < 4; ++i)
    {
        D3DXVECTOR3 pos = eyePt;
        pos += extent * (i & 1 ? kCameraWidth : -kCameraWidth);
        if(i & 2)
            pos += axis * lookAhead;

        max.x = __max(max.x, pos.x);
        min.x = __min(min.x, pos.x);

        max.y = __max(max.y, pos.z);
        min.y = __min(min.y, pos.z);
    }

    extern POINT numModels;
    extern float modelSpacing;

    min -= D3DXVECTOR2(modelSpacing, modelSpacing);
    max += D3DXVECTOR2(modelSpacing, modelSpacing);

    POINT lo;
    lo.x = int(min.x / modelSpacing + (numModels.x - 1) * 0.5f);
    lo.y = int(min.y / modelSpacing + (numModels.y - 1) * 0.5f);
    lo.x = __max(lo.x, 0);
    lo.y = __max(lo.y, 0);
    POINT hi;
    hi.x = int(max.x / modelSpacing + (numModels.x - 1) * 0.5f);
    hi.y = int(max.y / modelSpacing + (numModels.y - 1) * 0.5f);
    hi.x = __min(hi.x, numModels.x - 1);
    hi.y = __min(hi.y, numModels.y - 1);

    extern ObjectList objects;

    float height = QueryTerrainHeight(eyePt) + 1.0f;

    for(int y = lo.y; y <= hi.y; ++y)
    {
        for(int x = lo.x; x <= hi.x; ++x)
        {
            // Terrain is first object in queue. Ignore it.
            Object* object = objects[x + y * numModels.x + 1];

            float objHeight = CheckObjectHit(object, eyePt, extent, kCameraWidth, axis, lookAhead) + 0.25f;
            height = __max(height, objHeight);
        }
    }
    return height;
}

float CameraDrive::CheckObjectHit(const Object* object, 
    const D3DXVECTOR3& pos,
    const D3DXVECTOR3& biAxis, const float extent,
    const D3DXVECTOR3& axis, const float length) const
{
    const D3DXMATRIX& l2w = object->GetLocalToWorld();
    D3DXMATRIX w2l;
    D3DXMatrixInverse(&w2l, NULL, &l2w);

    const Bounds& localBnd = object->GetLocalBounds();
    const D3DXVECTOR3& localMin = localBnd.Min();
    const D3DXVECTOR3& localMax = localBnd.Max();
    int mask = 0xf;
    for(int i = 0; i < 4; ++i)
    {
        D3DXVECTOR3 corner = pos;
        corner += biAxis * (i & 1 ? extent : -extent);
        if(i & 2)
            corner += axis * length;
        D3DXVECTOR4 pt;
        D3DXVec3Transform(&pt, &corner,  &w2l);

        if(pt.x > localMin.x)
            mask &= ~0x1;
        if(pt.x < localMax.x)
            mask &= ~0x2;
        if(pt.y > localMin.y)
            mask &= ~0x4;
        if(pt.y < localMax.y)
            mask &= ~0x8;
    }
    if(mask)
        return -FLT_MAX;

    const D3DXVECTOR3 biStart = pos - biAxis * extent;
    const float minBiDot = D3DXVec3Dot(&biStart, &biAxis);
    const float maxBiDot = minBiDot + D3DXVec3Dot(&biAxis, &biAxis);
    const D3DXVECTOR3 centerStart = pos;
    const D3DXVECTOR3 centerEnd = centerStart + axis * length;
    const float minAxDot = D3DXVec3Dot(&centerStart, &axis);
    const float maxAxDot = D3DXVec3Dot(&centerEnd, &axis);

    const Bounds& worldBnd = object->GetWorldBounds();
    const D3DXVECTOR3& worldMin = worldBnd.Min();
    const D3DXVECTOR3& worldMax = worldBnd.Max();
    mask = 0xf;
    for(int j = 0; j < 2; ++j)
    {
        for(int i = 0; i < 2; ++i)
        {
            D3DXVECTOR3 crn(
                i & 1 ? worldMax.x : worldMin.x,
                0.0f,
                j & 1 ? worldMax.z : worldMin.z);

            const float biDot = D3DXVec3Dot(&crn, &biAxis);
            if(biDot > minBiDot)
                mask &= ~0x1;
            if(biDot < maxBiDot)
                mask &= ~0x2;
            const float dot = D3DXVec3Dot(&crn, &axis);
            if(dot > minAxDot)
                mask &= ~0x4;
            if(dot < maxAxDot)
                mask &= ~0x8;
        }
    }
    if(mask)
        return -FLT_MAX;

    return worldBnd.Max().y;
}

bool CameraDrive::BeginBuildLog()
{
    log.clear();
    m_LogMode = kLogBuild;
    return true;
}

bool CameraDrive::EndBuildLog(const WCHAR* const fileName)
{
    m_LogMode = kLogNone;
    if(fileName)
    {
        SaveLog(fileName);
        log.clear();
        return true;
    }
    return false;
}

bool CameraDrive::SaveLog(const WCHAR* const fileName) const
{
    if(log.size())
    {
        FILE* fp;
        if(!_wfopen_s(&fp, fileName, L"wt"))
        {
            int numElem = log.size();
            fwprintf_s(fp, L"%d\n", numElem);

            for(int i = 0; i < numElem; ++i)
            {
                const LogEntry& entry = log[i];

                fwprintf_s(fp, L"%f ", entry.eyePt.x);
                fwprintf_s(fp, L"%f ", entry.eyePt.y);
                fwprintf_s(fp, L"%f\n", entry.eyePt.z);

                fwprintf_s(fp, L"%f ", entry.lookAtPt.x);
                fwprintf_s(fp, L"%f ", entry.lookAtPt.y);
                fwprintf_s(fp, L"%f\n", entry.lookAtPt.z);

                fwprintf_s(fp, L"%f ", entry.upDir.x);
                fwprintf_s(fp, L"%f ", entry.upDir.y);
                fwprintf_s(fp, L"%f\n", entry.upDir.z);
            }
            fclose(fp);
            return true;
        }
    }
    return false;
}

bool CameraDrive::LoadLog(const WCHAR* const fileName)
{
    log.clear();

    FILE* fp = NULL;
    if( !_wfopen_s(&fp, fileName, L"rt") )
    {
        assert(fp);
        int numElem = 0;
        fwscanf_s(fp, L"%d", &numElem);
        if(numElem)
        {
            for(int i = 0; i < numElem; ++i)
            {
                LogEntry entry;
                fwscanf_s(fp, L"%f", &entry.eyePt.x);
                fwscanf_s(fp, L"%f", &entry.eyePt.y);
                fwscanf_s(fp, L"%f", &entry.eyePt.z);

                fwscanf_s(fp, L"%f", &entry.lookAtPt.x);
                fwscanf_s(fp, L"%f", &entry.lookAtPt.y);
                fwscanf_s(fp, L"%f", &entry.lookAtPt.z);

                fwscanf_s(fp, L"%f", &entry.upDir.x);
                fwscanf_s(fp, L"%f", &entry.upDir.y);
                fwscanf_s(fp, L"%f", &entry.upDir.z);

                log.push_back(entry);
            }
        }
        fclose(fp);
    }
    return log.size() > 0;
}

bool CameraDrive::BeginFromLog(const WCHAR* const fileName)
{
    if(fileName)
    {
        LoadLog(fileName);
    }
    if(log.size() > 0)
    {
        m_LogMode = kLogFrom;
        m_LogCurrent = 0;
        m_LogLoop = 0;
        m_LogTime = DXUTGetGlobalTimer()->GetAbsoluteTime();
        return true;
    }
    return false;
}

bool CameraDrive::EndFromLog()
{
    m_LogMode = kLogNone;
    return true;
}

void CameraDrive::AddToLog()
{
    if(ToLog())
    {
        LogEntry entry;
        entry.eyePt = m_vEyePt;
        entry.lookAtPt = m_vLookAtPt;
        entry.upDir = m_vUpDir;
        log.push_back(entry);
    }
}

bool CameraDrive::SetFromLog()
{
    if(FromLog())
    {
        const LogEntry& entry = log[m_LogCurrent];
        m_vEyePt = entry.eyePt;
        m_vLookAtPt = entry.lookAtPt;
        m_vUpDir = entry.upDir;

        if(++m_LogCurrent >= int(log.size()))
        {
            SaveLogTime();
            m_LogCurrent = 0;
            ++m_LogLoop;
            m_LogMode = kLogNone;
        }

        return true;
    }
    return false;
}

void CameraDrive::SaveLogTime()
{
    const double startTime = m_LogTime;
    m_LogTime = DXUTGetGlobalTimer()->GetAbsoluteTime();
    double secs = m_LogTime - startTime;
    FILE* fp;
    if(!_wfopen_s(&fp, L"LogTimes.txt", L"a+t"))
    {
        fwprintf_s(fp, L"loop %d: %d frames %.4f s %.2f fps\n", 
            m_LogLoop, 
            log.size(), 
            secs,
            double(log.size()) / secs
            );
        
        fclose(fp);
    }
    // Reset the start time, in case the file IO took any considerable time.
    m_LogTime = DXUTGetGlobalTimer()->GetAbsoluteTime();
}

