
#ifndef SHARED_STRUCT_H
#define SHARED_STRUCT_H

#include <assert.h>

///////////////////////////////////////////////////////////////////////////////////////////
// SharedStruct:
// Templatized (generic) wrapper for a memory object shared across multiple processes.
///////////////////////////////////////////////////////////////////////////////////////////

template <typename T> class SharedStruct
{
public:

    // Constructor, initialize all members to blank state.
    SharedStruct()
        : m_hData(0), m_pData((T*)0), m_bWritable(false)
    {
    }

    // Close down any open data ports.
    virtual ~SharedStruct()
    {
        Close();
    }

    // Create the memory object with associated ID.
    // Returns true on success. 
    bool OpenMaster(const LPTSTR name)
    {
	    m_hData = CreateFileMapping(
                 INVALID_HANDLE_VALUE,  // use paging file
                 NULL,                  // default security
                 PAGE_READWRITE,        // read/write access
                 0,                     // maximum object size (high-order DWORD)
                 sizeof(T),             // maximum object size (low-order DWORD)
                 name);                 // name of mapping object

	    m_pData = (T*) MapViewOfFile(m_hData,   // handle to map object
                        FILE_MAP_ALL_ACCESS,    // read/write permission (mafinch - on Win, WRITE is equivalent to ALL_ACCESS (i.e. read/write)).
                        0,                      // Offset high
                        0,                      // Offset low
                        sizeof(T));             // Number bytes to map

		m_bWritable = true;
        return m_pData != (T*)0;
    }

    // Open existing shared object by identifier. If "writable", write access is granted,
    // otherwise only read access. This is checked in accessors. See T* Data() vs const T* Data().
    // Returns true on success. Failure generally means Master process hasn't set up the 
    // shared object yet. Try again later.
    bool OpenClient(const LPTSTR name, bool writable=false)
    {
        const DWORD access = writable ? FILE_MAP_ALL_ACCESS : FILE_MAP_READ;
        m_hData = OpenFileMapping(
                   access,   // read/write access
                   FALSE,                 // do not inherit the name
                   name);               // name of mapping object

		m_pData = (T*)MapViewOfFile(m_hData,   // handle to map object
                        access, // read/write permission
                        0,
                        0,
                        sizeof(T));

        m_bWritable = writable;

        return m_pData != (T*)(0);
    }

    // Close up and release all resources.
    void Close()
    {
        if(m_pData)
        {
		    UnmapViewOfFile(m_pData);
		    m_pData = (T*)0;
        }

        if(m_hData)
        {
		    CloseHandle(m_hData);
            m_hData = 0;
        }
    }

    // True if shared data is currently accessible.
    bool Valid() const { return m_pData != (T*)0; }

    // Access data read/write using pointer syntax, e.g. mySharedStruct->myDataField
    T* operator -> () { return Data(); }
    // Access read only data using pointer syntax, e.g. mySharedStruct->myDataField
    const T* operator -> () const { return Data(); }

    // Access pointer to writable data struct. Note that this checks that you have writable access.
    // If you get an assertion here for reading a member of your data, take a const ref or pointer
    // and use that, e.g.:
    //      const SharedStruct<MyDataType>& tmp = mySharedSruct;
    //      float sum = tmp->x + tmp->y;
    // If you get this assertion for modifying a member of your data, you need to open writable
    // and/or worry about synchronization.
    T* Data() { assert(m_bWritable); return m_pData; }

    // Access readonly pointer to data struct.
    const T* Data() const { return m_pData; }

private:
    // Handle to memory mapped file.
    HANDLE  m_hData;
    // Pointer to shared memory.
    T*      m_pData;
    // Store whether data was opened with write permissions.
    bool    m_bWritable;
};

#endif // SHARED_STRUCT_H
