#include "pch.h"

#include <windows.h>
#include <d3d11.h>
//#include <d3dcompiler.h>
#include <DirectXMath.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include "Utility.h"

#include "CompiledShader.h"

#include "BlitQuad.h"
#include "SaveRenderState.h"

struct BlitConstBuffer
{
    XMVECTOR		m_vScaleOffSrc;
    XMVECTOR        m_vScaleOffDst;
    XMVECTOR        m_vTint;
};

bool BlitQuad::m_Forced = false;

#pragma warning(push)
#pragma warning(disable : 4482)
BlitQuad::BlitQuad()
	:	m_DstRect(0.0f, 0.0f, 1.0f, 1.0f),
		m_SrcRect(0.0f, 0.0f, 1.0f, 1.0f),
        m_Tint(1.0f, 1.0f, 1.0f, 1.0f),
		m_pVertexShader(NULL),
		m_pPixelShader(NULL),
		m_pRasterizerState(NULL),
		m_pDepthStencilState(NULL),
        m_pTexSampler(NULL),
        m_bPointSample(false),
		m_pConstantBuffer(NULL),
        m_pBlendState(NULL),
        m_Blend(Blend::kNone)
{
}

void BlitQuad::Release()
{
	SAFE_RELEASE(m_pVertexShader);
	SAFE_RELEASE(m_pPixelShader);
	SAFE_RELEASE(m_pRasterizerState);
	SAFE_RELEASE(m_pDepthStencilState);
    SAFE_RELEASE(m_pTexSampler);
	SAFE_RELEASE(m_pConstantBuffer);
    SAFE_RELEASE(m_pBlendState);
}

HRESULT BlitQuad::Init(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const WCHAR* entry, const WCHAR* shaderModel)
{
	HRESULT hr = S_OK;
	if(shaderFile == NULL)
	{
		shaderFile = L"BlitQuad.fx";
	}
	if(entry == NULL)
	{
		entry = L"Def";
	}
	if(shaderModel == NULL)
	{
		shaderModel = L"4_0";
	}
	hr = LoadShaders(pd3dDevice, shaderFile, entry, shaderModel);
	if(FAILED(hr))
		return hr;

	hr = InitState(pd3dDevice);
	if(FAILED(hr))
		return hr;

    hr = InitSampler(pd3dDevice);
    if(FAILED(hr))
        return hr;

	hr = InitConstantBuffers(pd3dDevice);
	if(FAILED(hr))
		return hr;

	return hr;
}

HRESULT BlitQuad::LoadShaders(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const WCHAR* entry, const WCHAR* sm)
{
	WCHAR vsEntry[256];
	swprintf_s(vsEntry, L"VS_%s", entry);
	WCHAR vsSM[256];
	swprintf_s(vsSM, L"vs_%s", sm);
	HRESULT hr = LoadVertexShader(pd3dDevice, shaderFile, vsEntry, vsSM);
	if(FAILED(hr))
		return hr;

	WCHAR psEntry[256];
	swprintf_s(psEntry, L"PS_%s", entry);
	WCHAR psSM[256];
	swprintf_s(psSM, L"ps_%s", sm);
	hr = LoadPixelShader(pd3dDevice, shaderFile, psEntry, psSM);
	if(FAILED(hr))
		return hr;

    m_Forced = false;

	return hr;
}

void BlitQuad::SetConstants(ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* tex)
{
	immediateContext->PSSetShaderResources( 0, 1, &tex );

	BlitConstBuffer buff;
	buff.m_vScaleOffDst = XMLoadFloat4(&XMFLOAT4(
		(m_DstRect.hi.x - m_DstRect.lo.x) * 2.0f,
		(m_DstRect.hi.y - m_DstRect.lo.y) * 2.0f,
		m_DstRect.lo.x * 2.0f - 1.0f,
		m_DstRect.lo.y * 2.0f - 1.0f));

	buff.m_vScaleOffSrc = XMLoadFloat4(&XMFLOAT4(
		m_SrcRect.hi.x - m_SrcRect.lo.x,
		m_SrcRect.hi.y - m_SrcRect.lo.y,
		m_SrcRect.lo.x,
		m_SrcRect.lo.y));

    buff.m_vTint = XMLoadFloat4(&m_Tint);

	immediateContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &buff, 0, 0 );
    immediateContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
    immediateContext->PSSetConstantBuffers( 0, 1, &m_pConstantBuffer );

}

void BlitQuad::UnSetConstants(ID3D11DeviceContext* immediateContext)
{
	ID3D11ShaderResourceView* nullArray = NULL;
	immediateContext->PSSetShaderResources( 0, 1, &nullArray );
}

void BlitQuad::SetState(ID3D11RenderTargetView* rt, ID3D11DeviceContext* immediateContext)
{
	if(rt != NULL)
	{
        ID3D11RenderTargetView* aRTViews[ 1 ] = { rt };
        immediateContext->OMSetRenderTargets( 1, aRTViews, NULL );
	}

    immediateContext->OMSetDepthStencilState( m_pDepthStencilState, 0 );

    immediateContext->RSSetState( m_pRasterizerState );

    const float blendFactor[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    const UINT sampleMask = 0xffffffff;
    immediateContext->OMSetBlendState(m_pBlendState, blendFactor, sampleMask);

    // Setup the viewport to match the backbuffer
    D3D11_VIEWPORT vp;
    vp.Width = (float)m_DstResolution.x;
    vp.Height = (float)m_DstResolution.y;
    vp.MinDepth = 0.0f;
    vp.MaxDepth = 1.0f;
    vp.TopLeftX = 0;
    vp.TopLeftY = 0;
    immediateContext->RSSetViewports( 1, &vp );


    immediateContext->IASetInputLayout( NULL );
    immediateContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );

    immediateContext->VSSetShader( m_pVertexShader, NULL, 0 );
    immediateContext->PSSetShader( m_pPixelShader, NULL, 0 );

    immediateContext->PSSetSamplers( 0, 1, &m_pTexSampler);

}

void BlitQuad::Render(ID3D11DeviceContext* immediateContext, ID3D11RenderTargetView* rt, ID3D11ShaderResourceView* tex)
{
	SaveRenderStateAuto state(immediateContext);

	SetConstants(immediateContext, tex);

	SetState(rt, immediateContext);

    immediateContext->Draw( 4, 0 );

	UnSetConstants(immediateContext);
}

HRESULT BlitQuad::InitState(ID3D11Device* pd3dDevice)
{
    D3D11_RASTERIZER_DESC drd = {
        D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
        D3D11_CULL_NONE,//D3D11_CULL_MODE CullMode;
        FALSE, //BOOL FrontCounterClockwise;
        0, //INT DepthBias;
        0.0f,//FLOAT DepthBiasClamp;
        0.0f,//FLOAT SlopeScaledDepthBias;
        FALSE,//BOOL DepthClipEnable;
        FALSE,//BOOL ScissorEnable;
        FALSE,//BOOL MultisampleEnable;
        FALSE//BOOL AntialiasedLineEnable;        
    };
    HRESULT hr = pd3dDevice->CreateRasterizerState(&drd, &m_pRasterizerState);
	if( FAILED( hr ) )
		return hr;

    // Depth stencil state
    D3D11_DEPTH_STENCIL_DESC DSDesc;
    ZeroMemory( &DSDesc, sizeof( D3D11_DEPTH_STENCIL_DESC ) );
    DSDesc.DepthEnable = FALSE;
    DSDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ZERO;
    DSDesc.DepthFunc = D3D11_COMPARISON_ALWAYS;
    DSDesc.StencilEnable = FALSE;
	hr = pd3dDevice->CreateDepthStencilState( &DSDesc, &m_pDepthStencilState );
	if( FAILED(hr) )
		return hr;

    hr = InitBlendState(pd3dDevice);
    if(FAILED(hr))
        return hr;

	return hr;
}
HRESULT BlitQuad::InitSampler(ID3D11Device* pd3dDevice)
{
    HRESULT hr = S_OK;
	if(!m_pTexSampler)
	{
		D3D11_SAMPLER_DESC sampDesc;
		ZeroMemory( &sampDesc, sizeof(sampDesc) );
		sampDesc.Filter = m_bPointSample 
                        ? D3D11_FILTER_MIN_MAG_MIP_POINT 
                        : D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
		sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
		sampDesc.MinLOD = 0;
		sampDesc.MaxLOD = D3D11_FLOAT32_MAX;
		hr = pd3dDevice->CreateSamplerState( &sampDesc, &m_pTexSampler );
		if( FAILED(hr) )
			return hr;
	}
    return S_OK;
}
void BlitQuad::SetPointSample(ID3D11Device* pd3dDevice, bool b)
{ 
    if(m_bPointSample != b)
    {
        m_bPointSample = b; 
        SAFE_RELEASE(m_pTexSampler);
        InitSampler(pd3dDevice);
    }
}

HRESULT BlitQuad::InitConstantBuffers(ID3D11Device* pd3dDevice)
{
	HRESULT hr = S_OK;

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(BlitConstBuffer);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer );
    if( FAILED( hr ) )
        return hr;

	return hr;
}

HRESULT BlitQuad::InitBlendState(ID3D11Device* pd3dDevice)
{
    HRESULT hr = S_OK;

    D3D11_BLEND_DESC blendDesc;
    ZeroMemory(&blendDesc, sizeof(blendDesc));
    blendDesc.AlphaToCoverageEnable = false;
    blendDesc.IndependentBlendEnable = false;

    // Aliasing here for brevity and readability.
    D3D11_RENDER_TARGET_BLEND_DESC& rtBlend = blendDesc.RenderTarget[0];
    rtBlend.RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE_ALL;
    rtBlend.SrcBlendAlpha = D3D11_BLEND_ONE;
    rtBlend.DestBlendAlpha = D3D11_BLEND_INV_SRC_ALPHA;
    rtBlend.BlendOp = D3D11_BLEND_OP_ADD;
    rtBlend.BlendOpAlpha = D3D11_BLEND_OP_ADD;

    switch(m_Blend)
    {
    case Blend::kSrcAlpha:
        rtBlend.BlendEnable = true;
        rtBlend.SrcBlend = D3D11_BLEND_SRC_ALPHA;
        rtBlend.DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
        break;
    case Blend::kSrcAlphaPM:
        rtBlend.SrcBlend = D3D11_BLEND_ONE;
        rtBlend.DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
        rtBlend.BlendEnable = true;
        break;
    case Blend::kInvSrcAlpha:
        rtBlend.SrcBlend = D3D11_BLEND_INV_SRC_ALPHA;
        rtBlend.DestBlend = D3D11_BLEND_SRC_ALPHA;
        rtBlend.BlendEnable = true;
        break;
    case kAdd:
        rtBlend.SrcBlend = D3D11_BLEND_ONE;
        rtBlend.DestBlend = D3D11_BLEND_ONE;
        rtBlend.BlendEnable = true;
        break;

    case Blend::kNone:
        rtBlend.BlendEnable = false;
        break;

    default:
        assert(false);
        rtBlend.BlendEnable = false;
        break;
    }

    hr = pd3dDevice->CreateBlendState(&blendDesc, &m_pBlendState);
    if( FAILED(hr) )
        return hr;

    return hr;
}

void BlitQuad::MakeDefines(D3D_SHADER_MACRO* defines, bool vtx)
{
    std::vector<D3D_SHADER_MACRO> macroVec;

    if(vtx)
        AppendVertexDefines(macroVec);
    else
        AppendPixelDefines(macroVec);

    for(int i = 0; i < int(macroVec.size()); ++i)
    {
        defines[i] = macroVec[i];
    }
    D3D_SHADER_MACRO end;
    end.Name = NULL;
    end.Definition = NULL;
    defines[macroVec.size()] = end;
}

HRESULT BlitQuad::LoadVertexShader(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const WCHAR* entry, const WCHAR* sm)
{
	HRESULT hr = S_OK;

#if 0
    // Compile the vertex shader
    ID3DBlob* pVSBlob = NULL;
    // You should use the lowest possible shader profile for your shader to enable various feature levels. These
    // shaders are simple enough to work well within the lowest possible profile, and will run on all feature levels
    DWORD dwShaderFlags = D3DCOMPILE_ENABLE_STRICTNESS;
#if defined( DEBUG ) || defined( _DEBUG )
    // Set the D3DCOMPILE_DEBUG flag to embed debug information in the shaders.
    // Setting this flag improves the shader debugging experience, but still allows 
    // the shaders to be optimized and to run exactly the way they will run in 
    // the release configuration of this program.
    dwShaderFlags |= D3DCOMPILE_DEBUG;
	/// SKIP_OPTIMIZATION flag will currently cause the compiler to fail on the symbolic differentiation.
	//dwShaderFlags |= D3DCOMPILE_SKIP_OPTIMIZATION;
#endif

    D3D_SHADER_MACRO defines[128];
    MakeDefines(defines, true);

    ID3DBlob* pErrMsg = NULL;
    hr = D3DX11CompileFromFile( shaderFile, defines, NULL, entry, sm, dwShaderFlags, 0, NULL,
                                     &pVSBlob, &pErrMsg, NULL );
    if( FAILED( hr ) )
    {
		const char* string = (const char*)pErrMsg->GetBufferPointer();
        MessageBox( NULL,
                    L"The FX file cannot be compiled.  Please run this executable from the directory that contains the FX file.", L"Error", MB_OK );
		string = NULL;
        return hr;
    }

    // Create the vertex shader
    hr = pd3dDevice->CreateVertexShader( pVSBlob->GetBufferPointer(), pVSBlob->GetBufferSize(), NULL, &m_pVertexShader );
    if( FAILED( hr ) )
    {    
        pVSBlob->Release();
        return hr;
    }
#else
    std::vector<D3D_SHADER_MACRO> defines;
    AppendVertexDefines(defines);

    CompiledShader compiler;
    compiler.SetForced(m_Forced);
    compiler.SetShaderFolder(L"media\\MAFShaders");
    compiler.AddDefines(defines);
    compiler.Compile(shaderFile, sm, entry);

    hr = pd3dDevice->CreateVertexShader( compiler.GetBufferPointer(), compiler.GetBufferSize(), NULL, &m_pVertexShader);
#endif

	return hr;
}

HRESULT BlitQuad::LoadPixelShader(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const WCHAR* entry, const WCHAR* sm)
{
	HRESULT hr = S_OK;

#if 0
    // Compile the pixel shader
    ID3DBlob* pPSBlob = NULL;
    DWORD dwShaderFlags = D3DCOMPILE_ENABLE_STRICTNESS;
#if defined( DEBUG ) || defined( _DEBUG )
    // Set the D3DCOMPILE_DEBUG flag to embed debug information in the shaders.
    // Setting this flag improves the shader debugging experience, but still allows 
    // the shaders to be optimized and to run exactly the way they will run in 
    // the release configuration of this program.
    dwShaderFlags |= D3DCOMPILE_DEBUG;
	/// SKIP_OPTIMIZATION flag will currently cause the compiler to fail on the symbolic differentiation.
	//dwShaderFlags |= D3DCOMPILE_SKIP_OPTIMIZATION;
#endif
    D3D_SHADER_MACRO defines[128];
    MakeDefines(defines, false);

    ID3DBlob* pErrMsg = NULL;
    hr = D3DX11CompileFromFile( shaderFile, defines, NULL, entry, sm, dwShaderFlags, 0, NULL,
                                     &pPSBlob, &pErrMsg, NULL );
    if( FAILED( hr ) )
    {
		const char* string = (const char*)pErrMsg->GetBufferPointer();
        MessageBox( NULL,
                    L"The FX file cannot be compiled.  Please run this executable from the directory that contains the FX file.", L"Error", MB_OK );
		string = NULL;
        return hr;
    }

    // Create the pixel shader
    hr = pd3dDevice->CreatePixelShader( pPSBlob->GetBufferPointer(), pPSBlob->GetBufferSize(), NULL, &m_pPixelShader );
    pPSBlob->Release();

#else
    std::vector<D3D_SHADER_MACRO> defines;
    AppendPixelDefines(defines);

    CompiledShader compiler;
    compiler.SetForced(m_Forced);
    compiler.SetShaderFolder(L"media\\MAFShaders");
    compiler.AddDefines(defines);
    compiler.Compile(shaderFile, sm, entry);

    hr = pd3dDevice->CreatePixelShader( compiler.GetBufferPointer(), compiler.GetBufferSize(), NULL, &m_pPixelShader);
#endif

	return hr;
}

#pragma warning(pop)