#include "pch.h"

#include <windows.h>
#include <assert.h>
#include <d3d11.h>

#include "Utility.h"
#include "Cascade.h"
#include "OffscreenRT.h"

static void MakeJitterTable();
static const int _jitterSamples = 300;
static XMFLOAT2 _jitterTable[_jitterSamples];
bool Cascade::m_bJitterEnabled = true;

static const UINT kPreviousColorSlot = 26;

void Cascade::Init()
{
    MakeJitterTable();
}

Cascade::Cascade()
	: 	m_fDuplication(1.0f),
        m_iPing(0),
        m_Jitter(0.0f, 0.0f),
        m_fJitterScale(1.0f),
        m_fRefreshInterval(0.0f),
        m_fRefreshCountDown(0.0f)
{
    m_aOffscreenRT[0] = m_aOffscreenRT[1] = NULL;

    m_mLastCameraToNDC = XMMatrixIdentity();
    m_mLastWorldToCamera = XMMatrixIdentity();

    m_iJitterIndex = int(double(rand()) / RAND_MAX * _jitterSamples);
}

Cascade::~Cascade()
{
	Release();
}

void Cascade::PingPong()
{
    m_iPing = m_iPing ? 0 : 1;
}

int Cascade::GetWidth() const
{
    return m_aOffscreenRT[m_iPing]->GetWidth();
}
int Cascade::GetHeight() const
{
    return m_aOffscreenRT[m_iPing]->GetHeight();
}

void Cascade::Release()
{
	SAFE_DELETE(m_aOffscreenRT[0]);
	SAFE_DELETE(m_aOffscreenRT[1]);
}

HRESULT Cascade::Create(ID3D11Device* pd3dDevice, int width, int height, int msaa, float dup)
{
    HRESULT hr = S_OK;

	m_aOffscreenRT[0] = new OffscreenRT();
	V_RETURN( m_aOffscreenRT[0]->Create(pd3dDevice, width, height, msaa) );

	m_aOffscreenRT[1] = new OffscreenRT();
	V_RETURN( m_aOffscreenRT[1]->Create(pd3dDevice, width, height, msaa) );

	m_fDuplication = dup;

    return hr;
}

HRESULT Cascade::CreateBase(ID3D11Device* pd3dDevice, int width, int height, int msaa)
{
    HRESULT hr = S_OK;

	m_aOffscreenRT[0] = new OffscreenRT();
	V_RETURN( m_aOffscreenRT[0]->Create(pd3dDevice, width, height, msaa) );

	m_aOffscreenRT[1] = new OffscreenRT();
	V_RETURN( m_aOffscreenRT[1]->Create(pd3dDevice, width, height, msaa) );

    m_fDuplication = 0;

    return hr;
}

static POINT InvertCursor(const POINT& gazePos, const float fbHeight)
{
    POINT pos;
    pos.x = gazePos.x;
    pos.y = int(fbHeight - gazePos.y + 0.5f);
    return pos;
}
#if defined(_DEBUG) || defined(DEBUG)
static bool AboutEqual(float v0, float v1)
{
    const float smallEnough = 1.0e-5f;
    return abs(v0 - v1) / abs(v0 + v1) < smallEnough;
}
#endif // defined(_DEBUG) || defined(DEBUG)
static XMMATRIX CreateInsertProjection(const XMMATRIX& xCam2NDC, 
    const float fbWidth, const float fbHeight, 
    const float cWidth, const float cHeight, 
    const float hither, const float yon,
    const XMFLOAT2& jitter,
    const float duplication,
    const POINT& gazePos)
{
    XMFLOAT4X4 cam2NDC;
    XMStoreFloat4x4(&cam2NDC, xCam2NDC);
    // Note: hither and yon can be extracted from the projection matrix as below,
    // but precision issues make the results unreliable. Best just pass them in,
    // as any camera system should know what it is using for the hither and yon
    // anyway and exactly. If you really really don't know, you could use these.
    // They are close, roughly about 0.1% off, depending of course on the ratio
    // of your hither to yon.
    //const float hither = -double(cam2NDC.m[3][2]) / double(cam2NDC.m[2][2]);
    //const float yon = hither * cam2NDC.m[3][2] / (hither + cam2NDC.m[3][2]);

    const float oldRightPlusLeft = -cam2NDC.m[2][0] / cam2NDC.m[0][0] * 2.0f * hither;
    const float oldRightMinusLeft = 2.0f * hither / cam2NDC.m[0][0];
    const float oldLeft = (oldRightPlusLeft - oldRightMinusLeft) * 0.5f;
    //const float oldRight = (oldRightPlusLeft + oldRightMinusLeft) * 0.5f; // interesting, but useless

    const float oldTopPlusBot = -cam2NDC.m[2][1] / cam2NDC.m[1][1] * 2.0f * hither;
    const float oldTopMinusBot = 2.0f * hither / cam2NDC.m[1][1];
    const float oldBot = (oldTopPlusBot - oldTopMinusBot) * 0.5f;
    //const float oldTop = (oldTopPlusBot + oldTopMinusBot) * 0.5f; // interesting, but useless

    //// Test how it's going. We should repro the original matrix here. 
    //D3DXMatrixPerspectiveOffCenterLH( &newCam2NDC, oldLeft, oldRight, oldBot, oldTop, hither, yon );

    const float newRightMinusLeft = float(cWidth) / fbWidth * oldRightMinusLeft * duplication;
    const float newTopMinusBot = float(cHeight) / fbHeight * oldTopMinusBot * duplication;

    const float jitterX = jitter.x * newRightMinusLeft / cWidth;
    const float jitterY = jitter.y * newTopMinusBot / cHeight;

    const float centerX = oldLeft + float(gazePos.x) / fbWidth * oldRightMinusLeft + jitterX;
    const float centerY = oldBot + float(gazePos.y) / fbHeight * oldTopMinusBot + jitterY;

    const float newLeft = centerX - float(cWidth) * 0.5f / fbWidth * oldRightMinusLeft * duplication;
    const float newRight = centerX + float(cWidth) * 0.5f / fbWidth * oldRightMinusLeft * duplication;
    assert(AboutEqual(newRight - newLeft, newRightMinusLeft));

    const float newBot = centerY - float(cHeight) * 0.5f / fbHeight * oldTopMinusBot * duplication;
    const float newTop = centerY + float(cHeight) * 0.5f / fbHeight * oldTopMinusBot * duplication;
    assert(AboutEqual(newTop - newBot, newTopMinusBot));

    XMMATRIX newCam2NDC = XMMatrixPerspectiveOffCenterLH(
        newLeft, newRight,
        newBot, newTop,
        hither, yon);

    return newCam2NDC;
}

static XMMATRIX CreateBaseProjection(const XMMATRIX& xCam2NDC,
    //const float fbWidth, const float fbHeight,
    const float cWidth, const float cHeight,
    const float hither, const float /*yon*/,
    const XMFLOAT2& jitter)
{
    XMFLOAT4X4 cam2NDC;
    XMStoreFloat4x4(&cam2NDC, xCam2NDC);

    const float rightMinusLeft = 2.0f * hither / cam2NDC.m[0][0];
    //const float rightPlusLeft = -cam2NDC.m[2][0] / cam2NDC.m[0][0] * 2.0f * hither;
    //const float right = (rightPlusLeft + rightMinusLeft) * 0.5f;
    //const float left = (rightPlusLeft - rightMinusLeft) * 0.5f;

    const float topMinusBot = 2.0f * hither / cam2NDC.m[1][1];
    //const float topPlusBot = -cam2NDC.m[2][1] / cam2NDC.m[1][1] * 2.0f * hither;
    //const float top = (topPlusBot + topMinusBot) * 0.5f;
    //const float bot = (topPlusBot - topMinusBot) * 0.5f;

    const float jitterX = jitter.x * rightMinusLeft / cWidth;
    const float jitterY = jitter.y * topMinusBot / cHeight;

    // Could construct a new matrix with D3DXMatrixPerspectiveOffCenterLH by adding the
    // jitter to left, right, top, and bottom. But the result would be the same as
    // adding it into the skew channel of the projection matrix.
    // Note that r-l and t-b cancel out with the computations of jitterX/Y above, but
    // leaving in for clarity.
    XMFLOAT4X4 newCam2NDC = cam2NDC;
    newCam2NDC.m[2][0] += jitterX * 2.0f / -rightMinusLeft;
    newCam2NDC.m[2][1] += jitterY * 2.0f / -topMinusBot;

    // Test that recalculation of matrix would provide same result as incremental approach used above.
    //D3DXMATRIX tst;
    //D3DXMatrixPerspectiveOffCenterLH( &tst, left + jitterX, right + jitterX, bot + jitterY, top + jitterY, hither, yon);

    return XMLoadFloat4x4(&newCam2NDC);
}

bool Cascade::ShouldRefresh(float dt)
{
    if(m_fRefreshInterval <= 0)
    {
        return true;
    }
    if((m_fRefreshCountDown -= dt) <= 0)
    {
        m_fRefreshCountDown = m_fRefreshInterval - fmod(-m_fRefreshCountDown, m_fRefreshInterval);
        assert(m_fRefreshCountDown > 0);
        assert(m_fRefreshCountDown <= m_fRefreshInterval);
        return true;
    }
    return false;
}

XMMATRIX Cascade::Push(ID3D11DeviceContext* pd3dImmediateContext, const POINT& inGazePos, 
    const XMMATRIX& origCamToNDC, const FLOAT nearClip, const FLOAT farClip)
{
    // Move to new offscreen
    PingPong();

	UINT numViewports = 10;
	D3D11_VIEWPORT viewports[10];
	pd3dImmediateContext->RSGetViewports(&numViewports, viewports);
	float fbWidth = viewports[0].Width;
	float fbHeight = viewports[0].Height;

    POINT gazePos = InvertCursor(inGazePos, fbHeight);

    // Generate a new jitter amount. This is stored to be used later in the frame for resampling this offscreen.
    m_Jitter = NextJitterSample();

    XMMATRIX newCamToNDC;
    if(m_fDuplication)
    {
        newCamToNDC = CreateInsertProjection(origCamToNDC, 
            fbWidth, fbHeight, 
            float(m_aOffscreenRT[m_iPing]->GetWidth()), float(m_aOffscreenRT[m_iPing]->GetHeight()), 
            nearClip, farClip,
            m_Jitter,
            m_fDuplication,
            gazePos);
    }
    else
    {
        // The base cascade is handled a little differently, because it matches the backbuffer's 
        // aspect ratio (though not resolution) and is unaffected by changes in gaze position.
        // It is flagged by a zero duplication, because it's the only cascade for which duplication
        // would have no meaning.
        newCamToNDC = CreateBaseProjection(origCamToNDC,
            float(m_aOffscreenRT[m_iPing]->GetWidth()), float(m_aOffscreenRT[m_iPing]->GetHeight()), 
            nearClip, farClip,
            m_Jitter);
    }

	m_aOffscreenRT[m_iPing]->Push(pd3dImmediateContext);

    m_LastGazePos = inGazePos;

 //   ID3D11ShaderResourceView* previousDS = offscreenRT[!m_iPing]->DSShaderResource();
	//pd3dImmediateContext->PSSetShaderResources( 3, 1, &previousDS );
    ID3D11ShaderResourceView* previousColor = m_aOffscreenRT[!m_iPing]->RTShaderResource();
	pd3dImmediateContext->PSSetShaderResources( kPreviousColorSlot, 1, &previousColor );

    return newCamToNDC;
}

void Cascade::Clear(ID3D11DeviceContext* pd3dImmediateContext, const float * const color)
{
	m_aOffscreenRT[m_iPing]->Clear(pd3dImmediateContext, color);
}


void Cascade::Pop(ID3D11DeviceContext* pd3dImmediateContext)
{
	ID3D11ShaderResourceView* nullArray = NULL;
	//pd3dImmediateContext->PSSetShaderResources( 3, 1, &nullArray );
	pd3dImmediateContext->PSSetShaderResources( kPreviousColorSlot, 1, &nullArray );

    m_aOffscreenRT[m_iPing]->Pop(pd3dImmediateContext);
}

ID3D11ShaderResourceView* Cascade::GetTexture() const
{
	return m_aOffscreenRT[m_iPing]->RTShaderResource();
}

ID3D11ShaderResourceView* Cascade::GetPreviousTex() const
{
    return m_aOffscreenRT[!m_iPing]->RTShaderResource();
}

ID3D11ShaderResourceView* Cascade::GetPreviousDS() const
{
    return m_aOffscreenRT[!m_iPing]->DSShaderResource();
}

HRESULT Cascade::Save(const WCHAR* fileName, ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext)
{
    return m_aOffscreenRT[m_iPing]->Save(fileName, pd3dDevice, pd3dImmediateContext);
}

static float Halton(int idx, int base)
{
    float ret = 0;
    float denom = 1.0f / float(base);
    while(idx > 0)
    {
        ret += denom * float(idx % base);
        idx = idx / base;
        denom = denom / base;
    }
    return ret;
}
static void MakeJitterTable()
{
    for(int i = 0; i < _jitterSamples; ++i)
    {
        const float x = Halton(i, 2);
        const float y = Halton(i, 3);
        _jitterTable[i] = XMFLOAT2(x - 0.5f, y - 0.5f);
    }
}
XMFLOAT2 Cascade::NextJitterSample()
{
    if(++m_iJitterIndex >= _jitterSamples)
        m_iJitterIndex = 0;
    return m_bJitterEnabled
        ? XMFLOAT2(_jitterTable[m_iJitterIndex].x * m_fJitterScale, _jitterTable[m_iJitterIndex].y * m_fJitterScale)
        : XMFLOAT2(0.0f, 0.0f);
}
