#include "pch.h"

#include <windows.h>
#include <assert.h>
#include <d3d11.h>

#include "Utility.h"
#include "FauxV8Mgr.h"
#include "Cascade.h"
#include "BlitComposite.h"
#include "OffsetMap.h"
#include "BlitDepth.h"

FauxV8Mgr::FauxV8Mgr()
    :   m_iNumLayers(0),
        m_pBlitComposite(NULL),
        m_Cascades(NULL),
        m_aSize(NULL),
        m_aJitterScale(NULL),
        m_aMSRead(NULL),
        m_aFrameDecay(NULL),
        m_aRenderInterval(NULL),
        m_aRefreshSecs(NULL),
        m_aBlitQuads(NULL),
        m_iMSAA(0),
        m_iMSAAInner(0),
        m_fCompositeBlendStart(0.6f),
        m_bUseMiddle(true),
        m_dbgDiscards(0),
        m_DbgReprojEnable(true),
        m_fAspect(1.0f),
        m_bFrameOdd(false),
        m_pSamplePrevious(NULL),
#ifdef MAF_RANDOM_TEXTURE
        m_pOffsetMap(NULL),
        m_iOffsetBase(0),
#endif // MAF_RANDOM_TEXTURE
        m_pReprojectionCB(NULL)
{
    Cascade::Init();
    SetupDefaults();
}

FauxV8Mgr::~FauxV8Mgr()
{
    Release();
    ReleaseArrays();
}

void FauxV8Mgr::Init( ID3D11Device* pd3dDevice )
{
    InitCascades(pd3dDevice);
    InitBlitComposite(pd3dDevice);
    InitBlitQuads(pd3dDevice);
    InitOffsetMap(pd3dDevice);
    InitSamplers(pd3dDevice);
    InitConstantBuffers(pd3dDevice);
    InitRenderIntervals();
}

// free all owned resources, making ready for shutdown or rebuilding with different parameters.
void FauxV8Mgr::Release()
{
    SAFE_RELEASE(m_pSamplePrevious);
    SAFE_RELEASE(m_pReprojectionCB);
    
#ifdef MAF_RANDOM_TEXTURE
    SAFE_DELETE(m_pOffsetMap);
#endif // MAF_RANDOM_TEXTURE

    SAFE_DELETE(m_pBlitComposite);

    for(int i = 0; i < m_iNumLayers; ++i)
    {
        SAFE_DELETE(m_Cascades[i]);

        SAFE_DELETE(m_aBlitQuads[i]);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
FauxV8Mgr::eRenderInterval operator&( const FauxV8Mgr::eRenderInterval& lhs, int rhs ) 
{
    return FauxV8Mgr::eRenderInterval(int(lhs) & rhs);
}
FauxV8Mgr::eRenderInterval operator|( const FauxV8Mgr::eRenderInterval& lhs, const FauxV8Mgr::eRenderInterval& rhs )
{
    return FauxV8Mgr::eRenderInterval(int(lhs) | int(rhs));
}
bool FauxV8Mgr::ShouldRefresh( int layer, float dt )
{
    if((layer == 1) && !IGetMiddleActive())
        return false;

    const bool onFrame = (m_aRenderInterval[layer] & ((m_bFrameOdd ? kRenderOdd : kRenderEven) | kRenderInit)) != kRenderNever;
    if(onFrame)
    {
        const bool onTime = m_Cascades[layer]->ShouldRefresh(dt);
        return onTime;
    }
    return false;
}

XMMATRIX FauxV8Mgr::PushLayer( int layer, ID3D11DeviceContext* pd3dImmediateContext )
{
    m_iCurrentLayer = layer;

    XMMATRIX camToNDC = m_Cascades[layer]->Push(pd3dImmediateContext, 
                        m_FoveaCenter, 
                        m_CameraInfo.m_mCameraToNDC, 
                        m_CameraInfo.m_fNearClip, m_CameraInfo.m_fFarClip);

    m_mLastWorldToCamera = m_Cascades[layer]->GetLastWorldToCamera();
    m_mLastCameraToNDC = m_Cascades[layer]->GetLastCameraToNDC();
    m_mLastWorldToNDC = m_Cascades[layer]->GetLastWorldToNDC();
    m_Cascades[layer]->SetLastWorldToCamera(m_CameraInfo.m_mWorldToCamera);
    m_Cascades[layer]->SetLastCameraToNDC(camToNDC);

    PushConstants( layer, pd3dImmediateContext );

    return camToNDC;
}

void FauxV8Mgr::Clear( ID3D11DeviceContext* pd3dImmediateContext, const float* const pColor )
{

    m_Cascades[m_iCurrentLayer]->Clear(pd3dImmediateContext, pColor);

#if 1
    if(m_aBlitQuads[m_iCurrentLayer])
    {
        if(m_iCurrentLayer)
        {
            if(m_iCurrentLayer < m_iNumLayers - 1)
            {
                // middle
                const float radius = float(m_Cascades[m_iCurrentLayer-1]->GetWidth())
                    / m_Cascades[m_iCurrentLayer]->GetWidth()
                    * m_Cascades[m_iCurrentLayer-1]->GetDuplication()
                    / m_Cascades[m_iCurrentLayer]->GetDuplication()
                    * m_fCompositeBlendStart;
                m_aBlitQuads[m_iCurrentLayer]->SetRadiusSquared(radius * radius);
            }
            else
            {
                // base
                const float radius = float(m_Cascades[m_iCurrentLayer-1]->GetWidth())
                    * m_Cascades[m_iCurrentLayer-1]->GetDuplication()
                    * m_fCompositeBlendStart 
                    * 0.5f;
                m_aBlitQuads[m_iCurrentLayer]->SetRadiusSquared(radius * radius);

                const POINT dims = m_pBlitComposite->GetDstResolution();
                const float fbWidth = float(dims.x);
                const float fbHeight = float(dims.y);

                const XMFLOAT2 center = XMFLOAT2(
                    float(m_FoveaCenter.x), 
                    (fbHeight - float(m_FoveaCenter.y) - 1.0f));
                m_aBlitQuads[m_iCurrentLayer]->SetCenter(center);
            
                const XMFLOAT2 size = XMFLOAT2(fbWidth, fbHeight);
                m_aBlitQuads[m_iCurrentLayer]->SetSize(size);
            }
        }
        m_aBlitQuads[m_iCurrentLayer]->Render(pd3dImmediateContext, NULL, NULL);
    }
#endif
}

void FauxV8Mgr::PopLayer( ID3D11DeviceContext* pd3dImmediateContext )
{
    if(GetDebugDiscardMask() & (1 << m_iCurrentLayer))
    {
#define MAF_DOT_HACK
#ifndef MAF_DOT_HACK
        const float black[] = { 0.0f, 0.0f, 0.0f, 1.0f };
#else
        const float black[] = { 1.0f, 1.0f, 1.0f, 1.0f };
#endif
        Clear(pd3dImmediateContext, black);
    }
    m_Cascades[m_iCurrentLayer]->Pop(pd3dImmediateContext);
    m_aRenderInterval[m_iCurrentLayer] = m_aRenderInterval[m_iCurrentLayer] & ~kRenderInit;
}

void FauxV8Mgr::FinishFrame( ID3D11DeviceContext* pd3dImmediateContext, ID3D11RenderTargetView* pRTV )
{
	UINT numViewports = 10;
	D3D11_VIEWPORT viewports[10];
	pd3dImmediateContext->RSGetViewports(&numViewports, viewports);
	const LONG width = (LONG)(viewports[0].Width + 0.5f);
	const LONG height = (LONG)(viewports[0].Height + 0.5f);
	const POINT dims = { width, height };

    BlitComposite* bq = m_pBlitComposite;

	bq->SetDstResolution(dims);
	const BlitQuad::BlitRect dstRect( XMFLOAT2(0.0f, 0.0f), XMFLOAT2(1.0f, 1.0f) );
	bq->SetDstRect(dstRect);

	ID3D11ShaderResourceView* texList[BlitComposite::kMaxLayers] = { NULL };
	BlitQuad::BlitRect rectList[BlitComposite::kMaxLayers];

    XMFLOAT2 srcSizes[BlitComposite::kMaxLayers];

    XMFLOAT2 jitterOffsets[BlitComposite::kMaxLayers];
	for(int i = m_iNumLayers - 2; i >= 0; --i)
	{
		if(m_Cascades[i])
		{
			const float duplication = m_Cascades[i]->GetDuplication();

			const float blitWidth = m_aSize[i] / viewports[0].Width * duplication;
			const float blitHeight = m_aSize[i] / viewports[0].Height * duplication;
            const POINT& gazePos = m_Cascades[i]->GetGazePos();
			const XMFLOAT2 center(gazePos.x / (float)width, 1.0f - gazePos.y / (float)height);
			const XMFLOAT2 dstMin(center.x - blitWidth * 0.5f, center.y - blitHeight * 0.5f);
			const XMFLOAT2 dstMax(center.x + blitWidth * 0.5f, center.y + blitHeight * 0.5f);

			const BlitQuad::BlitRect dstRect( dstMin, dstMax );
			rectList[i] = dstRect;
			texList[i] = m_Cascades[i]->GetTexture();

            srcSizes[i] = XMFLOAT2(float(m_aSize[i]), float(m_aSize[i]));

            jitterOffsets[i] = m_Cascades[i]->GetJitter();
		}
	}
    {
        const int base = m_iNumLayers - 1;
        const Cascade* cascade = m_Cascades[base];
        texList[BlitComposite::kBaseLayer] = cascade->GetTexture();
        srcSizes[BlitComposite::kBaseLayer] = XMFLOAT2(
            float(cascade->GetWidth()), float(cascade->GetHeight()));
        jitterOffsets[BlitComposite::kBaseLayer] = cascade->GetJitter();
    }

	bq->SetDstRects(rectList, m_iNumLayers - 1);
	bq->SetTextures(texList, srcSizes);
    bq->SetJitterOffsets(jitterOffsets);

	bq->Render(pd3dImmediateContext, pRTV, NULL);

    m_bFrameOdd = !m_bFrameOdd;
}

void FauxV8Mgr::SetFBWidthOverHeight(float aspect)
{
    m_fAspect = aspect;
    Release();
}

void FauxV8Mgr::SetFBWidthAndHeight( UINT width, UINT height )
{
    SetFBWidthOverHeight(float(width) / float(height));
}

POINT FauxV8Mgr::GetResolution(int layer) const
{
    if(layer < m_iNumLayers - 1)
    {
        if((layer == 1) && !IGetMiddleActive())
        {
            const POINT ret = { 0, 0 };
            return ret;
        }
        else
        {
            const POINT ret = { m_aSize[layer], m_aSize[layer] };
            return ret;
        }
    }
    assert(layer == m_iNumLayers-1);
    const POINT baseDim = IComputeBaseDims(m_fAspect, m_aSize[layer]);
    return baseDim;
}
////////////////////////////////////////////////////////////////////
// CameraInfo helper class
void FauxV8Mgr::CameraInfo::Save( const XMMATRIX& w2c, const XMMATRIX& c2ndc, const float nearClip, const float farClip )
{
    m_mWorldToCamera = w2c;
    m_mCameraToNDC = c2ndc;
    m_fNearClip = nearClip;
    m_fFarClip = farClip;
}

void FauxV8Mgr::SetCamera( const XMMATRIX& worldToCamera, const XMMATRIX& cameraToNDC, float nearClip, float farClip )
{
    m_CameraInfo.Save(worldToCamera, cameraToNDC, nearClip, farClip);
}
void FauxV8Mgr::SetFoveaCenter( const POINT& pixPos )
{
    m_FoveaCenter = pixPos;
}

void FauxV8Mgr::PushConstants(int layer, ID3D11DeviceContext* pd3dImmediateContext)
{
    const UINT kPreviousTextureSlot = 15;
    pd3dImmediateContext->PSSetSamplers( kPreviousTextureSlot, 1, &m_pSamplePrevious);

#ifdef MAF_RANDOM_TEXTURE
    const UINT kRandomShaderSlot = 11;
    m_pOffsetMap->Set(pd3dImmediateContext, kRandomShaderSlot);
#endif // MAF_RANDOM_TEXTURE

    const float frameDecay = m_DbgReprojEnable ? m_aFrameDecay[layer] : 0.0f;
    ReprojectionBlock cb;
    cb.m_Line0 = XMLoadFloat4(&XMFLOAT4(
        frameDecay,
        0.0f,
        0.0f,
        0.0f));

#ifdef MAF_RANDOM_TEXTURE
    const POINT fbSize = GetResolution(layer);
    const float tableSize = float(m_pOffsetMap->GetLength());
    cb.m_Line1 = XMLoadFloat4(&XMFLOAT4(
        float(fbSize.x),
        float(fbSize.y),
        1.0f / tableSize,
        0.5f / tableSize));
#else //  MAF_RANDOM_TEXTURE
    cb.m_Line1 = XMLoadFloat4(&XMFLOAT4(0.0f, 0.0f, 0.0f, 0.0f));
#endif // MAF_RANDOM_TEXTURE
    pd3dImmediateContext->UpdateSubresource( m_pReprojectionCB, 0, NULL, &cb, 0, 0 );

    const UINT kConstantBufferSlot = 10;
    pd3dImmediateContext->VSSetConstantBuffers( kConstantBufferSlot, 1, &m_pReprojectionCB );
    pd3dImmediateContext->PSSetConstantBuffers( kConstantBufferSlot, 1, &m_pReprojectionCB );

}
void FauxV8Mgr::DebugDisableReproj(ID3D11DeviceContext* pd3dImmediateContext)
{
    ReprojectionBlock cb;
    ZeroMemory(&cb, sizeof(cb));
    pd3dImmediateContext->UpdateSubresource( m_pReprojectionCB, 0, NULL, &cb, 0, 0 );
    const UINT kConstantBufferSlot = 10;
    pd3dImmediateContext->PSSetConstantBuffers( kConstantBufferSlot, 1, &m_pReprojectionCB );
}
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// Bookkeeping helpers. Leaving the number of layers a settable parameter
// lends a feature that will probably be useless in the mature system.
// But it will be useful until the optimal number of layers is determined.
void FauxV8Mgr::AllocateArrays()
{
    ReleaseArrays();
    m_Cascades = new Cascade* [m_iNumLayers];
    ZeroMemory(m_Cascades, m_iNumLayers * sizeof(*m_Cascades));
    m_aSize = new int[m_iNumLayers];
    m_aJitterScale = new float[m_iNumLayers];
    m_aMSRead = new bool[m_iNumLayers];
    m_aFrameDecay = new float[m_iNumLayers];
    m_aRenderInterval = new eRenderInterval[m_iNumLayers];
    ZeroMemory(m_aRenderInterval, m_iNumLayers * sizeof(*m_aRenderInterval));
    m_aRefreshSecs = new float[m_iNumLayers];
    ZeroMemory(m_aRefreshSecs, m_iNumLayers * sizeof(*m_aRefreshSecs));

    m_aBlitQuads = new BlitDepth* [m_iNumLayers];
    ZeroMemory(m_aBlitQuads, m_iNumLayers * sizeof(*m_aBlitQuads));
}
void FauxV8Mgr::ReleaseArrays()
{
    SAFE_DELETE_ARRAY(m_Cascades);
    SAFE_DELETE_ARRAY(m_aSize);
    SAFE_DELETE_ARRAY(m_aJitterScale);
    SAFE_DELETE_ARRAY(m_aMSRead);
    SAFE_DELETE_ARRAY(m_aFrameDecay);
    SAFE_DELETE_ARRAY(m_aRenderInterval);
    SAFE_DELETE_ARRAY(m_aRefreshSecs);
    SAFE_DELETE_ARRAY(m_aBlitQuads);
}

HRESULT FauxV8Mgr::InitCascades( ID3D11Device* pd3dDevice )
{
    HRESULT hr = S_OK;

    const int numInner = m_iNumLayers - 1;
    for(int i = 0; i < numInner; ++i)
    {
        if(m_aSize[i])
        {
		    m_Cascades[i] = new Cascade();
            const int msaa = i ? m_iMSAA : m_iMSAAInner;
            const int dup = 1 << i;
		    V_RETURN( m_Cascades[i]->Create(pd3dDevice, m_aSize[i], m_aSize[i], msaa, float(dup)) );
            const float jitter = m_aJitterScale[i];
            m_Cascades[i]->SetJitterScale(jitter);
            m_Cascades[i]->SetRefreshInterval(m_aRefreshSecs[i]);
        }
    }
    const int outerLayer = numInner;
    m_Cascades[outerLayer] = new Cascade();
    const POINT baseDim = IComputeBaseDims(m_fAspect, m_aSize[outerLayer]);
    V_RETURN( m_Cascades[outerLayer]->CreateBase(pd3dDevice, baseDim.x, baseDim.y, m_iMSAA) );
    m_Cascades[outerLayer]->SetJitterScale(m_aJitterScale[outerLayer]);
    m_Cascades[outerLayer]->SetRefreshInterval(m_aRefreshSecs[outerLayer]);

    return hr;
}
HRESULT FauxV8Mgr::InitBlitComposite( ID3D11Device* pd3dDevice )
{
    m_pBlitComposite = new BlitComposite();
    for(int i = 0; i < m_iNumLayers; ++i)
    {
        m_pBlitComposite->SetMSAA(i, m_aMSRead[i]);
    }

    const int numBlitLayers = IGetMiddleActive() ? m_iNumLayers - 1 : (m_iNumLayers > 1 ? m_iNumLayers - 2 : 0);
    m_pBlitComposite->SetNumLayers(numBlitLayers);
    m_pBlitComposite->SetDecayStart(m_fCompositeBlendStart);

    return m_pBlitComposite->Init(pd3dDevice, L"BlitComposite.fx");
}
HRESULT FauxV8Mgr::InitBlitQuads(ID3D11Device* pd3dDevice)
{
    HRESULT hr = S_OK;

#if 0
    m_aBlitQuads[0] = new BlitDepth();
    const POINT dstRes0 = { m_Cascades[0]->GetWidth(), m_Cascades[0]->GetHeight() };
    m_aBlitQuads[0]->SetDstResolution(dstRes0);
    V_RETURN(m_aBlitQuads[0]->Init(pd3dDevice, L"BlitDepth.fx", L"Inner"));

    for(int i = 1; i < m_iNumLayers - 1; ++i)
    {
        m_aBlitQuads[i] = new BlitDepth();
        const POINT dstRes = { m_Cascades[i]->GetWidth(), m_Cascades[i]->GetHeight() };
        m_aBlitQuads[i]->SetDstResolution(dstRes);
        V_RETURN( m_aBlitQuads[i]->Init(pd3dDevice, L"BlitDepth.fx", L"Middle") );
    }

    m_aBlitQuads[m_iNumLayers - 1] = new BlitDepth();
    const POINT dstResN = { m_Cascades[m_iNumLayers - 1]->GetWidth(), m_Cascades[m_iNumLayers - 1]->GetHeight() };
    m_aBlitQuads[m_iNumLayers - 1]->SetDstResolution(dstResN);
    V_RETURN( m_aBlitQuads[m_iNumLayers - 1]->Init(pd3dDevice, L"BlitDepth.fx", L"Base") );
#endif

    return hr;
}
HRESULT FauxV8Mgr::InitOffsetMap(ID3D11Device* pd3dDevice)
{
#ifdef MAF_RANDOM_TEXTURE
    assert(!m_pOffsetMap);
    m_pOffsetMap = new OffsetMap();
    const int kOffsetMapLen = 256;
    return m_pOffsetMap->Create(pd3dDevice, kOffsetMapLen);
#else // MAF_RANDOM_TEXTURE
    return S_OK;
#endif // MAF_RANDOM_TEXTURE
}
HRESULT FauxV8Mgr::InitSamplers( ID3D11Device* pd3dDevice )
{
    assert(!m_pSamplePrevious);

    D3D11_SAMPLER_DESC sampDesc;
    ZeroMemory( &sampDesc, sizeof(sampDesc) );
    sampDesc.Filter = D3D11_FILTER_MIN_MAG_MIP_LINEAR;
    sampDesc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
    sampDesc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
    sampDesc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
    sampDesc.ComparisonFunc = D3D11_COMPARISON_NEVER;
    sampDesc.MinLOD = 0;
    sampDesc.MaxLOD = 0;
    HRESULT hr = pd3dDevice->CreateSamplerState( &sampDesc, &m_pSamplePrevious );
    
    assert(SUCCEEDED(hr));

    return hr;
}
HRESULT FauxV8Mgr::InitConstantBuffers( ID3D11Device* pd3dDevice )
{
    HRESULT hr = S_OK;
    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    assert(!m_pReprojectionCB);
    if(!m_pReprojectionCB)
    {
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.CPUAccessFlags = 0;

        bd.ByteWidth = sizeof(ReprojectionBlock);
        hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pReprojectionCB);
        if( FAILED( hr ) )
            return hr;
    }

    return hr;
}
////////////////////////////////////////////////////////////////////
// SetupDefaults will set the system to its recommended parameters. In the mature
// system, none of the other setters are likely to ever need to be called, and
// the defaults will likely be optimal for any setting. They are included here
// for fine tuning the system to obtain the appropriate defaults.
// These must be called before Init(), or else Init() must be called again
// between calling any of these setter and the next Render().
void FauxV8Mgr::SetupDefaults()
{
    m_iNumLayers = 3;
    AllocateArrays();

    ISetSize(0, 188);
    ISetSize(1, 333);
    ISetSize(2, 383);

    // MAF_DOT_HACK
    //ISetSize(0, 64);
    //ISetSize(1, 133);
    //ISetSize(2, 183);

	//ISetSize(0, 229);
 //   ISetSize(1, 431);
 //   ISetSize(2, 1082);

    ISetJitterScale(0, 1.0f);
    ISetJitterScale(1, 1.0f);
    ISetJitterScale(2, 1.0f);

    ISetMSRead(0, false);
    ISetMSRead(1, false);
    ISetMSRead(2, false);

    //ISetFrameDecay(0, 0.3f);
    //ISetFrameDecay(1, 0.47f);
    //ISetFrameDecay(2, 0.7f);
    ISetFrameDecay(0, 0.3f);
    ISetFrameDecay(1, 0.47f);
    ISetFrameDecay(2, 0.7f);

    ISetRenderInterval(0, kRenderAlways);
    ISetRenderInterval(1, kRenderEven);
    ISetRenderInterval(2, kRenderOdd);
    InitRenderIntervals();

    m_iMSAA = 4;
    m_iMSAAInner = 4;

    m_fCompositeBlendStart = 0.6f;
    m_bUseMiddle = true;
}

void FauxV8Mgr::InitRenderIntervals()
{
    for(int i = 0; i < m_iNumLayers; ++i)
    {
        ISetRenderInterval(i, m_aRenderInterval[i] | kRenderInit);
    }
}

bool FauxV8Mgr::NeedInit() const
{
    return !m_pBlitComposite;
}

int FauxV8Mgr::IGetSize( int layer ) const
{
    return m_aSize[layer];
}

float FauxV8Mgr::IGetFrameDecay( int layer ) const
{
    return m_aFrameDecay[layer];
}

void FauxV8Mgr::ISetNumLayers( int n )
{
    Release();
    m_iNumLayers = n;
    AllocateArrays();
}

void FauxV8Mgr::ISetSize( int layer, int res )
{
    Release();
    assert(m_aSize);
    m_aSize[layer] = res;
}

void FauxV8Mgr::ISetJitterScale( int layer, float jitter )
{
    m_aJitterScale[layer] = jitter;
    if(m_Cascades[layer])
    {
        m_Cascades[layer]->SetJitterScale(jitter);
    }
}

void FauxV8Mgr::ISetMSAA( int msaa )
{
    Release();
    m_iMSAA = msaa;
}

void FauxV8Mgr::ISetMSAAInner( int msaa )
{
    Release();
    m_iMSAAInner = msaa;
}

void FauxV8Mgr::ISetMSRead( int layer, bool b )
{
    Release();
    m_aMSRead[layer] = b;
}

void FauxV8Mgr::ISetFrameDecay( int layer, float f )
{
    m_aFrameDecay[layer] = f;
}

void FauxV8Mgr::ISetCompositeBlendStart( float f )
{
    m_fCompositeBlendStart = f;
    if(m_pBlitComposite)
        m_pBlitComposite->SetDecayStart(m_fCompositeBlendStart);
}

POINT FauxV8Mgr::IComputeBaseDims(const float fWidthOverHeight, const int iMaxDim) const
{
    int width = 0;
    int height = 0;
    if(fWidthOverHeight > 1.0f)
    {
        width = iMaxDim;
        height = int(width / fWidthOverHeight + 0.5f);
    }
    else
    {
        height = iMaxDim;
        width = int(height * fWidthOverHeight + 0.5f);
    }
    POINT ret = { width, height };
    return ret;
}

void FauxV8Mgr::ISetRefreshSecs( int layer, float dt )
{
    m_aRefreshSecs[layer] = dt;
    if(m_Cascades[layer])
    {
        m_Cascades[layer]->SetRefreshInterval(dt);
    }
}

void FauxV8Mgr::ISetRenderInterval( int layer, eRenderInterval e )
{
    m_aRenderInterval[layer] = e | kRenderInit;
}

void FauxV8Mgr::ISetUseMiddle( bool b )
{
    m_bUseMiddle = b;
    Release();
}

struct PerspectiveSides
{
    float   m_Left;
    float   m_Right;
    float   m_Bottom;
    float   m_Top;
    float   m_Near;
    float   m_Far;
};
static PerspectiveSides DecomposeToSides(const XMMATRIX& src, const float hither, const float yon)
{
    XMFLOAT4X4 cam2NDC;
    XMStoreFloat4x4(&cam2NDC, src);

    PerspectiveSides ret;
    ret.m_Near = hither;
    ret.m_Far = yon;

    const float rightPlusLeft = -cam2NDC.m[2][0] / cam2NDC.m[0][0] * 2.0f * hither;
    const float rightMinusLeft = 2.0f * hither / cam2NDC.m[0][0];
    ret.m_Left = (rightPlusLeft - rightMinusLeft) * 0.5f;
    ret.m_Right = (rightPlusLeft + rightMinusLeft) * 0.5f; 

    const float topPlusBot = -cam2NDC.m[2][1] / cam2NDC.m[1][1] * 2.0f * hither;
    const float topMinusBot = 2.0f * hither / cam2NDC.m[1][1];
    ret.m_Bottom = (topPlusBot - topMinusBot) * 0.5f;
    ret.m_Top = (topPlusBot + topMinusBot) * 0.5f;

    return ret;
}

static XMMATRIX GenPerspective(const PerspectiveSides& sides)
{
    const XMMATRIX cam2NDC = XMMatrixPerspectiveOffCenterLH( 
        sides.m_Left, 
        sides.m_Right, 
        sides.m_Bottom, 
        sides.m_Top, 
        sides.m_Near, 
        sides.m_Far );

    return cam2NDC;
}
bool FauxV8Mgr::IntersectCull( XMMATRIX& isect, const XMMATRIX& lhs, const XMMATRIX& rhs, float nearClip, float farClip )
{
    const PerspectiveSides leftSides = DecomposeToSides(lhs, nearClip, farClip);
    const PerspectiveSides rightSides = DecomposeToSides(rhs, nearClip, farClip);
    PerspectiveSides isectSides;
    isectSides.m_Near = nearClip;
    isectSides.m_Far = farClip;

    isectSides.m_Left = __max(leftSides.m_Left, rightSides.m_Left);
    isectSides.m_Right = __min(leftSides.m_Right, rightSides.m_Right);
    if(isectSides.m_Right <= isectSides.m_Left)
        return false;

    isectSides.m_Bottom = __max(leftSides.m_Bottom, rightSides.m_Bottom);
    isectSides.m_Top = __min(leftSides.m_Top, rightSides.m_Top);
    if(isectSides.m_Top <= isectSides.m_Bottom)
        return false;

    isect = GenPerspective(isectSides);

    return true;
}

