
#ifndef FAUX_V8_MGR_H
#define FAUX_V8_MGR_H

class Cascade;
class BlitComposite;
class BlitDepth;
class OffsetMap;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// See the sample code at the bottom of this file.
//////////////////////////////////////////////////////////////////////////

class FauxV8Mgr
{
public:
    enum eRenderInterval
    {
        kRenderNever = 0x0,
        kRenderEven = 0x1,
        kRenderOdd  = 0x2,
        kRenderAlways = kRenderEven | kRenderOdd,
        kRenderInit = 0x1000
    };

    ////////////////////////////////////////////////////////////
    // Creating and teardown.
    FauxV8Mgr();

    virtual ~FauxV8Mgr();

    // create all necessary graphics resources.
    void Init(ID3D11Device* pd3dDevice);

    // free all owned resources, making ready for shutdown or rebuilding with different parameters.
    void Release();
    ////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////
    // Setup to render another frame. First set parameters constant for the frame.
    //
    //  set the camera parameters for this render.
    void SetCamera(const XMMATRIX& worldToCamera, const XMMATRIX& cameraToNDC, float nearClip, float farClip);
    //
    //  set the position in window pixel coordinates.
    void SetFoveaCenter(const POINT& pixPos);
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // This section covers the render to each layer.
    //
    // get the number of layers to loop over using PushLayer/PopLayer.
    int GetNumLayers() const { return m_iNumLayers; }
    //
    // should this layer be refreshed this frame? if not, skip out of the PushLayer/PopLayer and rendering.
    bool ShouldRefresh(int layer, float dt);
    //
    //  setup rendering state for rendering to an eccentricity layer. returns the modified projection
    //  transform which should replace the global transform passed into SetCamera() in any external code,
    //  especially including culling.
    XMMATRIX PushLayer(int layer, ID3D11DeviceContext* pd3dImmediateContext);
    //
    //  helper to clear the color and depth buffer. They are set (after PushLayer), so you could just get the
    //  current color and depth buffers and clear them yourself.
    void Clear(ID3D11DeviceContext* pd3dImmediateContext, const float* const pColor);
    //
    //  get the last WorldToNDC matrix for the current layer. only valid between PushLayer/PopLayer.
    //  this matrix is used for the reprojection, concatenated with each object's previous local to world to
    //  get the previous frame's local to NDC. LastWorldToNDC == LastWorldToCamera * LastCameraToNDC
    const XMMATRIX& GetLastWorldToNDC() const { return m_mLastWorldToNDC; }
    //  get the last WorldToCamera matrix for the current layer. only valid between PushLayer/PopLayer.
    //  this matrix can be used for the reprojection. LastWorldToNDC == LastWorldToCamera * LastCameraToNDC
    const XMMATRIX& GetLastWorldToCamera() const { return m_mLastWorldToCamera; }
    //  get the last CameraToNDC matrix for the current layer. only valid between PushLayer/PopLayer.
    //  this matrix can be used for the reprojection. LastWorldToNDC == LastWorldToCamera * LastCameraToNDC
    const XMMATRIX& GetLastCameraToNDC() const { return m_mLastCameraToNDC; }
    //
    ////////// Render your scene here. //////////
    //
    //  pop current layer state after rendering to it.
    void PopLayer(ID3D11DeviceContext* pd3dImmediateContext);
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // and finally perform the composite of layers into the indicated render target.
    void FinishFrame(ID3D11DeviceContext* pd3dImmediateContext, ID3D11RenderTargetView* pRTV);
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////////////////////////////
    // Generate a camera to NDC matrix that is the intersection of two input matrices,
    // to be used for generating a cull transform. This should be moved to a helper.
    static bool IntersectCull(XMMATRIX& isect, const XMMATRIX& lhs, const XMMATRIX& rhs, float nearClip, float farClip);
    ////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////
    // Set parameters governing the creation and subsequent behavior of the system. Changing
    // most of these parameters will require Init() to be called again before the next render.
    // Check NeedInit() to be sure. You can change any number of parameters as long as at least
    // one call to Init() is made before next render iteration.
    //
    // Return whether parameters have changed requiring another call to Init().
    bool NeedInit() const;
    
    // Return the resolution of the current layer. Only valid between PushLayer and PopLayer.
    POINT GetResolution(int layer) const;

    // Return which layer we are on. The app should know this, because it has told us in Push,
    // but we know here and it's convenient, so no need to be a jerk about it.
    // Obviously, only valid between PushLayer and PopLayer.
    int GetCurrentLayer() const { return m_iCurrentLayer; }

    void SetFBWidthOverHeight(float aspect);
    void SetFBWidthAndHeight(UINT width, UINT height);

    void SetDebugDiscardMask(int m) { m_dbgDiscards = m; }
    int GetDebugDiscardMask() const { return m_dbgDiscards; }

    void SetDebugReprojEnable(bool b) { m_DbgReprojEnable = b; }
    bool GetDebugReprojEnable() const { return m_DbgReprojEnable; }

    void DebugDisableReproj(ID3D11DeviceContext* pd3dImmediateContext);

protected:

    void PushConstants(int layer, ID3D11DeviceContext* pd3dImmediateContext);

    // set the number of eccentricity layers. Must call Init() again after
    void ISetNumLayers(int n);
    int IGetNumLayers() const { return m_iNumLayers; }
    //
    // set the size of an eccentricity layer
    void ISetSize(int layer, int res);
    int IGetSize(int layer) const;
    //
    void ISetJitterScale(int layer, float jitter);
    float IGetJitterScale(int layer) const { return m_aJitterScale[layer]; }
    //
    void ISetMSRead(int layer, bool b);
    bool IGetMSRead(int layer) const { return m_aMSRead[layer]; }
    //
    void ISetFrameDecay(int layer, float f);
    float IGetFrameDecay(int layer) const;
    //
    void ISetRenderInterval(int layer, eRenderInterval e);
    eRenderInterval IGetRenderInterval(int layer) const { return m_aRenderInterval[layer]; }

    void ISetRefreshSecs(int layer, float dt);
    float IGetRefreshSecs(int layer) const { return m_aRefreshSecs[layer]; }
    //
    void ISetMSAA(int msaa);
    int IGetMSAA() const { return m_iMSAA; }
    
    void ISetMSAAInner(int b);
    int IGetMSAAInner() const { return m_iMSAAInner; }

    void ISetUseMiddle(bool b);
    bool IGetUseMiddle() const { return m_bUseMiddle; }

    bool IGetMiddleActive() const { return IGetUseMiddle() && (m_aSize[1] > 0); }

    //
    void ISetCompositeBlendStart(float f);
    float IGetCompositeBlendStart() const { return m_fCompositeBlendStart; }

    POINT IComputeBaseDims(const float fWidthOverHeight, const int iMaxDim) const;

    Cascade* IGetCascade(int layer) const { return m_Cascades[layer]; }

private:
    FauxV8Mgr(const FauxV8Mgr&)
    {
        assert(false);
    }
    FauxV8Mgr& operator=(const FauxV8Mgr&)
    {
        assert(false);
        return *this;
    }

    //
    void SetupDefaults();
    void AllocateArrays();
    void ReleaseArrays();

    void InitRenderIntervals();
    HRESULT InitCascades(ID3D11Device* pd3dDevice);
    HRESULT InitBlitComposite(ID3D11Device* pd3dDevice);
    HRESULT InitBlitQuads(ID3D11Device* pd3dDevice);
    HRESULT InitOffsetMap(ID3D11Device* pd3dDevice);
    HRESULT InitSamplers(ID3D11Device* pd3dDevice);
    HRESULT InitConstantBuffers(ID3D11Device* pd3dDevice);

    bool ShouldRender(int layer);
    ////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////
    // Temporary state, valid for the render of a frame.
    POINT           m_FoveaCenter;
    struct CameraInfo
    {
        XMMATRIX      m_mWorldToCamera;
        XMMATRIX      m_mCameraToNDC;
        float           m_fNearClip;
        float           m_fFarClip;

        void Save(const XMMATRIX& w2c, const XMMATRIX& c2ndc, const float nearClip, const float farClip);
    };
    CameraInfo m_CameraInfo;

    struct ReprojectionBlock
    {
        XMVECTOR     m_Line0; // blend, jitter scale, jitter table offset, unused
        XMVECTOR     m_Line1; // width, height, 0.5 / width, 0.5 / height
    };

    ////////////////////////////////////////////////////////////////////
    // Temporary state, valid for the render of a single layer
    int             m_iCurrentLayer;
    XMMATRIX      m_mLastWorldToNDC;
    XMMATRIX      m_mLastWorldToCamera;
    XMMATRIX      m_mLastCameraToNDC;
    ////////////////////////////////////////////////////////////////////

    int             m_iNumLayers;
    BlitComposite*  m_pBlitComposite;
    
#ifdef MAF_RANDOM_TEXTURE
    OffsetMap*      m_pOffsetMap;
    int             m_iOffsetBase;
#endif // MAF_RANDOM_TEXTURE

    Cascade**           m_Cascades;
    int*                m_aSize;
    float*              m_aJitterScale;
    bool*               m_aMSRead;
    float*              m_aFrameDecay;
    eRenderInterval*    m_aRenderInterval;
    float*              m_aRefreshSecs;
    BlitDepth**         m_aBlitQuads;

    int             m_iMSAA;
    int             m_iMSAAInner;
    float           m_fCompositeBlendStart;
    bool            m_bUseMiddle;
    
    int             m_dbgDiscards;
    bool            m_DbgReprojEnable;

    float           m_fAspect;

    bool            m_bFrameOdd;

    ID3D11SamplerState* m_pSamplePrevious;
    ID3D11Buffer*       m_pReprojectionCB;

};

//////////////////////////////////////////////////////////////////////////
#ifdef NEVER_DEFINE_SAMPLE_CODE_ONLY
// This is working sample code. It's based on the DXUT system, but doesn't rely on it.
// It only requires the same camera info that any rendering engine will have 
// and gaze position that any eye-tracker will give.
{

    // Send this frame's camera information.
    fv8Mgr.SetCamera(*g_Camera.GetViewMatrix(), *g_Camera.GetProjMatrix(), g_Camera.GetNearClip(), g_Camera.GetFarClip());
    // Send the gaze position within the window.
    fv8Mgr.SetFoveaCenter(GetGazePosition());

    // Loop over the eccentricity layers.
    for(int iLayer = 0; iLayer < fv8Mgr.GetNumLayers(); ++iLayer)
    {
        // Different layers refresh at different rates. Check if this layer is due.
        if(fv8Mgr.ShouldRefresh(iLayer, DXUTGetElapsedTime()))
        {
            // Pushing this layer as a render target returns a modified projection matrix.
            // Be sure to update your cull volume based on the new projection.
            D3DXMATRIX cam2NDC = fv8Mgr.PushLayer(iLayer, pd3dImmediateContext);

            // Clear the screen. You can do this yourself if you rather, the device has
            // been set to the appropriate offscreen rendertargets as of the PushLayer().
            fv8Mgr.Clear(pd3dImmediateContext, ClearColor);

            // Render the scene as normal, using the modified projection matrix. This
            // presumes the cull volume is generated from the camera transforms. Also,
            // LOD and effects quality can be reduced for increasing layer index.
            RenderInnerLoop(pd3dImmediateContext, cam2NDC, iLayer);

            // Finish this layer, resetting state to before the PushLayer() above.
            fv8Mgr.PopLayer(pd3dImmediateContext);
        }
    }
    // Now composite the layers into the destination render target (the back buffer here).
    ID3D11RenderTargetView* pRTV = DXUTGetD3D11RenderTargetView();
    fv8Mgr.FinishFrame(pd3dImmediateContext, pRTV);#endif TEST_CODE_ONLY
}
#endif // NEVER_DEFINE_SAMPLE_CODE_ONLY

#endif // FAUX_V8_MGR_H

